# Databricks notebook source
# https://danijar.com/variable-sequence-lengths-in-tensorflow/  : Variable length sequences to RNN

# COMMAND ----------

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob
import json
from datetime import datetime
import itertools

from sklearn.linear_model import LinearRegression

assert np.__version__ == "1.15.4" and pd.__version__ == "0.23.4", "np: {}, pd: {}".format(np.__version__, pd.__version__)

# COMMAND ----------

base_url = '/dbfs/mnt/renewablesstudents/user-write/'
#subset = False
#c1_tmp_files = glob.glob(base_url + 'circuit_1/DTS_new_format/*.parquet')  # For subset of data
#df_tmp = pd.read_parquet(df_tmp[0] if subset else base_url + 'circuit_1/DTS_new_format/', columns=['34327.421', 'timestamp'])

df_tmp = pd.read_csv(base_url + 'circuit_1/part1tmp.csv')
df_tmp['new_timestamp'] = pd.to_datetime(df_tmp.timestamp, unit='s')
df_tmp['sampling_interval'] = df_tmp.timestamp - df_tmp.timestamp.shift(1)
df_tmp = df_tmp.iloc[1:,:]
df_tmp = df_tmp[df_tmp.new_timestamp < pd.to_datetime('2019 2 1')]
print('temp loaded')

def df_from_json(filepath):
  with open(filepath) as file:
    js = json.load(file)
  return pd.DataFrame(js['timeSeriesList'][0]['timeSeries'])

df_current = df_from_json('/dbfs/mnt/renewablesstudents/dow-dts-v2/bazefield_data/90636_DOW-WFO-OS-132kV-30E01+R01 GT1-CurrentL3_AVG.json')[['t', 'v']]
df_current['t'] = df_current['t'] *1e-3
df_current['new_timestamp'] = pd.to_datetime(df_current.t, unit='s')
df_current = df_current[(df_current.t + 3000>= df_tmp.timestamp.iloc[0]) & (df_current.t <= df_tmp.timestamp.iloc[-1])]
df_current = df_current.set_index('new_timestamp').resample('10T').mean().reset_index()
df_current['v'] = df_current['v'].interpolate()
df_current['t'] = df_current['t'].interpolate()
print('current loaded')

df_seatemp = pd.read_json(base_url + 'from_bz/40406_DOW-EFS-WaterTemp.json')[['t', 'v']]
df_seatemp['t'] = df_seatemp['t'] *1e-3
df_seatemp['new_timestamp'] = pd.to_datetime(df_seatemp.t, unit='s')
df_seatemp = df_seatemp[(df_seatemp.t + 3000 >= df_tmp.timestamp.iloc[0]) & (df_seatemp.t <= df_tmp.timestamp.iloc[-1])]
print('seatemp loaded')

segment = '34327.421'
list(df_tmp)

# COMMAND ----------

# Model a : current interpolated to fit temperature sampling rate.
df_curr = df_current.copy()

#df_curr = df_curr[(df_curr.t >= df_tmp.timestamp.iloc[0]) & (df_curr.t <= df_tmp.timestamp.iloc[-1])]
df_curr = df_curr.set_index('t')

df_model_a =  df_tmp[[segment, 'timestamp']].set_index('timestamp').join(df_curr['v'], how='outer')
df_model_a = df_model_a.join(df_seatemp[['v', 't']].set_index('t'), how='outer', rsuffix='_temp')
df_model_a['v'] = df_model_a['v'].interpolate()
df_model_a['v_temp'] = df_model_a['v_temp'].interpolate()
df_model_a = df_model_a.dropna()

df_model_a = df_model_a.reset_index()
df_model_a.columns = ["timestamp", df_model_a.columns[1], 'current', 'sea_temp']
df_model_a['sampling_int'] = df_model_a.timestamp - df_model_a.timestamp.shift(1)
df_model_a.loc[0, 'sampling_int'] = 0

df_model_a = df_model_a.drop(['timestamp'], axis=1)

df_model_a.head()

input_df = df_model_a.loc[:, [segment, 'current', 'sea_temp', 'sampling_int']]
data = input_df.values
intervals = df_model_a.loc[:, 'sampling_int'].values
print(data.shape, intervals.shape)
input_df.head()

# COMMAND ----------

def MSE(preds, targets, normalizer=None):
  assert preds.shape == targets.shape, 'shape of predictions must match shape of targets!'
  if normalizer:
    preds = normalizer.undo_normalize(preds, preds=True)
    targets = normalizer.undo_normalize(targets, preds=True)
  return np.mean(np.square(preds-targets))      

def get_inputs_targets_intervals_AD(data, interval_list, input_steps, pred_steps, output_steps=1, pred_dim=0):
  size = len(data) - input_steps - output_steps + 1
  dim = data.shape[1]
  inputs = np.zeros((size, input_steps, dim))
  targets = np.zeros((size, pred_steps, 1))
  intervals = np.zeros((size, pred_steps, output_steps))
  
  too_long_intervals = []
  max_interval_length = 2218.0
  min_interval_length = 554.5
  
  for i in range(size):
    inputs[i] = data[i:i+input_steps,:]
    targets[i] = data[i+input_steps-pred_steps+1:i+input_steps+output_steps,pred_dim:pred_dim+1]
    intervals[i] = np.expand_dims(interval_list[i+input_steps-pred_steps+1:i+input_steps+output_steps],1)
    if intervals[i][-1] > max_interval_length or intervals[i][-1] < min_interval_length:
      too_long_intervals.append(i)
  print('{} samples removed with too long interval until prediction'.format(len(too_long_intervals)))
  inputs = np.delete(inputs, too_long_intervals, axis=0)
  targets = np.delete(targets, too_long_intervals, axis=0)
  intervals = np.delete(intervals, too_long_intervals, axis=0)
  return inputs, targets, intervals        
        
# Z-normalize data
class normalizer:
  means = -1
  stds = -1
  
  def __init__(self, temp_data, curr_data, sea_data, interval_data=np.array([])):
    means = [np.mean(temp_data), np.mean(curr_data), np.mean(sea_data)]
    stds = [np.std(temp_data), np.std(curr_data), np.std(sea_data)]
    if not interval_data.size==0:
      means.append(np.mean(interval_data))
      stds.append(np.std(interval_data))
    self.means = np.array(means)
    self.stds = np.array(stds)
    
  def normalize(self, data, preds=False):
    normalized = (data - self.means)/ self.stds if not preds else (data - self.means[0])/ self.stds[0]
    assert normalized.shape == data.shape, "new shape {} not equal to old shape {}".format(normalized.shape, data.shape)
    return normalized

  def undo_normalize(self, data, preds=False):
    un_normalized = data*self.stds + self.means if not preds else data*self.stds[0] + self.means[0]
    assert un_normalized.shape == data.shape, "new shape {} not equal to old shape {}".format(un_normalized.shape, data.shape)
    return un_normalized

class int_normalizer(normalizer):
    def __init__(self, data):
      self.means = np.mean(data, axis=0)
      self.stds = np.std(data, axis=0)

# COMMAND ----------

def range_array(df, column, i, fold_data):
  return df[(df.new_timestamp >= fold_data[i]['start_date']) & (df.new_timestamp <= fold_data[i]['end_date'])][column].values

def make_normalizer_input_array(df, val_fold, test_fold, fold_data, column):
  return np.concatenate([range_array(df, column, i, fold_data) for i in range(n_folds) if i !=val_fold and i != test_fold])

# COMMAND ----------

n_input_steps = 50
n_folds = 5
n_dims = 4
fold_size = len(df_tmp)//n_folds
fold_ranges = []

fold_data = []
for k in range(n_folds):
  fold_range = range(k*fold_size,(k+1)*fold_size) if k < n_folds-1 else range(k*fold_size, len(df_tmp))
  fold_ranges.append(fold_range)
  fold_inputs, fold_targets, fold_intervals = get_inputs_targets_intervals_AD(data[fold_range], intervals[fold_range], n_input_steps, 1)
  sd = df_tmp[['new_timestamp']].iloc[fold_range].iloc[0].values[0]
  ed = df_tmp[['new_timestamp']].iloc[fold_range].iloc[-1].values[0]
  fold_data.append({'inputs': fold_inputs, 'targets':fold_targets, 'intervals':fold_intervals, 'range':fold_range, 'start_date': sd, 'end_date':ed})
  print(fold_data[-1]['start_date'], fold_data[-1]['end_date'])
  print(fold_inputs.shape, fold_targets.shape, fold_intervals.shape)

# COMMAND ----------

is_testing = True
#n_epochs = 1000
#batch_size = 128
#initial_patience = 50

MSEs_train = []
MSEs_val = []
MSEs_test = []
MSEs_train_norm = []
MSEs_val_norm = []
MSEs_test_all = []


MSEs_val_int = []
MSEs_train_int = []

MSEs_test_int = []
MSEs_train_int = []
MSEs_val_int = []

all_MSEs_test_int = []
all_MSEs_val_int = []
all_MSEs_train_int = []

for test_fold in range(n_folds):
  if not is_testing:
    test_fold = -1
  for val_fold in range(n_folds):
    if test_fold != val_fold or not is_testing:

      val_inputs, val_targets, val_intervals = fold_data[val_fold]['inputs'], fold_data[val_fold]['targets'], fold_data[val_fold]['intervals']
      train_inputs = np.concatenate([fold_data[i]['inputs'] for i in range(n_folds) if i !=val_fold and i != test_fold])
      train_targets = np.concatenate([fold_data[i]['targets'] for i in range(n_folds) if i !=val_fold and i != test_fold])
      train_intervals = np.concatenate([fold_data[i]['intervals'] for i in range(n_folds) if i !=val_fold and i != test_fold])
      
      # Normalize data:
      data_normalizer = normalizer(make_normalizer_input_array(df_tmp, test_fold, val_fold, fold_data, segment),
                                   make_normalizer_input_array(df_current, test_fold, val_fold, fold_data, 'v'),
                                   make_normalizer_input_array(df_seatemp, test_fold, val_fold, fold_data, 'v'),
                                   make_normalizer_input_array(df_tmp, test_fold, val_fold, fold_data, 'sampling_interval'))
      interval_normalizer = int_normalizer(np.concatenate([intervals[fold_data[i]['range']] for i in range(n_folds) if i !=val_fold and i != test_fold]))
      val_inputs, val_targets, val_intervals = data_normalizer.normalize(val_inputs), data_normalizer.normalize(val_targets, preds=True), interval_normalizer.normalize(val_intervals)
      train_inputs, train_targets, train_intervals = data_normalizer.normalize(train_inputs), data_normalizer.normalize(train_targets, preds=True), interval_normalizer.normalize(train_intervals)

      if is_testing:
        test_inputs, test_targets, test_intervals = fold_data[test_fold]['inputs'], fold_data[test_fold]['targets'], fold_data[test_fold]['intervals']
        test_inputs, test_targets, test_intervals = data_normalizer.normalize(test_inputs), data_normalizer.normalize(test_targets, preds=True), interval_normalizer.normalize(test_intervals)
      
      reg_with_interval = LinearRegression().fit(np.concatenate([train_inputs.reshape(-1, n_input_steps*n_dims),
                                                                 train_intervals.reshape(-1,1)], axis=1), train_targets.flatten())

      # Get predictions on val data.
      val_preds_int = reg_with_interval.predict(np.concatenate([val_inputs.reshape(-1, n_input_steps*n_dims), val_intervals.reshape(-1,1)], axis=1))
      train_preds_int = reg_with_interval.predict(np.concatenate([train_inputs.reshape(-1, n_input_steps*n_dims), train_intervals.reshape(-1,1)], axis=1))
      test_preds_int = reg_with_interval.predict(np.concatenate([test_inputs.reshape(-1, n_input_steps*n_dims), test_intervals.reshape(-1,1)], axis=1))

      # Undo normalization, calculate MSE

      MSEs_val_int.append(MSE(val_preds_int, val_targets.flatten(), data_normalizer))
      MSEs_train_int.append(MSE(train_preds_int, train_targets.flatten(), data_normalizer))
      MSEs_test_int.append(MSE(test_preds_int, test_targets.flatten(), data_normalizer))
      print('t{}v{}: train: {}, val: {}, test: {}'.format(test_fold, val_fold, MSEs_train_int[-1], MSEs_val_int[-1], MSEs_test_int[-1]))
  if not is_testing:
    break
  else:
    all_MSEs_test_int.append(MSEs_test_int)
    all_MSEs_val_int.append(MSEs_val_int)
    all_MSEs_train_int.append(MSEs_train_int)
    MSEs_test_int = []
    MSEs_train_int = []
    MSEs_val_int = []

    
print("With interval:")
print('Test & {} & {} & {} & {} & {}'.format(np.mean(all_MSEs_test_int[0]),np.mean(all_MSEs_test_int[1]),np.mean(all_MSEs_test_int[2]),np.mean(all_MSEs_test_int[3]),np.mean(all_MSEs_test_int[4])))
print('Val & {} & {} & {} & {} & {}'.format(np.mean(all_MSEs_val_int[0]),np.mean(all_MSEs_val_int[1]),np.mean(all_MSEs_val_int[2]),np.mean(all_MSEs_val_int[3]),np.mean(all_MSEs_val_int[4])))
print('Train & {} & {} & {} & {} & {}'.format(np.mean(all_MSEs_train_int[0]),np.mean(all_MSEs_train_int[1]),np.mean(all_MSEs_train_int[2]),np.mean(all_MSEs_train_int[3]),np.mean(all_MSEs_train_int[4])))

# COMMAND ----------

print('LR & {} & {} & {}'.format(np.mean(all_MSEs_train_int), np.mean(all_MSEs_val_int), np.mean(all_MSEs_test_int)))

# COMMAND ----------

for i, t in enumerate(all_MSEs_test_int):
  s = str(i+1)
  
  t_copy = t.copy()
  t_copy.insert(i, 0)
  
  for ind in t_copy:
    s += ' & {}'.format(ind if ind else '\\-')
  print(s + '\\\ \hline')

# COMMAND ----------

for i, t in enumerate(all_MSEs_test_int):
  s = str(i+1)
  
  t_copy = t.copy()
  t_copy.insert(i, 0)
  
  for ind in t_copy:
    s += ' & {}'.format(ind if ind else '\\-')
  print(s + '\\\ \hline')

# COMMAND ----------

np.array(all_MSEs_test_int).flatten()

# COMMAND ----------

all_MSEs_test_int

# COMMAND ----------


