# Databricks notebook source
# https://danijar.com/variable-sequence-lengths-in-tensorflow/  : Variable length sequences to RNN

# COMMAND ----------

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob
import json
from datetime import datetime
from warnings import warn
import itertools

#tf imports
import tensorflow as tf
from tensorflow.nn import rnn_cell
from tensorflow.saved_model import tag_constants

assert np.__version__ == "1.15.4" and pd.__version__ == "0.23.4", "np: {}, pd: {}".format(np.__version__, pd.__version__)

# COMMAND ----------

log_dir = '/dbfs/mnt/renewablesstudents/user-write/logfiles/p1/experiments_2/'
model_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p1/experiments_2/'

if not os.path.exists(log_dir):
    os.makedirs(log_dir)
    print('dir made: {}'.format(log_dir))
  
if not os.path.exists(model_dir):
    os.makedirs(model_dir)
    print('dir made: {}'.format(model_dir))

# COMMAND ----------

base_url = '/dbfs/mnt/renewablesstudents/user-write/'
#subset = False
#c1_tmp_files = glob.glob(base_url + 'circuit_1/DTS_new_format/*.parquet')  # For subset of data
#df_tmp = pd.read_parquet(df_tmp[0] if subset else base_url + 'circuit_1/DTS_new_format/', columns=['34327.421', 'timestamp'])

df_tmp = pd.read_csv(base_url + 'circuit_1/part1tmp.csv')
df_tmp['new_timestamp'] = pd.to_datetime(df_tmp.timestamp, unit='s')
df_tmp['sampling_interval'] = df_tmp.timestamp - df_tmp.timestamp.shift(1)
df_tmp = df_tmp.iloc[1:,:]
df_tmp = df_tmp[df_tmp.new_timestamp < pd.to_datetime('2019 2 1')]
print('temp loaded')

def df_from_json(filepath):
  with open(filepath) as file:
    js = json.load(file)
  return pd.DataFrame(js['timeSeriesList'][0]['timeSeries'])

df_current = df_from_json('/dbfs/mnt/renewablesstudents/dow-dts-v2/bazefield_data/90636_DOW-WFO-OS-132kV-30E01+R01 GT1-CurrentL3_AVG.json')[['t', 'v']]
df_current['t'] = df_current['t'] *1e-3
df_current['new_timestamp'] = pd.to_datetime(df_current.t, unit='s')
df_current = df_current[(df_current.t + 3000>= df_tmp.timestamp.iloc[0]) & (df_current.t <= df_tmp.timestamp.iloc[-1])]
df_current = df_current.set_index('new_timestamp').resample('10T').mean().reset_index()
df_current['v'] = df_current['v'].interpolate()
df_current['t'] = df_current['t'].interpolate()
print('current loaded')

df_seatemp = pd.read_json(base_url + 'from_bz/40406_DOW-EFS-WaterTemp.json')[['t', 'v']]
df_seatemp['t'] = df_seatemp['t'] *1e-3
df_seatemp['new_timestamp'] = pd.to_datetime(df_seatemp.t, unit='s')
df_seatemp = df_seatemp[(df_seatemp.t + 3000 >= df_tmp.timestamp.iloc[0]) & (df_seatemp.t <= df_tmp.timestamp.iloc[-1])]
print('seatemp loaded')

segment = '34327.421'
list(df_tmp)

# COMMAND ----------



# COMMAND ----------

tmp_sub = df_tmp.loc[df_tmp.new_timestamp <= pd.to_datetime('2017-11-01 01:18:00')]

# COMMAND ----------

sea_sub = df_seatemp.loc[df_seatemp.new_timestamp <= pd.to_datetime('2017-11-01 01:18:00')]

# COMMAND ----------

curr_sub = df_current.loc[df_current.new_timestamp <= pd.to_datetime('2017-11-01 01:18:00')].iloc[0:]
curr_sub.t = curr_sub.t + 0.001

# COMMAND ----------

tmp_sub[['new_timestamp', segment]]

# COMMAND ----------

sea_sub

# COMMAND ----------

curr_sub

# COMMAND ----------

# Set up tf graph
tf.reset_default_graph()

learning_rate = 0.01
n_hidden = 64
dim_input = 5
n_input_steps = 50
n_output_steps = 1
pred_steps = 1
max_length = 197

assert n_input_steps >= pred_steps, 'input steps must be > pred steps'

# Graph input
x = tf.placeholder("float", [None, max_length, dim_input], name='x')
t = tf.placeholder("float", [None, pred_steps, n_output_steps], name='t')
y = tf.placeholder("float", [None, pred_steps ,n_output_steps], name='y')
l = tf.placeholder("int32", [None], name='l')

# Define weights
weights = {
    'out': tf.Variable(tf.random_normal([n_hidden+1,n_output_steps]))  # +1 because of concatenated time until prediction input
}
biases = {
    'out': tf.Variable(tf.zeros([n_output_steps]))
}

def last_relevant(output, length):
  batch_size = tf.shape(output)[0]
  max_length = tf.shape(output)[1]
  out_size = int(output.get_shape()[2])
  index = tf.range(0, batch_size) * max_length + (length - 1)
  flat = tf.reshape(output, [-1, out_size])
  relevant = tf.gather(flat, index)
  return relevant

cell = rnn_cell.LSTMCell(n_hidden, name='LSTM_cell')

outputs, _ = tf.nn.dynamic_rnn(cell, x, dtype=tf.float32, sequence_length=l) 

#print(outputs[:, -pred_steps:, :].shape)
#print(last_relevant(outputs, l).shape)
#print(tf.reshape(last_relevant(outputs, l),(-1,1,n_hidden)).shape)
#print(tf.reshape(t, (-1,1)).shape)

#preds = tf.matmul(tf.reshape(tf.concat([outputs[:, -pred_steps:, :], t], axis=2), (-1, (n_hidden+1))), weights['out']) + biases['out']
preds = tf.matmul(tf.concat([last_relevant(outputs, l), tf.reshape(t, (-1,1))], axis=1), weights['out']) + biases['out']
#print(preds.shape)
#preds = tf.matmul(tf.reshape(tf.concat([tf.reshape(last_relevant(outputs, l),(-1,1,n_hidden)), t], axis=2), (-1, (n_hidden+1))), weights['out']) + biases['out']
#print(preds.shape)
preds = tf.reshape(preds, (-1, pred_steps, n_output_steps), name='preds')
#print(preds.shape)

# Define loss and optimizer
loss_op = tf.losses.mean_squared_error(labels=y, predictions=preds)
optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
train_op = optimizer.minimize(loss_op)

init = tf.global_variables_initializer()
saver = tf.train.Saver()

# COMMAND ----------

def MSE(preds, targets, normalizer=None, model=None):
  assert preds.shape == targets.shape, 'shape of predictions must match shape of targets!'
  if normalizer:
    preds = normalizer.undo_normalize(preds, preds=True)
    targets = normalizer.undo_normalize(targets, preds=True)
  if model:
    preds = model.undo_normalize(preds)
    targets = model.undo_normalize(targets)
  return np.mean(np.square(preds-targets))

def write_to_log(text, filename):
  if not filename:
    print('not written to log: ' + text)
  else:
    with open(filename, 'a') as file:
      file.write(str(datetime.now()) + ':   ' + text+'\n')
      print(text)
      
def save_params(sess, epoch, test_fold, val_fold, model_filename, log_filename):
  if model_filename:
    epoch_dir = model_filename + '/m_t{}v{}/epoch_{}/'.format(test_fold, val_fold, epoch)
    if not os.path.exists(epoch_dir):
        os.makedirs(epoch_dir)
    fp = saver.save(sess, epoch_dir + 'model.ckpt')
    write_to_log('\tparams saved to {}'.format(fp), log_filename)
    return fp
  else: return None
  
def save_full_model(model_filename, params_path, test_fold, val_fold, val_lengths, log_filename, val_inputs, val_intervals, val_targets, min_val_loss):
  with tf.Session() as sess:
        sess.run(init)
        write_to_log('Saving full model', log_filename)
        saver.restore(sess, params_path)
        write_to_log('params loaded from {}'.format(params_path), log_filename)
        val_loss = MSE(sess.run(preds, feed_dict ={x:val_inputs, t:val_intervals, l:val_lengths})[:,-1,:], val_targets[:,-1,:])
        assert val_loss == min_val_loss, "val loss {} should be equal to min val loss {}".format(val_loss, min_val_loss)
        full_model_dir = model_filename + '/full_model_t{}v{}'.format(test_fold, val_fold) if test_fold else model_filename + '/full_model_v{}'.format(val_fold)
        tf.saved_model.simple_save(sess,
                                  full_model_dir,
                                  inputs={"x":x, "t":t, "y":y},
                                  outputs={"preds":preds, "loss_op":loss_op})
        write_to_log("Full model saved to {}".format(full_model_dir),log_filename)

def get_inputs_targets_intervals_AD(data, interval_list, input_steps, pred_steps, output_steps=1, pred_dim=0):
  size = len(data) - input_steps - output_steps + 1
  dim = data.shape[1]
  inputs = np.zeros((size, input_steps, dim))
  targets = np.zeros((size, pred_steps, 1))
  intervals = np.zeros((size, pred_steps, output_steps))
  
  too_long_intervals = []
  max_interval_length = 2218.0
  min_interval_length = 554.5
  
  for i in range(size):
    inputs[i] = data[i:i+input_steps,:]
    targets[i] = data[i+input_steps-pred_steps+1:i+input_steps+output_steps,pred_dim:pred_dim+1]
    intervals[i] = np.expand_dims(interval_list[i+input_steps-pred_steps+1:i+input_steps+output_steps],1)
    if intervals[i][-1] > max_interval_length or intervals[i][-1] < min_interval_length:
      too_long_intervals.append(i)
  print('{} samples removed with too long interval until prediction'.format(len(too_long_intervals)))
  inputs = np.delete(inputs, too_long_intervals, axis=0)
  targets = np.delete(targets, too_long_intervals, axis=0)
  intervals = np.delete(intervals, too_long_intervals, axis=0)
  return inputs, targets, intervals        
        
# Z-normalize data
class normalizer:
  means = -1
  stds = -1
  
  def __init__(self, temp_data, curr_data, sea_data, interval_data=np.array([])):
    means = [np.mean(temp_data), np.mean(curr_data), np.mean(sea_data)]
    stds = [np.std(temp_data), np.std(curr_data), np.std(sea_data)]
    if not interval_data.size==0:
      means.append(np.mean(interval_data))
      stds.append(np.std(interval_data))
    self.means = np.array(means)
    self.stds = np.array(stds)
    
  def normalize(self, data, preds=False):
    normalized = (data - self.means)/ self.stds if not preds else (data - self.means[0])/ self.stds[0]
    assert normalized.shape == data.shape, "new shape {} not equal to old shape {}".format(normalized.shape, data.shape)
    return normalized

  def undo_normalize(self, data, preds=False):
    un_normalized = data*self.stds + self.means if not preds else data*self.stds[0] + self.means[0]
    assert un_normalized.shape == data.shape, "new shape {} not equal to old shape {}".format(un_normalized.shape, data.shape)
    return un_normalized

class int_normalizer(normalizer):
    def __init__(self, data):
      self.means = np.mean(data, axis=0)
      self.stds = np.std(data, axis=0)

# COMMAND ----------

# input normalized data!
def one_dim(df_tmp, df_curr, df_sea):
  df_curr = df_curr.loc[:,['v', 't']]
  df_curr = df_curr.set_index('t')
  df_tmp = df_tmp.loc[:, [segment, 'timestamp']].set_index('timestamp') 
  df_curr = df_curr[(df_curr.index >= df_tmp.index[0]) & (df_curr.index <= df_tmp.index[-1])]
 
  df = df_tmp.join(df_curr, how='outer')
  df = df.reset_index()
  df.columns = ['t', df.columns[1],  df.columns[2]]
  df_sea = df_sea[(df_sea.t >= df_tmp.index[0]) & (df_sea.t <= df_tmp.index[-1])]
  df = df.set_index('t').join(df_sea.loc[:,['t','v']].set_index('t'), how='outer', rsuffix='_st', lsuffix='_c')
  df = df.reset_index()
  df.t = df.t.astype('int')
  df['interval_s'] = df.t - df.shift(1).t

  df['st_m'] = df['v_st'].notna().astype('int')
  df['t_m'] = df[segment].notna().astype('int')
  df['c_m'] = df['v_c'].notna().astype('int')
  
  print(df.head())
  
  df.loc[df['t_m'], 'st_m'] = 0
  df.loc[df['t_m'], 'c_m'] = 0
  df.loc[df['st_m'], 'c_m'] = 0
  print(df.head())
  df['values'] = df['v_c']
  df['values'].update(df['v_st'])
  df['values'].update(df[segment])
  print(df.head())
  df.loc[0,'interval_s'] = 0
  return df[['values', 'st_m', 't_m', 'c_m', 'interval_s']], df.iloc[-1].t

df2, ti2 = one_dim(df_tmp.iloc[1:51,:],  df_current, df_seatemp)
print(ti2)
df2

# COMMAND ----------

one_dim(tmp_sub, curr_sub, sea_sub)[0]

# COMMAND ----------

n_folds = 5
fold_size = len(df_tmp)//n_folds
fold_ranges = []

fold_data = []
for k in range(n_folds):
  fold_range = range(k*fold_size,(k+1)*fold_size) if k < n_folds-1 else range(k*fold_size, len(df_tmp))
  fold_ranges.append(fold_range)
  fold_temps = df_tmp.iloc[fold_range,:]
  sd = df_tmp[['new_timestamp']].iloc[fold_range].iloc[0].values[0]
  ed = df_tmp[['new_timestamp']].iloc[fold_range].iloc[-1].values[0]
  fold_currents = df_current.loc[(df_current.new_timestamp >= sd) & (df_current.new_timestamp <= ed)]
  fold_sea_temps = df_seatemp.loc[(df_seatemp.new_timestamp >= sd) & (df_seatemp.new_timestamp <= ed)]
  fold_data.append({'fold_temps':fold_temps, 'fold_currents':fold_currents, 'fold_sea_temps':fold_sea_temps, 'start_date': sd, 'end_date':ed})
  print(fold_data[-1]['start_date'], fold_data[-1]['end_date'])
  print(fold_temps.shape, fold_currents.shape, fold_currents.shape)

# COMMAND ----------

#plt.close()
#
#fig, (ax1,ax2, ax3, ax4) = plt.subplots(4,1, figsize=(15,5), sharex=True)
#ax1.plot(input_df[segment])
#ax2.plot(input_df['current'])
#ax3.plot(input_df['sea_temp'])
#ax4.plot(intervals)
#for r in fold_ranges:
#  for a in [ax1, ax2, ax3, ax4]:
#    a.axvline(r.stop, color='r')
#display(plt.show())

# COMMAND ----------

pd.options.mode.chained_assignment = "raise"
class Model_e:
  def __init__(self, test_fold, val_fold, fold_data, n_folds, input_steps=50, max_length=197, input_dim=5):
    self.test_fold = test_fold
    self.val_fold = val_fold
    train_temps = np.concatenate([fold_data[i]['fold_temps'][segment].values for i in range(n_folds) if i !=val_fold and i != test_fold])
    train_currs = np.concatenate([fold_data[i]['fold_currents']['v'].values for i in range(n_folds) if i !=val_fold and i != test_fold])
    train_sea = np.concatenate([fold_data[i]['fold_sea_temps']['v'].values for i in range(n_folds) if i !=val_fold and i != test_fold])
    train_intervals = np.concatenate([fold_data[i]['fold_temps']['sampling_interval'].values for i in range(n_folds) if i !=val_fold and i != test_fold])
    self.temp_mean = np.mean(train_temps)
    self.temp_std = np.std(train_temps)
    self.curr_mean = np.mean(train_currs)
    self.curr_std = np.std(train_currs)
    self.sea_mean = np.mean(train_sea)
    self.sea_std = np.std(train_sea)
    self.int_mean = -1
    self.int_std = -1
    
    self.input_steps = input_steps
    self.max_length = max_length
    self.fold_data = fold_data
    self.input_dim = input_dim
    
  def get_inputs_multiple_folds(self, folds):
    inputs = [None for _ in range(len(folds))]
    targets = [None for _ in range(len(folds))]
    intervals = [None for _ in range(len(folds))]
    lengths = [None for _ in range(len(folds))]
    for index, fold in enumerate(folds):
      assert fold != self.test_fold and fold != self.val_fold, 'should not be test or val fold'
      inputs[index], targets[index], intervals[index], lengths[index] = self.__get_fold_inputs(fold)
    inputs, targets, intervals, lengths = np.concatenate(inputs), np.concatenate(targets), np.concatenate(intervals), np.concatenate(lengths)
    print(inputs.shape, targets.shape, intervals.shape, lengths.shape)
    self.int_mean = np.mean(intervals)
    self.int_std = np.std(intervals)
    intervals = (intervals - self.int_mean) / self.int_mean
    return inputs, targets, intervals, lengths
       
  def get_inputs(self, fold):
    assert fold == self.test_fold or fold == self.val_fold, 'should be test or val fold'
    if self.int_mean == -1:
      raise ValueError('need to run get_inputs_multiple_folds on train data to set interval mean and std')
    inputs, targets, intervals, lengths = self.__get_fold_inputs(fold)
    intervals = (intervals - self.int_mean) / self.int_mean
    return inputs, targets, intervals, lengths
    
  def __get_fold_inputs(self, fold):
    temp_df = self.fold_data[fold]['fold_temps'].copy()
    current_df = self.fold_data[fold]['fold_currents'].copy()
    seatemp_df = self.fold_data[fold]['fold_sea_temps'].copy()
    
    current_df.t = current_df.t + 0.001
    seatemp_df.t = seatemp_df.t + 0.002
    current_df.v = (current_df.v - self.curr_mean) / self.curr_std
    temp_df[segment] = (temp_df[segment] - self.temp_mean) / self.temp_std
    seatemp_df.v = (seatemp_df.v - self.sea_mean) / self.sea_std
    
    size = len(temp_df) - self.input_steps - 1
    inputs = np.zeros((size, self.max_length, self.input_dim))
    targets = np.zeros((size, 1, 1))
    intervals = np.zeros((size, 1, 1))
    input_lengths = np.zeros((size), dtype='int')
    
    too_long_intervals = []
    max_interval_length = 2218.0
    min_interval_length = 554.5
    
    for i in range(size):  
      input_temps = temp_df.iloc[i:i+self.input_steps, :]
      input_vals, last_time = one_dim(input_temps, current_df, seatemp_df)
      input_lengths[i] = len(input_vals)
      inputs[i, :input_vals.shape[0],:] = input_vals.values
      targets[i] = temp_df.iloc[i+self.input_steps, :][segment]
      intervals[i] = temp_df.iloc[i+self.input_steps, :]['timestamp'] -  last_time
      if temp_df.iloc[i+self.input_steps, :]['timestamp'] - temp_df.iloc[i+self.input_steps-1, :]['timestamp'] > max_interval_length or temp_df.iloc[i+self.input_steps, :]['timestamp'] - temp_df.iloc[i+self.input_steps-1, :]['timestamp'] < min_interval_length:
        too_long_intervals.append(i)
    print('{} samples removed with too long interval until prediction'.format(len(too_long_intervals)))
    inputs = np.delete(inputs, too_long_intervals, axis=0)
    targets = np.delete(targets, too_long_intervals, axis=0)
    intervals = np.delete(intervals, too_long_intervals, axis=0)
    input_lengths = np.delete(input_lengths, too_long_intervals, axis=0)
    return inputs, targets, intervals, input_lengths #, targets, intervals
    
  def undo_normalize(self, temp_preds):
    return (temp_preds * self.temp_std) + self.temp_mean
  
#model_e = Model_e(0, 1, fold_data, 5)
#model_e.get_inputs_multiple_folds([2, 3, 4])
#model_e.get_inputs(1)

# COMMAND ----------

is_testing = True
n_epochs = 1000
batch_size = 128
initial_patience = 50

MSEs_train = []
MSEs_val = []
MSEs_test = []
MSEs_train_norm = []
MSEs_val_norm = []
MSEs_test_all = []

file_suffix = 'model_E_v2'
log_filename = log_dir + file_suffix + '.txt' if file_suffix else None
model_filename = model_dir + file_suffix if file_suffix else None
if log_filename: assert not os.path.isfile(log_filename), "path " + log_filename + " is already in use"
if not file_suffix:
  warn("Training model without saving or logging")
else:
  write_to_log('writing log to {} and models to {}'.format(log_filename, model_filename), log_filename)

args = {'lr': learning_rate,
        'n hidden': str(n_hidden),
        'dim input':dim_input,
        'input steps':n_input_steps,
        'output steps':n_output_steps,
        'pred steps':pred_steps,
        'n folds': n_folds,
        'n epochs': n_epochs,
        'batch size': batch_size,
        'patience': initial_patience,
       'fold ranges': fold_ranges,
       'max length': max_length}
write_to_log(str(args), log_filename)
  
for test_fold in range(n_folds):
  if not is_testing:
    test_fold = -1
  for val_fold in range(n_folds):
    if test_fold != val_fold or not is_testing:
      write_to_log('test fold {} val fold {}'.format(test_fold, val_fold), log_filename)
      model_e = Model_e(test_fold, val_fold, fold_data, n_folds)
      train_inputs, train_targets, train_intervals, train_lengths = model_e.get_inputs_multiple_folds([i for i in range(n_folds) if i !=val_fold and i != test_fold])
      val_inputs, val_targets, val_intervals, val_lengths = model_e.get_inputs(val_fold)
      if is_testing:
        test_inputs, test_targets, test_intervals, test_lengths = model_e.get_inputs(test_fold)
      epoch_indicies = [i for i in range(len(train_inputs))]
      min_val_loss, min_train_loss = np.inf, np.inf
      patience = initial_patience
      
      with tf.Session() as sess:
        sess.run(init)
        for epoch in range(n_epochs):
          np.random.shuffle(epoch_indicies)
          epoch_inputs, epoch_targets, epoch_t, epoch_l = train_inputs[epoch_indicies], train_targets[epoch_indicies], train_intervals[epoch_indicies], train_lengths[epoch_indicies]
          for batch in range(0, len(train_inputs), batch_size):
            batch_x, batch_y, batch_t, batch_l  = epoch_inputs[batch:batch+batch_size], epoch_targets[batch:batch+batch_size], epoch_t[batch:batch+batch_size], epoch_l[batch:batch+batch_size]
            sess.run(train_op, feed_dict={x:batch_x, t:batch_t, y:batch_y, l:batch_l})
          epoch_val_loss = MSE(sess.run(preds, feed_dict ={x:val_inputs, t:val_intervals, l:val_lengths})[:,-1,:], val_targets[:,-1,:], normalizer=False, model=model_e)
          epoch_train_loss = sess.run(loss_op, feed_dict ={x:train_inputs, t:train_intervals, y:train_targets[:,:,:], l:train_lengths})
          if epoch_val_loss < min_val_loss and patience:
            min_val_loss = epoch_val_loss
            val_preds_epoch = sess.run(preds, feed_dict={x:val_inputs, t:val_intervals, l:val_lengths})
            train_preds_epoch = sess.run(preds, feed_dict={x:train_inputs, t:train_intervals, l:train_lengths})
            train_MSE_epoch = MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:], normalizer=False, model=model_e)
            test_MSE = MSE(sess.run(preds, feed_dict={x:test_inputs, t:test_intervals, l:test_lengths})[:,-1,:], test_targets[:,-1,:], normalizer=False, model=model_e) if is_testing else -1
            min_train_loss = epoch_train_loss
            write_to_log("\tEpoch {} finished with training loss {}. For 1-step preds training loss is {} and validation loss {} test {}".format(epoch, min_train_loss, train_MSE_epoch, min_val_loss, test_MSE), log_filename)
            params_path = None #save_params(sess, epoch, test_fold, val_fold, model_filename, False)
            patience = initial_patience
          elif patience:
            patience -= 1
          else:
            write_to_log("Stopped training at epoch {} due to patience running out".format(epoch), log_filename)
            break
      MSE_train = MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:], normalizer=False, model=model_e)
      MSE_val = MSE(val_preds_epoch[:,-1,:], val_targets[:,-1,:], normalizer=False, model=model_e)
      MSEs_train.append(MSE_train)
      MSEs_val.append(MSE_val)
      MSEs_test.append(test_MSE)
      MSEs_train_norm.append(MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:]))
      MSEs_val_norm.append(MSE(val_preds_epoch[:,-1,:], val_targets[:,-1,:]))
      write_to_log("Accuracy for one step preds train:{} val:{}\n".format(MSE_train, MSE_val), log_filename)
      if params_path:
        save_full_model(model_filename, params_path, test_fold, val_fold, val_lengths, log_filename, val_inputs, val_intervals, val_targets, MSEs_val_norm[-1])
  if not is_testing:
    break
  else:
    MSEs_test_all.append(MSEs_test)
    MSEs_test = []
write_to_log('normalized average: train:{}, val:{}'.format(np.mean(MSEs_train_norm), np.mean(MSEs_val_norm)), log_filename)
write_to_log('average MSEs across all folds train:{} val:{} test:{}'.format(np.mean(MSEs_train), np.mean(MSEs_val), np.mean(MSEs_test)), log_filename)
write_to_log('MSE std across all folds train:{} val:{} test:{}'.format(np.std(MSEs_train), np.std(MSEs_val), np.std(MSEs_test)), log_filename)
if not is_testing:
  write_to_log('MSE val for each fold: {} & {} & {} & {} & {}'.format(MSEs_val[0],MSEs_val[1],MSEs_val[2],MSEs_val[3],MSEs_val[4]), log_filename)
else:
  write_to_log('average MSE test for each fold: {} & {} & {} & {} & {}'.format(np.mean(MSEs_test_all[0]),
                                                                               np.mean(MSEs_test_all[1]),
                                                                               np.mean(MSEs_test_all[2]),
                                                                               np.mean(MSEs_test_all[3]),
                                                                               np.mean(MSEs_test_all[4])), log_filename)
write_to_log('\n-----------------------------------------------------------------\n', log_filename)

# COMMAND ----------

# Restore full model
tf.reset_default_graph()
with tf.Session() as sess:
  tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], full_model_dir)
  loaded_x = tf.get_default_graph().get_tensor_by_name('x:0')
  loaded_t = tf.get_default_graph().get_tensor_by_name('t:0')
  loaded_preds = tf.get_default_graph().get_tensor_by_name('preds:0')
  lp = sess.run(loaded_preds, feed_dict={loaded_x:val_inputs, loaded_t:val_input_intervals})[:,-1,:]
  print('val MSE', MSE(sess.run(loaded_preds, feed_dict={loaded_x:val_inputs, loaded_t:val_input_intervals})[:,-1,:], val_targets[:,-1,:], unormalize=True))

# COMMAND ----------

dbutils.fs.ls('dbfs:/mnt/renewablesstudents/user-write/logfiles/p1/experiments_2/model_E.txt')

# COMMAND ----------

with open('/dbfs/mnt/renewablesstudents/user-write/logfiles/p1/experiments_2/model_E.txt') as file:
  for line in file:
    print(line)

# COMMAND ----------

fold_ranges

# COMMAND ----------


