# Databricks notebook source
import random

# COMMAND ----------

n_blocks= 12
n_folds = 5

# COMMAND ----------

base_url = '/dbfs/mnt/renewablesstudents/user-write/'
#subset = False
#c1_tmp_files = glob.glob(base_url + 'circuit_1/DTS_new_format/*.parquet')  # For subset of data
#df_tmp = pd.read_parquet(df_tmp[0] if subset else base_url + 'circuit_1/DTS_new_format/', columns=['34327.421', 'timestamp'])

df_tmp = pd.read_csv(base_url + 'circuit_1/part1tmp.csv')
df_tmp['new_timestamp'] = pd.to_datetime(df_tmp.timestamp, unit='s')
df_tmp['sampling_interval'] = df_tmp.timestamp - df_tmp.timestamp.shift(1)
df_tmp = df_tmp.iloc[1:,:]
df_tmp = df_tmp[df_tmp.new_timestamp < pd.to_datetime('2019 2 1')]
print('temp loaded into df_tmp')
df_tmp = df_tmp.reset_index(drop=True)

def df_from_json(filepath):
  with open(filepath) as file:
    js = json.load(file)
  return pd.DataFrame(js['timeSeriesList'][0]['timeSeries'])

df_current = df_from_json('/dbfs/mnt/renewablesstudents/dow-dts-v2/bazefield_data/90636_DOW-WFO-OS-132kV-30E01+R01 GT1-CurrentL3_AVG.json')[['t', 'v']]
df_current['t'] = df_current['t'] *1e-3
df_current['new_timestamp'] = pd.to_datetime(df_current.t, unit='s')
df_current = df_current[(df_current.t + 3000>= df_tmp.timestamp.iloc[0]) & (df_current.t <= df_tmp.timestamp.iloc[-1])]
df_current = df_current.set_index('new_timestamp').resample('10T').mean().reset_index()
df_current['v'] = df_current['v'].interpolate()
df_current['t'] = df_current['t'].interpolate()
print('current loaded into df_current')

df_seatemp = pd.read_json(base_url + 'from_bz/40406_DOW-EFS-WaterTemp.json')[['t', 'v']]
df_seatemp['t'] = df_seatemp['t'] *1e-3
df_seatemp['new_timestamp'] = pd.to_datetime(df_seatemp.t, unit='s')
df_seatemp = df_seatemp[(df_seatemp.t + 3000 >= df_tmp.timestamp.iloc[0]) & (df_seatemp.t <= df_tmp.timestamp.iloc[-1])]
print('seatemp loaded df_seatemp')

segment = '34327.421'
list(df_tmp)

# COMMAND ----------

def set_dirs(name):
  name = name if name.endswith('/') else name +'/'
  
  log_dir = '/dbfs/mnt/renewablesstudents/user-write/logfiles/p1/' + name 
  model_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p1/' + name

  if not os.path.exists(log_dir):
      os.makedirs(log_dir)
      print('dir made: {}'.format(log_dir))

  #if not os.path.exists(model_dir):
  #    os.makedirs(model_dir)
  #    print('dir made: {}'.format(model_dir))
  return log_dir, model_dir

# COMMAND ----------

def MSE(preds, targets, normalizer=None, model=None):
  assert preds.shape == targets.shape, 'shape of predictions must match shape of targets!'
  if normalizer:
    preds = normalizer.undo_normalize(preds, preds=True)
    targets = normalizer.undo_normalize(targets, preds=True)
  if model:
    preds = model.undo_normalize(preds)
    targets = model.undo_normalize(targets)
  return np.mean(np.square(preds-targets))

def write_to_log(text, filename):
  if not filename:
    print('not written to log: ' + text)
  else:
    with open(filename, 'a') as file:
      file.write(str(datetime.now()) + ':   ' + text+'\n')
      print(text)
      
def save_params(sess, epoch, test_fold, val_fold, model_filename, log_filename):
  if model_filename:
    epoch_dir = model_filename + '/m_t{}v{}/epoch_{}/'.format(test_fold, val_fold, epoch)
    if not os.path.exists(epoch_dir):
        os.makedirs(epoch_dir)
    fp = saver.save(sess, epoch_dir + 'model.ckpt')
    write_to_log('\tparams saved to {}'.format(fp), log_filename)
    return fp
  else: return None
  
def save_full_model(model_filename, params_path, test_fold, val_fold, log_filename, val_inputs, val_intervals, val_targets, min_val_loss):
  with tf.Session() as sess:
        sess.run(init)
        write_to_log('Saving full model', log_filename)
        saver.restore(sess, params_path)
        write_to_log('params loaded from {}'.format(params_path), log_filename)
        val_loss = MSE(sess.run(preds, feed_dict ={x:val_inputs, t:val_intervals})[:,-1,:], val_targets[:,-1,:])
        assert val_loss == min_val_loss, "val loss {} should be equal to min val loss {}".format(val_loss, min_val_loss)
        full_model_dir = model_filename + '/full_model_t{}v{}'.format(test_fold, val_fold) if test_fold else model_filename + '/full_model_v{}'.format(val_fold)
        tf.saved_model.simple_save(sess,
                                  full_model_dir,
                                  inputs={"x":x, "t":t, "y":y},
                                  outputs={"preds":preds, "loss_op":loss_op})
        write_to_log("Full model saved to {}".format(full_model_dir),log_filename)

def get_inputs_targets_intervals(data, interval_list, input_steps, pred_steps, output_steps=1, pred_dim=0):
  size = len(data) - input_steps - output_steps + 1
  dim = data.shape[1]
  inputs = np.zeros((size, input_steps, dim))
  targets = np.zeros((size, pred_steps, 1))
  intervals = np.zeros((size, pred_steps, output_steps))
  
  too_long_intervals = []
  max_interval_length = 2218.0
  min_interval_length = 554.5
  
  for i in range(size):
    inputs[i] = data[i:i+input_steps,:]
    targets[i] = data[i+input_steps-pred_steps+1:i+input_steps+output_steps,pred_dim:pred_dim+1]
    intervals[i] = np.expand_dims(interval_list[i+input_steps-pred_steps+1:i+input_steps+output_steps],1)
    if intervals[i][-1] > max_interval_length or intervals[i][-1] < min_interval_length:
      too_long_intervals.append(i)
  #print('{} samples removed with too long interval until prediction'.format(len(too_long_intervals)))
  inputs = np.delete(inputs, too_long_intervals, axis=0)
  targets = np.delete(targets, too_long_intervals, axis=0)
  intervals = np.delete(intervals, too_long_intervals, axis=0)
  return inputs, targets, intervals        

# COMMAND ----------

# Model a : current interpolated to fit temperature sampling rate.
def get_model_a_data():
  df_curr = df_current.copy()

  #df_curr = df_curr[(df_curr.t >= df_tmp.timestamp.iloc[0]) & (df_curr.t <= df_tmp.timestamp.iloc[-1])]
  df_curr = df_curr.set_index('t')

  df_model_a =  df_tmp[[segment, 'timestamp']].set_index('timestamp').join(df_curr['v'], how='outer')
  df_model_a = df_model_a.join(df_seatemp[['v', 't']].set_index('t'), how='outer', rsuffix='_temp')
  df_model_a['v'] = df_model_a['v'].interpolate()
  df_model_a['v_temp'] = df_model_a['v_temp'].interpolate()
  df_model_a = df_model_a.dropna()

  df_model_a = df_model_a.reset_index()
  df_model_a.columns = ["timestamp", df_model_a.columns[1], 'current', 'sea_temp']
  df_model_a['sampling_int'] = df_model_a.timestamp - df_model_a.timestamp.shift(1)
  df_model_a.loc[0, 'sampling_int'] = 0

  return df_model_a

# COMMAND ----------

def get_blocked_fold_ranges(n_blocks=60, n_folds=5):
  assert n_blocks % n_folds == 0, 'n_folds should divide n_blocks'
  block_size = len(df_tmp)//n_blocks
  block_ranges = []

  for k in range(n_blocks):
    block_range = range(k*block_size,(k+1)*block_size) if k < n_blocks-1 else range(k*block_size, len(df_tmp))
    block_ranges.append(block_range)
  random.seed(730)
  random.shuffle(block_ranges)
  block_ranges

  fold_ranges = [block_ranges[i*(n_blocks//n_folds):(i+1)*n_blocks//n_folds] for i in range(n_folds)]
  return fold_ranges

# COMMAND ----------

def get_model_f_sample_merged(df_tmp, df_current, df_seatemp):
  st = df_tmp.timestamp.iloc[0]
  et = df_tmp.timestamp.iloc[-1]
  
  df_current = df_current[(df_current.t > st) & (df_current.t < et)].copy()
  df_seatemp = df_seatemp[(df_seatemp.t > st) & (df_seatemp.t < et)].copy()
  
  df_current.t = df_current.t - st
  df_seatemp.t = df_seatemp.t - st
  df_tmp.timestamp = df_tmp.timestamp - st
  
  df_joined = df_tmp.set_index('timestamp')[[segment]].join(df_current.set_index('t')[['v']], how='outer')
  df_joined = df_joined.join(df_seatemp.set_index('t')[['v']], lsuffix='_c', rsuffix='_s', how='outer')
  df_joined = df_joined.fillna(method='bfill')
  df_joined = df_joined.fillna(method='ffill')
  
  df_joined = df_joined.reset_index()
  df_joined.columns = ['t', segment, 'v_c', 'v_s']
  
  return df_joined[['t', segment]], df_joined[['t', 'v_c']], df_joined[['t', 'v_s']]

#tm, cu, se = get_model_f_sample_merged(df_tmp[0:50].copy(), df_current, df_seatemp)
#tm.shape, cu.shape, se.shape

# COMMAND ----------

def get_fold_data():
  n_folds = 5
  
  fold_size = len(df_tmp)//n_folds
  fold_ranges = []

  fold_data = []
  for k in range(n_folds):
    fold_range = range(k*fold_size,(k+1)*fold_size) if k < n_folds-1 else range(k*fold_size, len(data))
    fold_ranges.append(fold_range)
    fold_inputs, fold_targets, fold_intervals = get_inputs_targets_intervals(data[fold_range,:], intervals[fold_range],  50, 1)
    sd = df_tmp[['new_timestamp']].iloc[fold_range].iloc[0].values[0]
    ed = df_tmp[['new_timestamp']].iloc[fold_range].iloc[-1].values[0]
    fold_data.append({'inputs': fold_inputs, 'targets':fold_targets, 'intervals':fold_intervals, 'range':fold_range, 'start_date': sd, 'end_date':ed})
  return fold_data
