# Databricks notebook source
# https://danijar.com/variable-sequence-lengths-in-tensorflow/  : Variable length sequences to RNN

# COMMAND ----------

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import glob
import json
from datetime import datetime
from warnings import warn
import itertools

#tf imports
import tensorflow as tf
from tensorflow.nn import rnn_cell
from tensorflow.saved_model import tag_constants

assert np.__version__ == "1.15.4" and pd.__version__ == "0.23.4", "np: {}, pd: {}".format(np.__version__, pd.__version__)

# COMMAND ----------

# MAGIC %run ./func

# COMMAND ----------

log_dir, model_dir = set_dirs('new_CV')

# COMMAND ----------

# Set up tf graph
tf.reset_default_graph()

learning_rate = 0.01
n_hidden = 64
dim_input = 2
n_input_steps = 50
n_output_steps = 1
pred_steps = 1

max_len = 200

assert n_input_steps >= pred_steps, 'input steps must be > pred steps'

# Graph input
x_t = tf.placeholder("float", [None, max_len, dim_input], name='x')
x_c = tf.placeholder("float", [None, max_len, dim_input], name='x')
x_s = tf.placeholder("float", [None, max_len, dim_input], name='x')

l = tf.placeholder('int32', [None])

t = tf.placeholder("float", [None, pred_steps, n_output_steps], name='t')
y = tf.placeholder("float", [None, pred_steps, n_output_steps], name='y')

# Define weights
weights = {
    'out': tf.Variable(tf.random_normal([n_hidden+1,n_output_steps])),  # +1 because of concatenated time until prediction input
    #'concat': tf.Variable(tf.random_normal([n_hidden * 3,n_hidden]))  # +1 because of concatenated time until prediction input
}
biases = {
    'out': tf.Variable(tf.zeros([n_output_steps])),
    #'concat': tf.Variable(tf.zeros([n_output_steps]))
}

def last_relevant(output, length):
  batch_size = tf.shape(output)[0]
  max_length = tf.shape(output)[1]
  out_size = int(output.get_shape()[2])
  index = tf.range(0, batch_size) * max_length + (length - 1)
  flat = tf.reshape(output, [-1, out_size])
  relevant = tf.gather(flat, index)
  return relevant

with tf.variable_scope('lstm_t'):
  cell_t = tf.contrib.rnn.PhasedLSTMCell(n_hidden)
  outputs_t, _ = tf.nn.dynamic_rnn(cell_t, (x_t[:,:,0:1], x_t[:,:,1:]), dtype=tf.float32, sequence_length=l)

with tf.variable_scope('lstm_c'):  
  cell_c = tf.contrib.rnn.PhasedLSTMCell(n_hidden)
  outputs_c, _ = tf.nn.dynamic_rnn(cell_c, (x_c[:,:,0:1],x_c[:,:,1:]), dtype=tf.float32, sequence_length=l)

with tf.variable_scope('lstm_s'):
  cell_s = tf.contrib.rnn.PhasedLSTMCell(n_hidden)
  outputs_s, _ = tf.nn.dynamic_rnn(cell_s, (x_s[:,:,0:1], x_s[:,:,1:]), dtype=tf.float32, sequence_length=l)

#concat outputs t, c and s
merged_outputs = tf.concat([outputs_t, outputs_c, outputs_s], axis=2)

with tf.variable_scope('lstm_merge'):
  cell = tf.contrib.rnn.PhasedLSTMCell(n_hidden)
  
  #input concatenated outputs
  outputs, _ = tf.nn.dynamic_rnn(cell, (x_s[:,:,0:1], merged_outputs), dtype=tf.float32, sequence_length=l)
  output = tf.reshape(last_relevant(outputs, l), (-1,1,n_hidden))


preds = tf.matmul(tf.reshape(tf.concat([output, t], axis=2), (-1, (n_hidden+1))), weights['out']) + biases['out']
preds = tf.reshape(preds, (-1, pred_steps, n_output_steps), name='preds')

# Define loss and optimizer
loss_op = tf.losses.mean_squared_error(labels=y, predictions=preds)
optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
train_op = optimizer.minimize(loss_op)

init = tf.global_variables_initializer()
saver = tf.train.Saver()

# COMMAND ----------

def get_time_range_df_subset(df, time_range):
  return df.loc[(df.new_timestamp >= time_range[0]) & (df.new_timestamp <= time_range[1]), :]

# COMMAND ----------

pd.options.mode.chained_assignment = "raise"
class Model_F:
  def __init__(self, test_fold, val_fold, time_ranges, input_steps=50, input_dim=2):
    assert test_fold != val_fold, 'test fold should not equal val fold'
    self.test_fold = test_fold
    self.val_fold = val_fold
    self.time_ranges = time_ranges
    
    train_temps = np.concatenate([df_tmp[(df_tmp['new_timestamp'] >= time_ranges[fold][block][0]) & (df_tmp['new_timestamp'] <= time_ranges[fold][block][1])][segment].values for fold in range(n_folds) for block in range(n_blocks) if fold !=val_fold and fold != test_fold])
    train_currs = np.concatenate([df_current[(df_current['new_timestamp'] >= time_ranges[fold][block][0]) & (df_current['new_timestamp'] <= time_ranges[fold][block][1])]['v'].values for fold in range(n_folds) for block in range(n_blocks) if fold !=val_fold and fold != test_fold])
    train_sea = np.concatenate([df_seatemp[(df_seatemp['new_timestamp'] >= time_ranges[fold][block][0]) & (df_seatemp['new_timestamp'] <= time_ranges[fold][block][1])]['v'].values for fold in range(n_folds) for block in range(n_blocks) if fold !=val_fold and fold != test_fold])
    train_intervals = np.concatenate([df_tmp[(df_tmp['new_timestamp'] >= time_ranges[fold][block][0]) & (df_tmp['new_timestamp'] <= time_ranges[fold][block][1])]['sampling_interval'].values for fold in range(n_folds) for block in range(n_blocks) if fold !=val_fold and fold != test_fold])
    
    self.temp_mean = np.mean(train_temps)
    self.temp_std = np.std(train_temps)
    self.curr_mean = np.mean(train_currs)
    self.curr_std = np.std(train_currs)
    self.sea_mean = np.mean(train_sea)
    self.sea_std = np.std(train_sea)
    self.int_mean = np.mean(train_intervals)
    self.int_std = np.std(train_intervals)
  
    self.input_steps = input_steps
    self.input_dim = input_dim
    
    self.max_len = max_len

  def get_inputs_multiple_folds(self, folds):
    temp_inputs = [None for _ in range(len(folds))]
    curr_inputs = [None for _ in range(len(folds))]
    sea_inputs = [None for _ in range(len(folds))]
    targets = [None for _ in range(len(folds))]
    intervals = [None for _ in range(len(folds))]
    lengths = [None for _ in range(len(folds))]
    for index, fold in enumerate(folds):
      assert fold != self.test_fold and fold != self.val_fold, 'should not be test or val fold'
      temp_inputs[index], curr_inputs[index], sea_inputs[index], lengths[index], targets[index], intervals[index] = self.__get_fold_inputs(fold)
    temp_inputs, curr_inputs, sea_inputs, lengths, targets, intervals = np.concatenate(temp_inputs), np.concatenate(curr_inputs), np.concatenate(sea_inputs), np.concatenate(lengths), np.concatenate(targets), np.concatenate(intervals)
    print(temp_inputs.shape, curr_inputs.shape, sea_inputs.shape, lengths.shape,  targets.shape, intervals.shape)
    return temp_inputs, curr_inputs, sea_inputs, lengths, targets, intervals
       
  def get_inputs(self, fold):
    assert fold == self.test_fold or fold == self.val_fold, 'should be test or val fold'
    return self.__get_fold_inputs(fold)
    
  def __get_fold_inputs(self, fold):
    all_temp_inputs, all_curr_inputs, all_sea_inputs, all_lengths, all_targets, all_intervals = [],[],[],[],[],[]
    for time_range in self.time_ranges[fold]:
      
      temp_df = get_time_range_df_subset(df_tmp, time_range).copy()
      current_df = get_time_range_df_subset(df_current, time_range).copy()
      seatemp_df = get_time_range_df_subset(df_seatemp, time_range).copy()
      
      current_df.v = (current_df.v - self.curr_mean) / self.curr_std
      temp_df[segment] = (temp_df[segment] - self.temp_mean) / self.temp_std
      seatemp_df.v = (seatemp_df.v - self.sea_mean) / self.sea_std
    
      size = len(temp_df) - self.input_steps - 1
    
      temp_inputs = np.zeros((size, self.max_len, self.input_dim))
      curr_inputs = np.zeros((size, self.max_len, self.input_dim))
      sea_inputs = np.zeros((size, self.max_len, self.input_dim))
      lengths = np.zeros((size), dtype='int')

      targets = np.zeros((size, 1, 1))
      intervals = np.zeros((size, 1, 1))

      too_long_intervals = []
      max_interval_length = 2218.0
      min_interval_length = 554.5

      for i in range(size):  
        input_temps = temp_df.iloc[i:i+self.input_steps, :].copy()
        tm, cu, se = get_model_f_sample_merged(input_temps, current_df, seatemp_df)
        lengths[i] = len(tm)

        temp_inputs[i,: tm.shape[0],:] = tm
        curr_inputs[i, :cu.shape[0],:] = cu.values
        sea_inputs[i,  :se.shape[0],:] = se.values

        targets[i] = temp_df.iloc[i+self.input_steps, :][segment]
        intervals[i] = temp_df.iloc[i+self.input_steps, :]['timestamp'] -  temp_df.iloc[i+self.input_steps-1, :]['timestamp']
        if temp_df.iloc[i+self.input_steps, :]['timestamp'] - temp_df.iloc[i+self.input_steps-1, :]['timestamp'] > max_interval_length or temp_df.iloc[i+self.input_steps, :]['timestamp'] - temp_df.iloc[i+self.input_steps-1, :]['timestamp'] < min_interval_length:
          too_long_intervals.append(i)
      #print('{} samples removed with too long interval until prediction'.format(len(too_long_intervals)))
      all_temp_inputs.append(np.delete(temp_inputs, too_long_intervals, axis=0))
      all_curr_inputs.append(np.delete(curr_inputs, too_long_intervals, axis=0))
      all_sea_inputs.append(np.delete(sea_inputs, too_long_intervals, axis=0))
      all_lengths.append(np.delete(lengths, too_long_intervals, axis=0))
      all_targets.append(np.delete(targets, too_long_intervals, axis=0))
      intervals = np.delete(intervals, too_long_intervals, axis=0)
      all_intervals.append((intervals - self.int_mean) / self.int_mean)
    return np.concatenate(all_temp_inputs), np.concatenate(all_curr_inputs), np.concatenate(all_sea_inputs), np.concatenate(all_lengths), np.concatenate(all_targets), np.concatenate(all_intervals)
    
  def undo_normalize(self, temp_preds):
    return (temp_preds * self.temp_std) + self.temp_mean
  
#model_f = Model_F(0, 1, time_ranges)
#temp_inputs, curr_inputs, sea_inputs, lengths, targets, intervals = model_f.get_inputs(0)
#temp_inputs, curr_inputs, sea_inputs, lengths, tars, ints = model_f.get_inputs_multiple_folds([0, 1, 2, 3, 4])
#print('max len:{}'.format(np.max(lengths)))   #max sea:21 max curr:126

# COMMAND ----------

temp_ranges = get_blocked_fold_ranges()
time_ranges = [[(df_tmp.iloc[temp_ranges[fold][block].start]['new_timestamp'], df_tmp.iloc[temp_ranges[fold][block].stop-1]['new_timestamp']) for block in range(n_blocks)] for fold in range(n_folds)]

# COMMAND ----------

is_testing = True
n_epochs = 1000
batch_size = 128
initial_patience = 50

MSEs_train = []
MSEs_val = []
MSEs_test = []
MSEs_train_norm = []
MSEs_val_norm = []
MSEs_test_all = []

file_suffix = 'model_F'
log_filename = log_dir + file_suffix + '.txt' if file_suffix else None
model_filename = model_dir + file_suffix if file_suffix else None
if log_filename: assert not os.path.isfile(log_filename), "path " + log_filename + " is already in use"
if not file_suffix:
  warn("Training model without saving or logging")
else:
  write_to_log('writing log to {} and models to {}'.format(log_filename, model_filename), log_filename)

args = {'lr': learning_rate,
        'n hidden': str(n_hidden),
        'dim input':dim_input,
        'input steps':n_input_steps,
        'output steps':n_output_steps,
        'pred steps':pred_steps,
        'n folds': n_folds,
        'n epochs': n_epochs,
        'batch size': batch_size,
        'patience': initial_patience,
       'fold ranges': temp_ranges}
write_to_log(str(args), log_filename)
  
for test_fold in range(n_folds):
  if not is_testing:
    test_fold = -1
  for val_fold in range(n_folds):
    if test_fold != val_fold or not is_testing:
      write_to_log('test fold {} val fold {}'.format(test_fold, val_fold), log_filename)
      
      model_f = Model_F(test_fold, val_fold, time_ranges)
      train_temps, train_currs, train_seas, train_lengths, train_targets, train_intervals = model_f.get_inputs_multiple_folds([i for i in range(n_folds) if i !=val_fold and i != test_fold])
      val_temps, val_currs, val_seas, val_lengths, val_targets, val_intervals = model_f.get_inputs(val_fold)
      print(val_temps.shape, val_currs.shape, val_seas.shape, val_lengths.shape, val_targets.shape, val_intervals.shape)
      
      if is_testing:
        test_temps, test_currs, test_seas,  test_lengths, test_targets, test_intervals = model_f.get_inputs(test_fold)
        pass
      epoch_indicies = [i for i in range(len(train_temps))]
      min_val_loss, min_train_loss = np.inf, np.inf
      patience = initial_patience
      
      with tf.Session() as sess:
        sess.run(init)
        for epoch in range(n_epochs):
          np.random.shuffle(epoch_indicies)
          epoch_temps, epoch_currs, epoch_seas = train_temps[epoch_indicies], train_currs[epoch_indicies], train_seas[epoch_indicies]
          epoch_targets, epoch_t, epoch_l =  train_targets[epoch_indicies], train_intervals[epoch_indicies], train_lengths[epoch_indicies]
          for batch in range(0, len(train_temps), batch_size):
            batch_xt, batch_xc, batch_xs = epoch_temps[batch:batch+batch_size], epoch_currs[batch:batch+batch_size], epoch_seas[batch:batch+batch_size]
            batch_y, batch_t, batch_l  =  epoch_targets[batch:batch+batch_size], epoch_t[batch:batch+batch_size], epoch_l[batch:batch+batch_size],
            sess.run(train_op, feed_dict={x_t:batch_xt, x_c:batch_xc, x_s:batch_xs, t:batch_t, y:batch_y, l:batch_l})
          epoch_val_loss = MSE(sess.run(preds, feed_dict ={x_t:val_temps, x_c:val_currs,x_s:val_seas,
                                                           t:val_intervals,
                                                           l:val_lengths})[:,-1,:], val_targets[:,-1,:], normalizer=False, model=model_f)
          epoch_train_loss = sess.run(loss_op, feed_dict ={x_t:train_temps,x_c:train_currs,x_s:train_seas, 
                                                           t:train_intervals,
                                                           y:train_targets[:,:,:],
                                                           l:train_lengths})
          if epoch_val_loss < min_val_loss and patience:
            min_val_loss = epoch_val_loss
            val_preds_epoch = sess.run(preds, feed_dict={x_t:val_temps,x_c:val_currs,x_s:val_seas,
                                                         t:val_intervals,
                                                         l:val_lengths})
            train_preds_epoch = sess.run(preds, feed_dict={x_t:train_temps,x_c:train_currs,x_s:train_seas,
                                                           t:train_intervals,
                                                           l:train_lengths,})
            train_MSE_epoch = MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:], normalizer=False, model=model_f)
            test_MSE = MSE(sess.run(preds, feed_dict={x_t:test_temps,x_c:test_currs,x_s:test_seas,
                                                      t:test_intervals,
                                                      l:test_lengths})[:,-1,:], test_targets[:,-1,:], normalizer=False, model=model_f) if is_testing else -1
            min_train_loss = epoch_train_loss
            write_to_log("\tEpoch {} finished with training loss {}. For 1-step preds training loss is {} and validation loss {} test {}".format(epoch, min_train_loss, train_MSE_epoch, min_val_loss, test_MSE), log_filename)
            params_path = None #save_params(sess, epoch, test_fold, val_fold, model_filename, False)
            patience = initial_patience
          elif patience:
            patience -= 1
          else:
            write_to_log("Stopped training at epoch {} due to patience running out".format(epoch), log_filename)
            break
      MSE_train = MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:], normalizer=False, model=model_f)
      MSE_val = MSE(val_preds_epoch[:,-1,:], val_targets[:,-1,:], normalizer=False, model=model_f)
      MSEs_train.append(MSE_train)
      MSEs_val.append(MSE_val)
      MSEs_test.append(test_MSE)
      MSEs_train_norm.append(MSE(train_preds_epoch[:,-1,:], train_targets[:,-1,:]))
      MSEs_val_norm.append(MSE(val_preds_epoch[:,-1,:], val_targets[:,-1,:]))
      write_to_log("Accuracy for one step preds train:{} val:{}\n".format(MSE_train, MSE_val), log_filename)
      if params_path:
        save_full_model(model_filename, params_path, test_fold, val_fold, val_lengths, log_filename, val_inputs, val_intervals, val_targets, MSEs_val_norm[-1])
  if not is_testing:
    break
  else:
    MSEs_test_all.append(MSEs_test)
    MSEs_test = []
write_to_log('normalized average: train:{}, val:{}'.format(np.mean(MSEs_train_norm), np.mean(MSEs_val_norm)), log_filename)
write_to_log('average MSEs across all folds train:{} val:{} test:{}'.format(np.mean(MSEs_train), np.mean(MSEs_val), np.mean(MSEs_test)), log_filename)
write_to_log('MSE std across all folds train:{} val:{} test:{}'.format(np.std(MSEs_train), np.std(MSEs_val), np.std(MSEs_test)), log_filename)
if not is_testing:
  write_to_log('MSE val for each fold: {} & {} & {} & {} & {}'.format(MSEs_val[0],MSEs_val[1],MSEs_val[2],MSEs_val[3],MSEs_val[4]), log_filename)
else:
  write_to_log('average MSE test for each fold: {} & {} & {} & {} & {}'.format(np.mean(MSEs_test_all[0]),
                                                                               np.mean(MSEs_test_all[1]),
                                                                               np.mean(MSEs_test_all[2]),
                                                                               np.mean(MSEs_test_all[3]),
                                                                               np.mean(MSEs_test_all[4])), log_filename)
write_to_log('\n-----------------------------------------------------------------\n', log_filename)

# COMMAND ----------


