# Databricks notebook source
from scipy.stats import norm

# COMMAND ----------

# MAGIC %run ./func_p2

# COMMAND ----------

SUBSET = False
df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
fold_ranges = get_fold_ranges(SUBSET)
n_columns = 9

# COMMAND ----------

for fr in fold_ranges:
  print(len(fr))

# COMMAND ----------

len(df_dts)

# COMMAND ----------

len(fold_ranges), len(fold_ranges[0])

# COMMAND ----------

def get_increase_multipliers(n):
  assert n%2, 'n must be odd but was {}'.format(n)
  quantiles = np.concatenate([np.linspace(2, 0, (n//2)+1)[:-1], np.linspace(0, 2, (n//2)+1)])
  return np.array([2*(1-norm.cdf(q)) for q in quantiles])

def get_increases(n, pct_inc):
  return pct_inc * get_increase_multipliers(n)

get_increases(9, 0.1), get_increase_multipliers(9)

# COMMAND ----------

 10*np.asarray([0.04550026, 0.1336144 , 0.31731051, 0.61707508, 1., 0.61707508, 0.31731051, 0.1336144 , 0.04550026])

# COMMAND ----------

plt.close()
fig, ax= plt.subplots(1,1, figsize=(8,4))
ax.bar([i for i in range(-5,6)], 0.1*np.asarray([0,0.04550026, 0.1336144 , 0.31731051, 0.61707508, 1., 0.61707508, 0.31731051, 0.1336144 , 0.04550026, 0]), align='center')
ax.set_title('Spatial distribution of temperature increases')
ax.set_ylabel('Temperature increase')
#ax.set_xlabel('temperature increase')
ax.set_xticks([i for i in range(-5,6)])
display(plt.show())

# COMMAND ----------

def _make_instant_faults(df, time_range, n_columns, center_column, data_path=None, metadata_path=None, pct_increase=0.1):
  assert n_columns % 2 == 1
  if not center_column: 
    center_column = random.randint(0, n_segments - n_columns//2)
  columns = [c for c in range(center_column - n_columns//2, center_column + n_columns//2+1) if c >= 0]
  segments = [list(df)[column] for column in columns]
  assert len(df) > 100, "Length of dataframe is assumed to be 100+"
  begin_step = 100
  increases = get_increases(n_columns, pct_increase)
  gen_df = df.loc[:, segments + ['timestamp']].copy().reset_index(drop=True)
  for col in range(n_columns):
    gen_df.iloc[begin_step:, col] = gen_df.iloc[begin_step:, col] * (1+increases[col])
  description = {"segments":segments, "columns":columns, "length_df":len(df), "begin_step":begin_step,
                 "pct_increase":pct_increase, 'start':str(time_range['start_timestamp']), 'end':str(time_range['end_timestamp']), "type":"instant"}
  if data_path:
    gen_df.to_csv(data_path, index=False)
    with open(metadata_path, 'w') as file:
      json.dump(description, file)
  else:
    return gen_df, description

def make_instant_local_faults(df, time_range, range_count, center_column, fold_dir=None, pct_increase=0.1):
  if fold_dir: 
    if not os.path.exists(fold_dir):
      os.makedirs(fold_dir)
    data_path = fold_dir + '/data_{:03d}.csv'.format(range_count)
    metadata_path = fold_dir +'/metadata_{:03d}.json'.format(range_count)
    _make_instant_faults(df, time_range, 1, center_column, data_path, metadata_path, pct_increase=0.1)
  else:
    return _make_instant_faults(df, time_range, 1, center_column, pct_increase=0.1)
  
def make_instant_nonlocal_faults(df, time_range, range_count, center_column, fold_dir=None, pct_increase=0.1):
  neighbouring = n_columns
  if fold_dir: 
    if not os.path.exists(fold_dir):
      os.makedirs(fold_dir)
    data_path = fold_dir + '/data_{:03d}.csv'.format(range_count)
    metadata_path = fold_dir +'/metadata_{:03d}.json'.format(range_count)
    _make_instant_faults(df, time_range, neighbouring, center_column, data_path, metadata_path, pct_increase=0.1)
  else:
    return _make_instant_faults(df, time_range, neighbouring, center_column, pct_increase=0.1)

# COMMAND ----------

block_df = df_by_time_range(df_dts, fold_ranges[0][1])
gen_df, description = make_instant_nonlocal_faults(block_df, fold_ranges[0][1], 0, None)

# COMMAND ----------

gen_df.count()

# COMMAND ----------

for el in list(gen_df):
  print(list(block_df).index(el))
list(block_df).index(list(gen_df)[0])-1, list(block_df).index(list(gen_df)[-2])+1

# COMMAND ----------

block_df.iloc[:,less-5:less].count()

# COMMAND ----------

block_df.iloc[:,more:more+5].count()

# COMMAND ----------

gen_df.count()

# COMMAND ----------

# PLOT TYPE I FAULT

block_df = df_by_time_range(df_dts, fold_ranges[4][1])
gen_df, description = make_instant_local_faults(block_df, fold_ranges[4][1], 0, 6290)
plt.close()
fig, (ax1, ax2) = plt.subplots(2,1, sharey=True, sharex=True, figsize=(9,4.4))



less = list(block_df).index(list(gen_df)[0])
more =  list(block_df).index(list(gen_df)[-2])+1

ax2.plot(block_df.iloc[:,less-5:less].values)
ax2.plot(gen_df.iloc[:,:-1])
ax2.plot(block_df.iloc[:,more:more+5].values)

ax1.plot(block_df.iloc[:,less-5:less].values)
ax1.plot(block_df.loc[:,list(gen_df.iloc[:,:-1])].values)
ax1.plot(block_df.iloc[:,more:more+5].values)

#ax1.plot(block_df.iloc[:,[list(block_df).index(list(gen_df)[0])-1, list(block_df).index(list(gen_df)[-2])+1]].values)



ax2.set_xlabel('Time step')
ax2.set_ylabel('DTS value')
ax1.set_ylabel('DTS value')
ax2.set_title('Constructed type I fault')
ax1.set_title('Block with no fault')

ax2.set_xlim(right=len(block_df))
#display(plt.show())

# COMMAND ----------

list(gen_df)

# COMMAND ----------

all_segments = list(block_df.iloc[:,:-3])
len(all_segments)

# COMMAND ----------

block_df.iloc[-1,:list(block_df).index(list(gen_df)[0])].shape, block_df.iloc[-1,list(block_df).index(list(gen_df)[-2])+1:-3].shape

# COMMAND ----------

[i for i in range(len(block_df.iloc[-1,:list(block_df).index(list(gen_df)[0])]), block_df.iloc[-1,:list(block_df).index(list(gen_df)[0])]+10)], gen_df.iloc[-1,:-1]

# COMMAND ----------

gen_df

# COMMAND ----------

list(block_df.iloc[-1:,6000:6600])

# COMMAND ----------

[segs[0], segs[-1]]

# COMMAND ----------

len(segs)

# COMMAND ----------

fig, (ax1) = plt.subplots(1, 1, sharey=True, sharex=True, figsize=(12,4.8))



ax1.plot(block_df.iloc[-1,6000:6600], label='normal data')
#ax2.plot(block_df.iloc[-1,6000:6600])
#ax2.plot(block_df.iloc[-1,:-3])
ax1.plot(gen_df.iloc[-1,:-1], '-',color='red',  label='constructed fault')


ax1.set_title('Part of export cable temperature profile for instant fault')
ax1.legend()
ax1.set_xlabel('Segment')
segs = list(block_df.iloc[-1:,6000:6600])
ax1.set_ylabel('DTS value')

ax1.set_xlim(right=6600, left=6000)
ax1.set_xticks([6005, 6284, 6595])
ax1.set_xticklabels([str(float(segs[5])), str(float(segs[290])), str(float(segs[-6]))])



display(plt.show())

# COMMAND ----------

fig, (ax1,ax2) = plt.subplots(2, 1, sharey=True, sharex=True, figsize=(9,4.8))

ax1.plot(block_df.iloc[-1,:-3])
ax2.plot(block_df.iloc[-1,:-3])
ax2.plot(gen_df.iloc[-1,:-1], color='orangered')
ax2.set_xlim(right=len(block_df.iloc[-1,:-3]))

ax1.set_title('Normal temperature profile')
ax2.set_title('Temperature profile with fault')
ax2.set_xlabel('Segment')
ax1.set_ylabel('DTS value')
ax2.set_ylabel('DTS value')

tick_list = [0, 10000, 20000, 30000, 39500]
ax2.set_xticks(tick_list)
ax2.set_xticklabels([str(float(list(block_df)[i])) for i in tick_list])

#first_part = block_df.iloc[-1,:list(block_df).index(list(gen_df)[0])]
#ax2.plot(first_part)

#ax2.plot(block_df.iloc[-1,list(block_df).index(list(gen_df)[-2])+1:-3])

display(plt.show())

# COMMAND ----------

## PLOT TYPE II FAULT

block_df = df_by_time_range(df_dts, fold_ranges[4][1])
gen_df, description = make_instant_nonlocal_faults(block_df, fold_ranges[4][1], 0, 6290)
plt.close()
fig, (ax1,ax2) = plt.subplots(2, 1, sharey=True, sharex=True, figsize=(9,4.8))

ax1.plot(block_df.iloc[:,list(block_df).index(list(gen_df)[0])-1].values, color='b')
ax1.plot(block_df.loc[:,list(gen_df.iloc[:,:-1])].values)
ax1.plot(block_df.iloc[:,list(block_df).index(list(gen_df)[-2])+1].values)

ax2.plot(block_df.iloc[:,list(block_df).index(list(gen_df)[0])-1].values, color='b')
ax2.plot(gen_df.iloc[:,:-1])
ax2.plot(block_df.iloc[:,list(block_df).index(list(gen_df)[-2])+1].values)

#ax1.plot(block_df.iloc[:,list(block_df).index(list(gen_df)[0])-1].values)
#ax1.plot(block_df.loc[:,list(gen_df.iloc[:,:-1])].values)
#ax1.plot(block_df.iloc[:,list(block_df).index(list(gen_df)[-2])+1].values)
#ax1.plot(block_df.iloc[:,[list(block_df).index(list(gen_df)[0])-1, list(block_df).index(list(gen_df)[-2])+1]].values)



ax2.set_xlabel('Time step')
ax2.set_ylabel('DTS value')
ax1.set_ylabel('DTS value')
ax2.set_title('Instant fault')
ax1.set_title('Normal data')

ax2.set_xlim(right=len(block_df))
display(plt.show())

# COMMAND ----------

def make_drifting_local_faults(df_tmp, time_ranges, fold_dir, center_column=None):
  _make_drifting_faults(df_tmp, time_ranges, n_columns=1, fold_dir=fold_dir, center_column=center_column)
  
def make_drifting_nonlocal_faults(df_tmp, time_ranges, fold_dir, center_column=None):
  _make_drifting_faults(df_tmp, time_ranges, n_columns=n_columns, fold_dir=fold_dir, center_column=center_column)

# COMMAND ----------

def _make_drifting_fault(df, time_range, increase, n_columns, center_column, data_path=None, metadata_path=None):
  assert n_columns % 2 == 1
  if not center_column:
    center_column = random.randint(0, n_segments - n_columns//2)
  columns = [c for c in range(center_column - n_columns//2, center_column + n_columns//2+1) if c >= 0]
  segments = [list(df)[column] for column in columns]
  gen_df = df.loc[:, segments + ['timestamp', 'new_timestamp']].copy().set_index('new_timestamp')
  
  increases = [{'to':1 + (increase['to']-1)*q, 'from':1 + (increase['from']-1)*q} for q in get_increase_multipliers(n_columns)]
  increases_columns = ["pct_increase_{}".format(i) for i in range(len(increases))]

  increase_dict = {increases_columns[i]: [increase['from'], increase['to']] for i, increase in enumerate(increases)}
  increase_dict["new_timestamp"] = [pd.to_datetime(time_range['start_timestamp']), pd.to_datetime(time_range['end_timestamp'])]
  
  increase_frame = pd.DataFrame.from_dict(increase_dict).set_index('new_timestamp')
  
  joined_df = gen_df.join(increase_frame, how='outer')
  joined_df[increases_columns] = joined_df[increases_columns].interpolate()
  joined_df = joined_df.dropna().reset_index(drop=True)
  
  increase_list = list(joined_df.loc[: ,"pct_increase_{}".format(n_columns//2)].values)
  
  joined_df[segments] =  joined_df[segments].multiply(joined_df[increases_columns].values, axis=0)
  joined_df = joined_df.drop(increases_columns, axis=1)
  
  description = {"segments":segments, "columns":columns, "length_df":len(df), "increase":increase, 'increase_list': increase_list,
                 'start':str(time_range['start_timestamp']), 'end':str(time_range['end_timestamp']), "type":"drifting"}
  if data_path:
    joined_df.to_csv(data_path, index=False)
    with open(metadata_path, 'w') as file:
      json.dump(description, file)
  else:
    return joined_df, description
  
def _make_drifting_faults(df, time_ranges, n_columns=1, days=90, repeats = 1, pct_increase=0.1, fold_dir=None, center_column=None):
  ranges_per_repeat = pd.Timedelta('{} days'.format(days)) // (pd.to_datetime(time_ranges[0]['end_timestamp']) - pd.to_datetime(time_ranges[0]['start_timestamp']))
  
  for rep in range(repeats):
    shuffled_ranges = []
    while len(shuffled_ranges) < ranges_per_repeat:
      temp_range = time_ranges.copy()
      random.shuffle(temp_range)
      shuffled_ranges.extend(temp_range)
      
    pct_list = np.linspace(1, 1+pct_increase, ranges_per_repeat+1)
    increase_ranges = [{"from":pct_list[i-1], "to":pct_list[i], "i":i} for i in range(1, len(pct_list))]
    for i, increase_dict in enumerate(increase_ranges):
      tr = shuffled_ranges[i]
      block_df = df_by_time_range(df, tr)
      if fold_dir:
        if not os.path.exists(fold_dir):
            os.makedirs(fold_dir)
        data_path = fold_dir + '/data_{:03d}_{:03d}.csv'.format(rep, i)
        metadata_path = fold_dir +'/metadata_{:03d}_{:03d}.json'.format(rep, i)
        _make_drifting_fault(block_df, tr, increase_dict, n_columns, center_column, data_path, metadata_path)
      else:
        _,_ = _make_drifting_fault(block_df, tr, increase_dict, n_columns, center_column)

# COMMAND ----------

block_df = df_by_time_range(df_dts, fold_ranges[0][2])
drift_df, desc = _make_drifting_fault(block_df, fold_ranges[0][2], {"from":1.09777778, "to":1.1, "i":0}, 9, None)
#desc

# COMMAND ----------

plt.close()
fig, (ax1, ax2) = plt.subplots(2,1, sharey=True)
ax1.plot(drift_df.iloc[:,:-1])
ax2.plot(block_df.loc[:,list(drift_df.iloc[:,:-1])].values)
#display(plt.show())

# COMMAND ----------

#plt.close()
#fig, (ax1, ax2) = plt.subplots(2,1, sharey=True)
#ax1.plot(drift_df.iloc[:,:-1])
#ax2.plot(block_df.loc[:,list(drift_df.iloc[:,:-1])].values)
#display(plt.show())



## PLOT TYPE IV FAULT

block_df = df_by_time_range(df_dts, fold_ranges[4][1])
drift_df1, desc = _make_drifting_fault(block_df, fold_ranges[4][1], {"from":1.09777778, "to":1.1, "i":0}, 9, 6290)
drift_df2, desc2 = _make_drifting_fault(block_df, fold_ranges[4][1], {"from":1.0, "to":1.0022222222222, "i":0}, 9, 6290)
drift_df3, desc3 = _make_drifting_fault(block_df, fold_ranges[4][1], {"from":1.048888888889, "to":1.05111111111, "i":0}, 9, 6290)

plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,1, sharey=True, sharex=True, figsize=(9,7.2))

ax1.plot(block_df.iloc[:,list(block_df).index(list(drift_df1)[0])-1].values, color='b')
#ax1.plot(block_df.loc[:,list(drift_df.iloc[:,:-1])].values)
ax1.plot(drift_df2.iloc[:,:-1])
ax1.plot(block_df.iloc[:,list(block_df).index(list(drift_df1)[-2])+1].values)

ax2.plot(block_df.iloc[:,list(block_df).index(list(drift_df1)[0])-1].values, color='b')
ax2.plot(drift_df3.iloc[:,:-1])
ax2.plot(block_df.iloc[:,list(block_df).index(list(drift_df1)[-2])+1].values)

ax3.plot(block_df.iloc[:,list(block_df).index(list(drift_df1)[0])-1].values, color='b')
ax3.plot(drift_df1.iloc[:,:-1])
ax3.plot(block_df.iloc[:,list(block_df).index(list(drift_df1)[-2])+1].values)


ax3.set_xlabel('Time step')
ax2.set_ylabel('DTS value')
ax1.set_ylabel('DTS value')
ax3.set_ylabel('DTS value')
ax1.set_title('Gradual fault first 2 days')
ax2.set_title('Gradual fault days 44 to 46')
ax3.set_title('Gradual fault last 2 days')
ax2.set_xlim(right=len(block_df))

display(plt.show())

# COMMAND ----------

fig, (ax1) = plt.subplots(1, 1, sharey=True, sharex=True, figsize=(12,4.8))

ax1.plot(block_df.iloc[-1,6250:6340], label='normal data')
#ax2.plot(block_df.iloc[-1,2000:6300])
#ax2.plot(block_df.iloc[-1,:-3])
ax1.plot(drift_df2.iloc[-1,:-1], '-',color='green', alpha=0.8, label='after 2 days')
ax1.plot(drift_df3.iloc[-1,:-1], '-',color='purple',  label='after 46 days')
ax1.plot(drift_df1.iloc[-1,:-1], '-',color='red',  label='last day')


ax1.set_title('Part of export cable temperature profile for gradual fault')
ax1.legend()
ax1.set_xlabel('Segment')
#segs = list(block_df.iloc[-1:,6000:6600])
ax1.set_ylabel('DTS value')

#ax1.set_xlim(right=6600, left=6000)
#ax1.set_xticks([6005, 6284, 6595])
#ax1.set_xticklabels([str(float(segs[5])), str(float(segs[290])), str(float(segs[-6]))])



display(plt.show())

# COMMAND ----------

#plt.close()
#fig, (ax1, ax2) = plt.subplots(2,1, sharey=True)
#ax1.plot(drift_df.iloc[:,:-1])
#ax2.plot(block_df.loc[:,list(drift_df.iloc[:,:-1])].values)
#display(plt.show())



## PLOT TYPE III FAULT

block_df = df_by_time_range(df_dts, fold_ranges[4][1])
drift_df, desc = _make_drifting_fault(block_df, fold_ranges[4][1], {"from":1.09777778, "to":1.1, "i":0}, 1, 6290)
drift_df2, desc2 = _make_drifting_fault(block_df, fold_ranges[4][1], {"from":1.0, "to":1.0022222222222, "i":0}, 1, 6290)
drift_df3, desc3 = _make_drifting_fault(block_df, fold_ranges[4][1], {"from":1.048888888889, "to":1.05111111111, "i":0}, 1, 6290)
plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,1, sharey=True, sharex=True, figsize=(9,6.6))


less = list(block_df).index(list(drift_df)[0])
more =  list(block_df).index(list(drift_df)[-2])+1

ax1.plot(block_df.iloc[:,less-5:less].values)
#ax1.plot(block_df.loc[:,list(drift_df.iloc[:,:-1])].values)
ax1.plot(drift_df2.iloc[:,:-1])
ax1.plot(block_df.iloc[:,more:more+5].values)

ax2.plot(block_df.iloc[:,less-5:less].values)
ax2.plot(drift_df3.iloc[:,:-1])
ax2.plot(block_df.iloc[:,more:more+5].values)

ax3.plot(block_df.iloc[:,less-5:less].values)
ax3.plot(drift_df.iloc[:,:-1])
ax3.plot(block_df.iloc[:,more:more+5].values)


ax3.set_xlabel('Time step')
ax2.set_ylabel('DTS value')
ax1.set_ylabel('DTS value')
ax3.set_ylabel('DTS value')
ax1.set_title('Constructed type III fault first 2 days')
ax2.set_title('Constructed type III fault days 44 to 46')
ax3.set_title('Constructed type III fault last 2 days')

ax2.set_xlim(right=len(block_df))
display(plt.show())

# COMMAND ----------

_make_drifting_faults(df_dts, fold_ranges[0])

# COMMAND ----------



# COMMAND ----------

#fault_dir = 'dbfs/mnt/renewablesstudents/user-write/faults_v2'
def make_faults(df_tmp, time_ranges, fold_dir):
  # Write faults to file, identify by fault ID and fault type (I, II, III, IV) by directory, then count within directory. 
  print('writing to', fold_dir)
  
  
  center_columns = [6290]#[17712, 32028, False]
  
  for cc in center_columns:
    column_dir = fold_dir + '/column_{}'.format(cc if cc>0 else 'random')
    for i, tr in enumerate(time_ranges):
      block_df = df_by_time_range(df_tmp, tr)
      make_instant_local_faults(block_df, tr, i, cc, column_dir + '/typeI')
      make_instant_nonlocal_faults(block_df, tr, i, cc, column_dir  + '/typeII')
    make_drifting_local_faults(df_tmp, time_ranges, column_dir+ '/typeIII', center_column=cc)
    make_drifting_nonlocal_faults(df_tmp, time_ranges, column_dir+ '/typeIV', center_column=cc)

# COMMAND ----------

for test_fold, tf_ranges in enumerate(fold_ranges):
  make_faults(df_dts, tf_ranges, '/dbfs/mnt/renewablesstudents/user-write/faults_allDataSUBSET/test_fold{}/'.format(test_fold))

# COMMAND ----------

dbutils.fs.ls('dbfs:/mnt/renewablesstudents/user-write/faults_allDataSUBSET/test_fold3/column_random/typeI')   # faults_allDataSUBSET

# COMMAND ----------

len(list(df_dts))

# COMMAND ----------

random.randint(0,39519)

# COMMAND ----------


