# Databricks notebook source
import tensorflow as tf

# COMMAND ----------

# MAGIC %run ./func_spatial

# COMMAND ----------

# MAGIC %run ../temporal/func_AD_evaluation

# COMMAND ----------

#Logs
#  /dbfs/mnt/renewablesstudents/user-write/logfiles/p2/spatial_AE/allData/CAE_4650_allDataApril.txt
#  /dbfs/mnt/renewablesstudents/user-write/logfiles/p2/spatial_AE/allData/CAE_4650_allDataApril_continued.txt

#Params
#  /dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_4650_allDataApril/params_t{}v{}/model.ckpt


#/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_148/params_t0v1/model.ckpt

# COMMAND ----------

SUBSET = False
fold_ranges = get_fold_ranges(subset=SUBSET)  
df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']
df_current = load_current_data(SUBSET)
df_seatemp = load_seatemp_data(SUBSET)

model_dir ='/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_4650_allDataApril/params_t{}v{}/model.ckpt'
#model_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_148/params_t0v1/model.ckpt'

# COMMAND ----------

#model_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_148/params_t0v1/model.ckpt'
#model_dir ='/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_4650_allDataApril/params_t0v1/model.ckpt'

# COMMAND ----------

def convolutionalAutoencoder(lr, log=False):
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()

  x = tf.placeholder("float", [None, n_segments])
  bs = tf.placeholder("int32")

  inp = tf.reshape(x, shape=[-1, n_segments, 1])  
  log.write('learning rate: {}'.format(lr))
  log.write('input {}'.format(inp))
      
  n_filters = 30
  ksize1 = [5, 1, n_filters]
  ksize2 = [5, n_filters, n_filters]
  ksize3 = [5, n_filters, n_filters]
  '''
  ksize4 = [5, n_filters, 4]
  p1 = 6
  p2 = 6
  p3 = 6
  p4 = 5
  
  '''
  ksize4 = [5, n_filters, 30]
  p1 = 4
  p2 = 4
  p3 = 4
  p4 = 4
  
  log.write('ksize 1: {}, max pool 1: {}'.format(ksize1, p1))
  log.write('ksize 2: {}, max pool 2: {}'.format(ksize2, p2))
  log.write('ksize 3: {}, max pool 3: {}'.format(ksize3, p3))
  log.write('ksize 4: {}, max pool 4: {}'.format(ksize4, p4))

  W = tf.get_variable('W', shape=(n_filters,1), initializer=tf.random_normal_initializer())
  B = tf.get_variable('B', shape=(1), initializer=tf.zeros_initializer())

  c1, s1 = conv(inp, ksize1, 1, 'conv1')
  log.write('conv1: {}'.format(c1))
  mp1 = max_pool(c1, p1, p1, 'mp1', ksize1[2], bs)
  log.write('mp1: {}'.format(mp1))
  c2, s2 = conv(mp1, ksize2, 1, 'conv2')
  log.write('conv2: {}'.format(c2))
  mp2 = max_pool(c2, p2, p2, 'mp2', ksize2[2], bs)
  log.write('mp2: {}'.format(mp2))
  c3, s3 = conv(mp2, ksize3, 1, 'conv3')
  log.write('c3: {}'.format(c3))
  mp3 = max_pool(c3, p3, p3, 'mp3', ksize3[2], bs)
  log.write('mp3: {}'.format(mp3))
  c4, s4 = conv(mp3, ksize4, 1, 'conv4')
  log.write('c4: {}'.format(c4))
  mp4 = max_pool(c4, p4, p4, 'mp4', ksize4[2], bs)
  log.write('mp4: {}'.format(mp4))
  
  encoding = tf.identity(mp4, name='encoding')
  
  dc4 = deconv(mp4, ksize4, s4, 'deconv4', int(np.ceil(n_segments/(p1*p2*p3*p4))), bs)
  log.write('dc4: {}'.format(dc4))
  up4 = upsample(dc4, 'unpool4', int(np.ceil(n_segments/(p1*p2*p3))), ksize4[1], bs)
  log.write('up4: {}'.format(up4))
  dc3 = deconv(up4, ksize3, s3,'deconv3', int(np.ceil(n_segments/(p1*p2*p3))), bs)
  log.write('dc3: {}'.format(dc3))
  up3 = upsample(dc3, 'unpool0', int(np.ceil(n_segments/(p1*p2))), ksize3[1], bs)
  log.write('up3: {}'.format(up3))
  dc2 = deconv(up3, ksize2, s2,'deconv2', int(np.ceil(n_segments/(p1*p2))), bs)
  log.write('dc2: {}'.format(dc2))
  up2 = upsample(dc2, 'unpool1', n_segments//p1, ksize2[1], bs)
  log.write('up2: {}'.format(up2))
  dc1 = deconv(up2, [ksize1[0],  ksize2[1], n_filters], s1,'deconv1', int(np.ceil(n_segments//p1)), bs)
  log.write('dc1: {}'.format(dc1))
  up1 = upsample(dc1, 'unpool1', n_segments, ksize1[2], bs)
  log.write('up1: {}'.format(up1))

  res = tf.matmul(tf.reshape(up1, shape = [-1, n_filters]), W) + B
  log.write('res: {}'.format(res))

  r = tf.reshape(res, shape=[-1, n_segments])

  loss_op = tf.losses.mean_squared_error(labels=x, predictions=r)
  optimizer = tf.train.AdamOptimizer(learning_rate=lr)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, r, x, bs

# COMMAND ----------

train_op, r, x, bs = convolutionalAutoencoder(lr = 0.001)
saver = tf.train.Saver()

# COMMAND ----------

def rec_error_likelihood(recs, mean, var):
  temp = recs - mean
  norm_arr = np.square(temp)/var
  if not len(reconstructed_sequence_global_f):
    global f_score_global
    f_score_global = norm_arr
  else:
    global n_score_global
    n_score_global = norm_arr
  return np.max(norm_arr)/np.mean(norm_arr), np.argmax(norm_arr)
  return np.sum(norm_arr), np.argmax(norm_arr)

class SpatialAnomalyDetector():
  """
  spatial AD model.
  """
  def __init__(self, model_dir):
    self.model_dir = model_dir
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.rec_error_mean = None
    self.rec_error_var = None
    self.anomaly_threshold = None
    
    self.params_path = None

  def _is_anomaly_from_reconstruction_error(self, recon_error):
    anomaly_likelihood_score, anomaly_location = rec_error_likelihood(recon_error, self.rec_error_mean, self.rec_error_var)
    #print(anomaly_likelihood_score)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return anomaly_location, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def set_mean_var_anomalyThreshold(self):
    val_data = self.data_by_fold.get_validation_data()
    val_rec_errors = get_rec_error_from_saved_params(self.params_path, val_data)
    self.rec_error_mean = val_rec_errors.mean(axis=0)
    self.rec_error_var = val_rec_errors.var(axis=0)
    scores = np.zeros(len(val_rec_errors))
    for i,val_error in enumerate(val_rec_errors):
      scores[i] = rec_error_likelihood(val_error, self.rec_error_mean, self.rec_error_var)[0]
    self.anomaly_threshold = scores.mean() + 3 * scores.std()
    
    
    print('anomaly threshold set to:', self.anomaly_threshold)
    

    
    train_data =self.data_by_fold.get_training_data()
    test_data = self.data_by_fold.get_test_data()
    
    train_rec_errors = get_rec_error_from_saved_params(self.params_path, train_data)
    test_rec_errors = get_rec_error_from_saved_params(self.params_path, test_data)
    
    train_MSE = np.mean(np.square(train_rec_errors))
    print('train_MSE', train_MSE)
    train_MSEs.append(train_MSE)
    val_MSEs.append(np.mean(np.square(val_rec_errors)))
    test_MSEs.append(np.mean(np.square(test_rec_errors)))
    

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = Spatial_data_by_fold(self.test_fold, self.val_fold, fold_ranges)
      self.params_path = self.model_dir.format(test_fold, val_fold)
      self.set_mean_var_anomalyThreshold()
      
      
      
      
  
  def get_score(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    recs = get_rec_error_from_saved_params(self.params_path, self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    scores = []
    global reconstructed_sequence_global_f
    global reconstructed_sequence_global_n
    global original_sequence_global_f
    global original_sequence_global_n
    for rec_error in recs:
      anomaly_score = rec_error_likelihood(rec_error, self.rec_error_mean, self.rec_error_var)
      #print('location', anomaly_score[1])
      scores.append(anomaly_score[0])
    
    if not len(reconstructed_sequence_global_f):
      print('hallo')
      original_sequence_global_f = self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values)
      reconstructed_sequence_global_f = get_recs_from_saved_params(self.params_path, self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    else:
      print('hallo2')
      original_sequence_global_n = self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values)
      reconstructed_sequence_global_n = get_recs_from_saved_params(self.params_path, self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
      #assert 1 == 3
 
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    recs = get_rec_error_from_saved_params(self.params_path, self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    is_anomaly = []
    for rec in recs:
      is_anomaly.append(self._is_anomaly_from_reconstruction_error(rec))
    return is_anomaly

# COMMAND ----------

# To plot reconstruction! Delete databyfold over, original is in func_AD

original_sequence_global_f = None
reconstructed_sequence_global_f = np.asarray([])
reconstructed_sequence_global_n = False
original_sequence_global_n = None
model_dir = model_dir ='/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_4650_allDataApril/params_t{}v{}/model.ckpt'
df_segment, segment = None, None

train_MSEs, val_MSEs, test_MSEs = [],[],[]

res = get_all_anomaly_scores(SpatialAnomalyDetector, model_dir, fault_dir,  columns=[6290], fault_types =['typeII'], folds=[1], #, 'typeIV'],,
                             save_path=False) #'/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/CAE4650_score_summed.json')

# COMMAND ----------

print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('{:.4} & {:.4} & {:.4}'.format(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

n_score_global.shape

# COMMAND ----------

plt.close()
fig, axis = plt.subplots(2,1, sharey=True, sharex=True)
axis[0].set_title('CAE individual anomaly scores')
axis[1].set_ylabel('Anomalous data')
axis[0].set_ylabel('Normal data')
axis[1].set_xlabel('segment index')

axis[0].axvline(6290, color='red', alpha=0.8, linestyle=':')
axis[1].axvline(6290, color='red', alpha=0.8, linestyle=':')
axis[0].plot(n_score_global)
axis[1].plot(f_score_global)
axis[1].legend(loc=3)
axis[1].set_xlim(-500, 40000)
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.02)
display(plt.show())


# COMMAND ----------

plt.close()
fig, axis = plt.subplots(2,1, sharey=True, sharex=True)
axis[0].set_title('CAE reconstructions')
axis[1].set_ylabel('Anomalous data')
axis[0].set_ylabel('Normal data')
axis[1].set_xlabel('segment index')
axis[0].set_yticks([])
axis[0].axvline(6290, color='red', alpha=0.8, linestyle=':')
axis[1].axvline(6290, color='red', alpha=0.8, linestyle=':')
axis[1].plot(original_sequence_global_f[-1,:], label='original')
axis[1].plot(reconstructed_sequence_global_f[-1,:], color='darkorange', label='reconstuction')
axis[0].plot(original_sequence_global_n[-1,:])
axis[0].plot(reconstructed_sequence_global_n[-1,:], color='darkorange')
axis[1].legend(loc=1)
axis[1].set_xlim(-500, 40000)
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.02)
display(plt.show())

# COMMAND ----------

result = json.loads(res)

tpr = result['true_positive_rates']['typeII']['6290']['1']
fpr = result['false_positive_rates']['typeII']['6290']['1']

plt.close()
plt.plot(fpr, tpr)
plt.xlim([-0.05, 1.05])
plt.ylim([-0.05, 1.05])
display(plt.show())

# COMMAND ----------

columns = [6290]

for column in []: #columns:
  log = Log('/dbfs/mnt/renewablesstudents/user-write/models/p2/evaluation/column_{}'.format(column), '/CAE_4noe_testSTD')
  if column != 'random':
    df_segment, segment  = load_segment_data(subset=SUBSET, column=column)
  anomaly_detector = SpatialAnomalyDetector(model_dir)
  ad_results = get_AD_results_V2(anomaly_detector, fault_dir, column=column, log=log, folds=[0])

# COMMAND ----------

#dbutils.fs.rm('dbfs:/mnt/renewablesstudents/user-write/models/p2/evaluation/column_6290/CAE_4650.txt')

# COMMAND ----------

fold = 0
fault_type = 'typeI'
column = 6290

fold_dir = fault_dir + 'test_fold{}/'.format(fold) + 'column_{}/'.format(column)

data_path = glob.glob(fold_dir + fault_type + '/*.csv')[0]
metadata_path = glob.glob(fold_dir + fault_type + '/*.json')[0]
fault_df, metadata = get_fault_data(data_path, metadata_path)
normal_df = get_normal_data(metadata_path, column)

# COMMAND ----------

metadata

# COMMAND ----------

fault_df.shape

# COMMAND ----------

def get_re(data, metadata):
  tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
  df_orig = df_by_time_range(df_dts, tr).copy()
  assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
  df_orig[metadata['segments']] = data[metadata['segments']].values
  recs = get_recs_from_saved_params(model_dir, dbf.normalize_dts(df_orig.iloc[:,:n_segments].values))
  err = get_rec_error_from_saved_params(model_dir, dbf.normalize_dts(df_orig.iloc[:,:n_segments].values))
  return recs[100,:], dbf.normalize_dts(df_orig.iloc[100,:n_segments].values), err[100,:]
  
f_rec, f_orig, f_err = get_re(fault_df, metadata) 
n_rec, n_orig, n_err = get_re(normal_df, metadata)
f_rec.shape, f_orig.shape

# COMMAND ----------

fig, (ax1, ax2, ax3) = plt.subplots(3,2, figsize=(10,5))

ax1[0].plot(f_rec)
ax2[0].plot(f_orig)
ax3[0].plot(f_err)
ax1[1].plot(n_rec)
ax2[1].plot(n_orig)
ax3[1].plot(n_err)

display(plt.show())

# COMMAND ----------

fig, (ax1, ax2, ax3) = plt.subplots(3,2, figsize=(10,5))

ax1[0].plot(f_rec)
ax2[0].plot(f_orig)
ax3[0].plot(f_err)
ax1[1].plot(n_rec)
ax2[1].plot(n_orig)
ax3[1].plot(n_err)

display(plt.show())

# COMMAND ----------

from sklearn.decomposition import PCA
dbf = Spatial_data_by_fold(0, 1, fold_ranges)
pca = PCA(n_components = 100, svd_solver='auto')
dts_norm = dbf.normalize_dts(df_dts.iloc[:,:n_segments].values)

# COMMAND ----------

dts_norm.shape

# COMMAND ----------

pca.fit(dts_norm[:, :])

# COMMAND ----------

sum(pca.explained_variance_ratio_), pca.explained_variance_ratio_

# COMMAND ----------

pca.components_.shape

# COMMAND ----------

import matplotlib.pyplot as plt
import seaborn as sns

plt.close()
sns.heatmap(pca.components_[:,:40])
display(plt.show())

# COMMAND ----------

sig = n_orig
tf = pca.transform(sig)
tra = pca.transform(sig.reshape(1,-1))
recon = (tra@pca.components_).reshape(-1)

plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,1, figsize=(10,5), sharey=True)

ax1.plot(sig)
ax2.plot(recon)
ax3.plot(sig-recon)
#ax1[1].plot(n_rec)
#ax2[1].plot(n_orig)
#ax3[1].plot(n_err)

display(plt.show())

# COMMAND ----------

f_orig.shape

# COMMAND ----------

tra.shape

# COMMAND ----------



# COMMAND ----------

recon.shape

# COMMAND ----------


