# Databricks notebook source
# MAGIC %run ./func_spatial

# COMMAND ----------

from scipy.sparse import diags
from scipy.sparse.linalg import factorized
from scipy.linalg import lu_factor, lu_solve

# COMMAND ----------

SUBSET = True

fold_ranges = get_fold_ranges(subset=SUBSET)  
df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']

# COMMAND ----------

df_dts.new_timestamp

# COMMAND ----------

def convolutionalAutoencoder(lr, log=False):
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()

  x = tf.placeholder("float", [None, n_segments])
  bs = tf.placeholder("int32")

  inp = tf.reshape(x, shape=[-1, n_segments, 1])  
  log.write('learning rate: {}'.format(lr))
  log.write('input {}'.format(inp))
      
  n_filters = 30
  ksize1 = [5, 1, n_filters]
  ksize2 = [5, n_filters, n_filters]
  ksize3 = [5, n_filters, n_filters]
  ksize4 = [5, n_filters, n_filters]
  p1 = 4
  p2 = 4
  p3 = 4
  p4 = 4
  
  log.write('ksize 1: {}, max pool 1: {}'.format(ksize1, p1))
  log.write('ksize 2: {}, max pool 2: {}'.format(ksize2, p2))
  log.write('ksize 3: {}, max pool 3: {}'.format(ksize3, p3))
  log.write('ksize 4: {}, max pool 4: {}'.format(ksize4, p4))

  W = tf.get_variable('W', shape=(n_filters,1), initializer=tf.random_normal_initializer())
  B = tf.get_variable('B', shape=(1), initializer=tf.zeros_initializer())

  c1, s1 = conv(inp, ksize1, 1, 'conv1')
  log.write('conv1: {}'.format(c1))
  mp1 = max_pool(c1, p1, p1, 'mp1', ksize1[2], bs)
  log.write('mp1: {}'.format(mp1))
  c2, s2 = conv(mp1, ksize2, 1, 'conv2')
  log.write('conv2: {}'.format(c2))
  mp2 = max_pool(c2, p2, p2, 'mp2', ksize2[2], bs)
  log.write('mp2: {}'.format(mp2))
  c3, s3 = conv(mp2, ksize3, 1, 'conv3')
  log.write('c3: {}'.format(c3))
  mp3 = max_pool(c3, p3, p3, 'mp3', ksize3[2], bs)
  log.write('mp3: {}'.format(mp3))
  c4, s4 = conv(mp3, ksize4, 1, 'conv4')
  log.write('c4: {}'.format(c4))
  mp4 = max_pool(c4, p4, p4, 'mp4', ksize4[2], bs)
  log.write('mp4: {}'.format(mp4))
  
  encoding = tf.identity(mp4, name='encoding')
  
  dc4 = deconv(mp4, ksize4, s4, 'deconv4', int(np.ceil(n_segments/(p1*p2*p3*p4))), bs)
  log.write('dc4: {}'.format(dc4))
  up4 = upsample(dc4, 'unpool4', int(np.ceil(n_segments/(p1*p2*p3))), ksize4[1], bs)
  log.write('up4: {}'.format(up4))
  dc3 = deconv(up4, ksize3, s3,'deconv3', int(np.ceil(n_segments/(p1*p2*p3))), bs)
  log.write('dc3: {}'.format(dc3))
  up3 = upsample(dc3, 'unpool0', int(np.ceil(n_segments/(p1*p2))), ksize3[1], bs)
  log.write('up3: {}'.format(up3))
  dc2 = deconv(up3, ksize2, s2,'deconv2', int(np.ceil(n_segments/(p1*p2))), bs)
  log.write('dc2: {}'.format(dc2))
  up2 = upsample(dc2, 'unpool1', n_segments//p1, ksize2[1], bs)
  log.write('up2: {}'.format(up2))
  dc1 = deconv(up2, [ksize1[0],  ksize2[1], n_filters], s1,'deconv1', int(np.ceil(n_segments//p1)), bs)
  log.write('dc1: {}'.format(dc1))
  up1 = upsample(dc1, 'unpool1', n_segments, ksize1[2], bs)
  log.write('up1: {}'.format(up1))

  res = tf.matmul(tf.reshape(up1, shape = [-1, n_filters]), W) + B
  log.write('res: {}'.format(res))

  r = tf.reshape(res, shape=[-1, n_segments])

  loss_op = tf.losses.mean_squared_error(labels=x, predictions=r)
  optimizer = tf.train.AdamOptimizer(learning_rate=lr)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, r, x, bs
#convolutionalAutoencoder(lr = 0.001)

# COMMAND ----------

models_dir = 'None'
file_suffix = False
log = Log(False, file_suffix)
train_op, r, x, bs = convolutionalAutoencoder(lr = 0.001, log=log)
saver = tf.train.Saver()

# COMMAND ----------

model_path = '/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/test/test_4650_n6/params_t0v1/model.ckpt'
train_data = Spatial_data_by_fold(0, 2, fold_ranges).get_training_data()
train_recs = get_recs_from_saved_params(model_path, train_data)
val_data = Spatial_data_by_fold(0, 2, fold_ranges).get_validation_data()
val_recs = get_recs_from_saved_params(model_path, val_data)

# COMMAND ----------

val_error = get_rec_error_from_saved_params(model_path, val_data)

# COMMAND ----------

train_error = get_rec_error_from_saved_params(model_path, train_data)

# COMMAND ----------

plt.close()
plt.plot(val_error[4,:])
#plt.plot(val_error[1000,:])
plt.plot(train_error[600,:], color='gold')
display(plt.show())
#pass

# COMMAND ----------

val_err = val_error[:, :]
val_err.shape

# COMMAND ----------

#val_cov = np.cov(val_err, rowvar=False)
val_mean = val_err.mean(axis=0)
val_var = val_err.var(axis=0)
#lu_piv = lu_factor(val_cov)
#print("Lu and piv constructed!", lu.shape, piv.shape)
#sparse_val_cov = band_matrix(1, val_cov.shape).multiply(val_cov)
#solver = factorized(sparse_val_cov)
#print('factorized!')

# COMMAND ----------

def band_matrix(band_width, shape):
  assert band_width % 2, 'band_width must be odd but was {}'.format(band_width)
  return diags([1 for _ in range(band_width)], [i for i in range(-(band_width//2), band_width//2+1)], shape=shape)

def rec_error_likelihood(recs, mean, cov=None, verbose=False, solver=None, lu_and_piv=False):
  if verbose:
    print('recs', recs.shape, 'mean', mean.shape)
  assert len(recs.shape) == 1, 'recs must be 1d but has shape {}'.format(recs.shape)
  temp = np.expand_dims(recs - mean, axis=1)
  if verbose:
    print('temp.T', temp.T.shape, 'temp',  temp.shape)
  if solver:
    x = solver(temp)
  elif cov:
    x = np.linalg.solve(cov, temp)
  elif lu_and_piv:
    x = lu_solve(lu_and_piv, temp)
  a = temp.T @ x
  
  if verbose:
    print((np.square(temp).squeeze()/val_var).shape)
    print(np.sum(np.square(temp).squeeze()/val_var), np.mean(np.square(temp).squeeze()/val_var))
  
  return float(np.squeeze(a))

# COMMAND ----------

def alt_likelihood(recs, mean, var):
  temp = recs - mean
  norm_arr = np.square(temp)/var
  print("sum", np.sum(norm_arr), "mean", np.mean(norm_arr), "max", np.max(norm_arr), "min", np.min(norm_arr), "argmax", np.argmax(norm_arr))
  return norm_arr, np.max(norm_arr)/np.mean(norm_arr)

# COMMAND ----------

# Need to see error in light of other error

index1 = 90
index2 = 1069

fault_index = 23000

fault1 = val_data[index1,:].copy()
fault1[fault_index] = fault1[fault_index] + 0.1 * np.abs(fault1[fault_index])
fault2 = val_data[index2,:].copy()
fault2[fault_index] = fault2[fault_index] + 0.1 * np.abs(fault2[fault_index])

fault1_rec = get_rec_error_from_saved_params(model_path, np.expand_dims(fault1, axis=0))
fault2_rec = get_rec_error_from_saved_params(model_path, np.expand_dims(fault2, axis=0))
no_fault1_rec = get_rec_error_from_saved_params(model_path, np.expand_dims(val_data[index1,:], axis=0))
no_fault2_rec = get_rec_error_from_saved_params(model_path, np.expand_dims(val_data[index2,:], axis=0))

# COMMAND ----------

n1, s1 = alt_likelihood(fault1_rec, val_mean, val_var)
n1n, s1n = alt_likelihood(no_fault1_rec, val_mean, val_var)
n2, s2 = alt_likelihood(fault2_rec, val_mean, val_var)
n2n, s2n = alt_likelihood(no_fault2_rec, val_mean, val_var)
s1, s1n, s2, s2n

# COMMAND ----------

plt.close()

fig, ax = plt.subplots(4,1)

ax[0].plot(fault1, color='r')
ax[0].plot(val_data[index1,:])

ax[1].plot(val_data[index1,:])
ax[1].plot(get_recs_from_saved_params(model_path, np.expand_dims(fault1, axis=0)).squeeze())
ax[1].plot(get_recs_from_saved_params(model_path, np.expand_dims(val_data[index1,:], axis=0)).squeeze(), color='r')

ax[2].plot(fault1_rec.squeeze(), color='r')
ax[2].plot(no_fault1_rec.squeeze())

ax[3].plot(n1.squeeze(), color='r')
ax[3].plot(n1n.squeeze())

display(plt.show())

# COMMAND ----------

plt.close()
fig, ax = plt.subplots(4,1)

ax[0].plot(fault2, color='r')
ax[0].plot(val_data[index2,:])

ax[1].plot(val_data[index2,:])
ax[1].plot(get_recs_from_saved_params(model_path, np.expand_dims(fault2, axis=0)).squeeze())
ax[1].plot(get_recs_from_saved_params(model_path, np.expand_dims(val_data[index2,:], axis=0)).squeeze(), color='r')

ax[2].plot(fault2_rec.squeeze(), color='r')
ax[2].plot(no_fault2_rec.squeeze())

ax[3].plot(n2.squeeze(), color='r')
ax[3].plot(n2n.squeeze())

display(plt.show())

# COMMAND ----------

a_vector = []
for i in range(len(val_err)):
  a_vector.append(np.sum(alt_likelihood(val_err[i,:], val_mean, val_var))[1])

# COMMAND ----------

#train_a_vector = []
#for i in range(len(train_error)):
#  train_a_vector.append(rec_error_likelihood(train_error[i,:], val_mean, solver=solver))
pass

# COMMAND ----------

av = np.array(a_vector)
anomaly_threshold = av.mean() + 2*av.std()
print('anomaly threshold', anomaly_threshold)
print("Non-anomalous val samples marked as anomalies {} out of {}".format(sum(av > anomaly_threshold), len(av)))

# COMMAND ----------

np.min(a_vector), np.max(a_vector), np.argmin(a_vector), np.argmax(a_vector)

# COMMAND ----------

plt.close()
plt.plot(a_vector[:])
#plt.plot(train_a_vector)
plt.axhline(anomaly_threshold)
display(plt.show())

# COMMAND ----------

plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,1)
ax1.plot(train_data[0,-100:], color='g')
ax1.plot(train_recs[0,-100:], color='r')
ax2.plot(train_recs[0,:], color='r')
ax2.plot(train_data[0,:])
ax3.plot(train_data[0,:]-train_recs[0,:])

display(plt.show())

# COMMAND ----------

plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,2, sharey=True)
ax1[0].plot(val_recs[np.argmin(a_vector),:100], color='r')
ax1[0].plot(val_data[np.argmin(a_vector),:100], color='g')
ax2[0].plot(val_recs[np.argmin(a_vector),:], color='r')
ax2[0].plot(val_data[np.argmin(a_vector),:], color='g')
ax3[0].plot(val_error[np.argmin(a_vector),:])

ax3[1].plot(val_error[np.argmax(a_vector),:]+0.2, color='g')
ax1[1].plot(val_recs[np.argmax(a_vector),:100], color='r')
ax1[1].plot(val_data[np.argmax(a_vector),:100], color='g')
ax2[1].plot(val_recs[np.argmax(a_vector),:], color='r')
ax2[1].plot(val_data[np.argmax(a_vector),:], color='g')
display(plt.show())

# COMMAND ----------

plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,2, sharey=True)
ax1[0].plot(train_recs[np.argmin(train_a_vector),:100], color='r')
ax1[0].plot(train_data[np.argmin(train_a_vector),:100], color='g')
ax2[0].plot(train_recs[np.argmin(train_a_vector),:], color='r')
ax2[0].plot(train_data[np.argmin(train_a_vector),:], color='g')
ax3[0].plot(train_error[np.argmin(train_a_vector),:])

ax3[1].plot(train_error[np.argmax(train_a_vector),:]+0.2, color='g')
ax1[1].plot(train_recs[np.argmax(train_a_vector),:100], color='r')
ax1[1].plot(train_data[np.argmax(train_a_vector),:100], color='g')
ax2[1].plot(train_recs[np.argmax(train_a_vector),:], color='r')
ax2[1].plot(train_data[np.argmax(train_a_vector),:], color='g')
display(plt.show())

# COMMAND ----------


