# Databricks notebook source
#import seaborn as sns

# COMMAND ----------

# MAGIC %run ./func_spatial

# COMMAND ----------

# MAGIC %run ../temporal/func_AD_evaluation

# COMMAND ----------

from sklearn.decomposition import PCA

# COMMAND ----------

SUBSET = False
fold_ranges = get_fold_ranges(subset=SUBSET)  
df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
df_current = load_current_data(SUBSET)
df_seatemp = load_seatemp_data(SUBSET)
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']

# COMMAND ----------

df_normalized = df_dts.iloc[:,:39516] - df_dts.iloc[:,:39516].mean(axis=0) / df_dts.iloc[:,:39516].std(axis=0)
PCA2000 = PCA(2000)
PCA2000 = PCA2000.fit(df_normalized)
for comps in [1, 10,25,50,100,250,500, 1000,2000]:
  print('{} & {:.4f}\% \\\ \hline'.format(comps, 100*sum(PCA2000.explained_variance_ratio_[:comps])))

# COMMAND ----------

class PCA_AnomalyDetector(AnomalyDetector):
  """
  spatial AD model.
  """
  def __init__(self, model_dir=''): #pca_components=100):
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.rec_error_mean = None
    self.rec_error_var = None
    self.anomaly_threshold = None
    self.pca = PCA(n_components=PCA_COMPONENTS, svd_solver='auto')

  def _is_anomaly_from_reconstruction_error(self, recon_error):
    anomaly_likelihood_score, anomaly_location = self.__rec_error_likelihood(recon_error)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return anomaly_location, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def __rec_error_likelihood(self, recs):
    temp = recs - self.rec_error_mean
    norm_arr = np.square(temp)/self.rec_error_var
    #global global_scores_f
    #global global_scores_n
    #if not len(reconstructed_sequence_global_f):
    #  global_scores_f = norm_arr
    #else:
    #  global_scores_n = norm_arr
    return np.max(norm_arr)/np.mean(norm_arr), np.argmax(norm_arr)
  
  def _get_pca_recs_error(self, data):
    transformed = self.pca.transform(data)
    #print(transformed.shape)
    #print('transformed -1', transformed[-1,:3])
    return data - self.pca.inverse_transform(transformed)  #transformed@self.pca.components_
  
  def set_mean_var_anomalyThreshold(self):
    val_data = self.data_by_fold.get_validation_data()
    val_rec_errors = self._get_pca_recs_error(val_data)
    self.rec_error_mean = val_rec_errors.mean(axis=0)
    self.rec_error_var = val_rec_errors.var(axis=0)
    scores = np.zeros(len(val_rec_errors))
    for i,val_error in enumerate(val_rec_errors):
      scores[i] = self.__rec_error_likelihood(val_error)[0]
    self.anomaly_threshold = np.quantile(scores, 1)
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = Spatial_data_by_fold(self.test_fold, self.val_fold, fold_ranges)
      #self.data_by_fold = dbf
      #global dbf 
      #dbf = self.data_by_fold
      self.pca.fit(self.data_by_fold.get_training_data())
      #global pca_global
      #pca_global = self.pca
      #self.pca = pca_global
      print('PCA fit with {} components! Explained variance ratio: {}'.format(self.pca.n_components_, sum(self.pca.explained_variance_ratio_)))
      self.set_mean_var_anomalyThreshold()
      
      #Calculate MS reconstruction error.
      
      #print('MSE train:', np.mean(np.square(self._get_pca_recs_error(self.data_by_fold.get_training_data()))))
      #print('MSE val:', np.mean(np.square(self._get_pca_recs_error(self.data_by_fold.get_validation_data()))))
      #print('MSE test:', np.mean(np.square(self._get_pca_recs_error(self.data_by_fold.get_test_data()))))
      train_MSEs.append(np.mean(np.square(self._get_pca_recs_error(self.data_by_fold.get_training_data()))))
      val_MSEs.append(np.mean(np.square(self._get_pca_recs_error(self.data_by_fold.get_validation_data()))))
      test_MSEs.append(np.mean(np.square(self._get_pca_recs_error(self.data_by_fold.get_test_data()))))
  
  def get_score(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    rec_errors = self._get_pca_recs_error(self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    scores = []
    for rec_error in rec_errors:
      scores.append(self.__rec_error_likelihood(rec_error)[0])
    #global reconstructed_sequence_global_f
    #global reconstructed_sequence_global_n
    #global original_sequence_global_f
    #global original_sequence_global_n
    #if not len(reconstructed_sequence_global_f):
    #  print('hallo')
     # original_sequence_global_f = self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values)
     # reconstructed_sequence_global_f =   self.pca.inverse_transform(self.pca.transform(original_sequence_global_f))
   # else:
     # print('hallo2')
    #  original_sequence_global_n = self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values)
      #reconstructed_sequence_global_n = self.pca.inverse_transform(self.pca.transform(original_sequence_global_n))
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    df_orig = df_by_time_range(df_dts, tr).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    
    rec_error = self._get_pca_recs_error(self.data_by_fold.normalize_dts(df_orig.iloc[:,:n_segments].values))
    is_anomaly = []
    for rec in rec_error:
      is_anomaly.append(self._is_anomaly_from_reconstruction_error(rec))
    return is_anomaly

# COMMAND ----------

PCA_COMPONENTS = 25

#original_sequence_global_f = None
#reconstructed_sequence_global_f = np.asarray([])
#reconstructed_sequence_global_n = False
#original_sequence_global_n = None
#global_scores_f = None
#global_scores_n = None
#df_segment, segment = None, None
train_MSEs, val_MSEs, test_MSEs = [],[],[]

res = get_all_anomaly_scores(PCA_AnomalyDetector, '', fault_dir, fault_types =['typeII'], #, 'typeIV'],
#                             columns=[6290], save_path=False, folds=[1])
                             save_path =False, columns=[6290])
                             #save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA_nc{}.json'.format(PCA_COMPONENTS))
print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('PCA & {} &{:.4} & {:.4} & {:.4}'.format(PCA_COMPONENTS, np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

PCA_COMPONENTS = 100
train_MSEs, val_MSEs, test_MSEs = [],[],[]

res = get_all_anomaly_scores(PCA_AnomalyDetector, '', fault_dir, fault_types =['typeII'], #, 'typeIV'],
#                             columns=[6290], save_path=False, folds=[1])
                             save_path =False, columns=[6290])
                             #save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA_nc{}.json'.format(PCA_COMPONENTS))
print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('PCA & {} &{:.4} & {:.4} & {:.4}'.format(PCA_COMPONENTS, np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

PCA_COMPONENTS = 50
train_MSEs, val_MSEs, test_MSEs = [],[],[]

res = get_all_anomaly_scores(PCA_AnomalyDetector, '', fault_dir, fault_types =['typeII'], #, 'typeIV'],
#                             columns=[6290], save_path=False, folds=[1])
                             save_path =False, columns=[6290])
                             #save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA_nc{}.json'.format(PCA_COMPONENTS))
print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('PCA & {} &{:.4} & {:.4} & {:.4}'.format(PCA_COMPONENTS, np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

PCA_COMPONENTS = 2000
train_MSEs, val_MSEs, test_MSEs = [],[],[]

res = get_all_anomaly_scores(PCA_AnomalyDetector, '', fault_dir, fault_types =['typeII'], #, 'typeIV'],
#                             columns=[6290], save_path=False, folds=[1])
                             save_path =False, columns=[6290])
                             #save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA_nc{}.json'.format(PCA_COMPONENTS))
print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('PCA & {} &{:.4} & {:.4} & {:.4}'.format(PCA_COMPONENTS, np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

train_MSEs

# COMMAND ----------

df_dts.iloc[:,:-3].mean(axis=0).shape
df_dts_norm = (df_dts.iloc[:,:-3] - df_dts.iloc[:,:-3].mean(axis=0)) / df_dts.iloc[:,:-3].std(axis=0)

# COMMAND ----------

plt.close()
fig, ax = plt.subplots(1,1, figsize=(15,5))
ax.plot(df_dts_norm.iloc[17837])
ax.set_xlim(-500,40000)
#ax.set_ylim(-3.2,2)
ax.set_title('Sequence of the  DTS values at a single time step')
ax.set_ylabel('normalized DTS values')
ax.set_xlabel('Segment')

display(plt.show())

# COMMAND ----------

print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('{:.4} & {:.4} & {:.4}'.format(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

pca_global.components_.shape

# COMMAND ----------

plt.close()
fig, ax = plt.subplots(4,1, sharex=True)
ax[0].plot(pca_global.components_[0,:])
ax[1].plot(pca_global.components_[1,:])
ax[2].plot(pca_global.components_[2,:])
ax[0].set_xlim(-1000, 40500)
ax[0].set_yticks([0, 0.005])
ax[0].set_ylim(-0.002, 0.006)
ax[1].set_ylim(-0.03, 0.06)
ax[1].set_yticks([0, 0.05])
ax[2].set_yticks([0, 0.05])
ax[3].set_yticks([-1, 0.5])
#ax[0].set_title('3 first components')
#ax[2].set_xlabel('segment')
#ax[0].set_ylabel('PC 1')
#ax[1].set_ylabel('PC 2')
#ax[2].set_ylabel('PC 3')
ax[0].set_xticks([])
ax[3].set_ylim(-1.5,1)

top_3 = 11.52685595*pca_global.components_[0,:] +  -28.24000627  *pca_global.components_[1,:] + 0.56571097*pca_global.components_[2,:]
#ax[3].plot(original_sequence_global_n[-1,:])
ax[3].plot(top_3)

display(plt.show())

# COMMAND ----------

original_sequence_global_f.shape

# COMMAND ----------

PCA_COMPONENTS = 4000

original_sequence_global_f = None
reconstructed_sequence_global_f = np.asarray([])
reconstructed_sequence_global_n = False
original_sequence_global_n = None
global_scores_f = None
global_scores_n = None
df_segment, segment = None, None
train_MSEs, val_MSEs, test_MSEs = [],[],[]
res = get_all_anomaly_scores(PCA_AnomalyDetector, '', fault_dir, fault_types =['typeII', 'typeIV'],
                             #columns=[6290], save_path=False) #, folds=[1])
                             save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA_nc{}.json'.format(PCA_COMPONENTS))
print(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs))
print('{:.4} & {:.4} & {:.4}'.format(np.mean(train_MSEs), np.mean(val_MSEs), np.mean(test_MSEs)))

# COMMAND ----------

pca_global.components_.shape

# COMMAND ----------

plt.close()
sns.heatmap(pca_global.components_[:50, 0:20])
display(plt.show())

# COMMAND ----------

plt.close()
sns.heatmap(pca_global.components_[:50, -20:])
display(plt.show())

# COMMAND ----------

pca_global.components_[:,0]

# COMMAND ----------

plt.close()
fig, axis = plt.subplots(2,1, sharey=False, sharex=True)
axis[0].set_title('Individual anomaly scores from PCA with 500 components')
axis[1].set_ylabel('Anomalous data')
axis[0].set_ylabel('Normal data')
axis[1].set_xlabel('segment index')

#axis[0].axvline(6290, color='red', alpha=0.8, linestyle=':')
#axis[1].axvline(6290, color='red', alpha=0.8, linestyle=':')
axis[0].plot(global_scores_n)
axis[1].plot(global_scores_f)
axis[1].legend(loc=3)
axis[0].set_xlim(-500, 40000)
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.06)
display(plt.show())


# COMMAND ----------

plt.close()
plt.plot(global_scores_f)
plt.plot(global_scores_n)
display(plt.show())

# COMMAND ----------

plt.close()
plt.plot(global_scores_f)
plt.plot(global_scores_n)
display(plt.show())

# COMMAND ----------

plt.close()
fig, axis = plt.subplots(2,1, sharey=True, sharex=True)
axis[0].set_title('Reconstruction using PCA with 500 components')
axis[1].set_ylabel('Anomalous data')
axis[0].set_ylabel('Normal data')
axis[1].set_xlabel('segment index')
axis[0].set_yticks([])
#axis[0].axvline(6290, color='green', alpha=0.5, linewidth=0.5)
#axis[1].axvline(6290, color='green', alpha=0.5, linewidth=0.5)
axis[0].plot(original_sequence_global_n[-1,:])
axis[0].plot(reconstructed_sequence_global_n[-1,:], color='darkorange')
axis[1].plot(original_sequence_global_f[-1,:], label='original')
axis[1].plot(reconstructed_sequence_global_f[-1,:], color='darkorange', label='reconstruction')
axis[1].legend(loc=3)
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.02)
display(plt.show())

# COMMAND ----------

plt.close()
fig, axis = plt.subplots(2,1, sharey=True, sharex=True)
axis[0].set_title('Reconstruction using PCA with 500 components')
axis[1].set_ylabel('Anomalous data')
axis[0].set_ylabel('Normal data')
axis[1].set_xlabel('segment index')
axis[0].set_yticks([])
#axis[0].axvline(6290, color='green', alpha=0.5, linewidth=0.5)
#axis[1].axvline(6290, color='green', alpha=0.5, linewidth=0.5)
axis[0].plot(original_sequence_global_n[-1,:])
axis[0].plot(reconstructed_sequence_global_n[-1,:], color='darkorange')
axis[1].plot(original_sequence_global_f[-1,:], label='original')
axis[1].plot(reconstructed_sequence_global_f[-1,:], color='darkorange', label='reconstruction')
axis[1].legend(loc=3)
axis[0].set_xlim(-500, 40000)
plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.02)
display(plt.show())

# COMMAND ----------

result = json.loads(res)

tpr = result['true_positive_rates']['typeII']['6290']['1']
fpr = result['false_positive_rates']['typeII']['6290']['1']

plt.close()
plt.plot(fpr, tpr)
plt.xlim([-0.05, 1.05])
plt.ylim([-0.05, 1.05])
display(plt.show())

# COMMAND ----------

columns = [6290, 17712, 32028] #, 'random'
#PCA_COMPONENTS = 50

for column in [6290]: #columns:
  log = Log() #Log('/dbfs/mnt/renewablesstudents/user-write/models/p2/evaluation/column_{}'.format(column), '/PCA_{}'.format(PCA_components))
  if column != 'random':
    df_segment, segment  = load_segment_data(subset=SUBSET, column=column)
  anomaly_detector = PCA_AnomalyDetector()
  ad_results = get_AD_results_V2(anomaly_detector, fault_dir, column=column, log=log)

# COMMAND ----------

fold = 0
fault_type = 'typeI'
column = 6290

fold_dir = fault_dir + 'test_fold{}/'.format(fold) + 'column_{}/'.format(column)

data_path = glob.glob(fold_dir + fault_type + '/*.csv')[0]
metadata_path = glob.glob(fold_dir + fault_type + '/*.json')[0]
fault_df, metadata = get_fault_data(data_path, metadata_path)
normal_df = get_normal_data(metadata_path, column)

# COMMAND ----------

def get_dts_frame_from(data, metadata):
  tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
  df_orig = df_by_time_range(df_dts, tr).copy()
  assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
  df_orig[metadata['segments']] = data[metadata['segments']].values
  return dbf.normalize_dts(df_orig.iloc[100,:n_segments].values)
  
f_orig= get_re(fault_df, metadata) 
n_orig= get_re(normal_df, metadata)
f_orig.shape, n_orig.shape

# COMMAND ----------

dbf = Spatial_data_by_fold(0, 1, fold_ranges)
pca = PCA(n_components = 100, svd_solver='auto')
dts_norm = dbf.normalize_dts(df_dts.iloc[:,:n_segments].values)

# COMMAND ----------

pca.fit(dbf.get_training_data())

# COMMAND ----------

sum(pca.explained_variance_ratio_), pca.explained_variance_ratio_

# COMMAND ----------

n_recon = (pca.transform(n_orig.reshape(1,-1))@pca.components_).reshape(-1)
f_recon = (pca.transform(f_orig.reshape(1,-1))@pca.components_).reshape(-1)

plt.close()
fig, (ax1, ax2, ax3) = plt.subplots(3,2, figsize=(10,5), sharey=True)

ax1[0].plot(n_orig)
ax2[0].plot(n_recon)
ax3[0].plot(n_orig-n_recon)
ax1[1].plot(f_orig)
ax2[1].plot(f_recon)
ax3[1].plot(f_orig-f_recon)

display(plt.show())

# COMMAND ----------

plt.close()
plt.plot(pca.explained_variance_ratio_)
display(plt.show())

# COMMAND ----------

pca.explained_variance_ratio_

# COMMAND ----------


