# Databricks notebook source
# MAGIC %run ../func_p2

# COMMAND ----------

#GPU
def get_reconstructions(sess, data):
  recs = []
  for batch in range(0, len(data), batch_size):
    recs.append(sess.run(r, feed_dict={x:data[batch:batch+batch_size], bs:len(data[batch:batch+batch_size])}))
  return np.concatenate(recs)

def get_loss(sess, data):
  recs = get_reconstructions(sess, data)
  losses = np.square(data-recs)
  return np.mean(losses)

# COMMAND ----------

'''
class AnomalyDetector:
  """
  AD model.
  """
  def __init__(self, model_dir=None):
    pass
  
  def is_anomalous(self, data, test_fold=-1):
    raise NotImplementedError
    
class Random_ad(AnomalyDetector):
  def __init__(self):
    pass
  
  def is_anomalous(self, data):
    return random.choice([True,False])
  
'''
pass

# COMMAND ----------

class Spatial_data_by_fold(Data_by_fold):
  def __init__(self, test_fold, val_fold, fold_ranges):
    super().__init__(test_fold, val_fold, fold_ranges, dts=True)
  
  def normalize_dts(self, dts_values):
    return (dts_values - self.dts_mean) / self.dts_std
  
  def __get_input_data_by_ranges(self, ranges):
    input_data = []
    for tr in ranges:
      range_values = df_by_time_range(df_dts, tr).iloc[:,:n_segments].values
      input_data.append(self.normalize_dts(range_values))
    return np.concatenate(input_data)
  
  def get_training_data(self):
    return self.__get_input_data_by_ranges(self.train_ranges)
    
  def get_validation_data(self):
    return self.__get_input_data_by_ranges(self.val_ranges)
  
  def get_test_data(self):
    return self.__get_input_data_by_ranges(self.test_ranges)

# COMMAND ----------

# Functions used to construct spatial ad model

def conv(x, ksize, stride, name):
  with tf.name_scope(name):
    W = tf.get_variable(name+'_W', shape=ksize, initializer=tf.random_normal_initializer())
    B = tf.get_variable(name+'_B', shape=(ksize[2]), initializer=tf.zeros_initializer())
    co = tf.nn.conv1d(x, W, stride=stride, padding="SAME")
    co = tf.nn.sigmoid(tf.nn.bias_add(co, B))
    return co, stride

def deconv(x, ksize, stride, name, output_len, bs):
  with tf.name_scope(name):
    W = tf.get_variable(name+'_W', shape=ksize, initializer=tf.random_normal_initializer())
    B = tf.get_variable(name+'_B', shape=(ksize[1]), initializer=tf.zeros_initializer())
    dc = tf.contrib.nn.conv1d_transpose(x, output_shape=[bs, output_len, ksize[1]], filter=W, stride=stride, padding="SAME")
    return tf.nn.sigmoid(tf.nn.bias_add(dc,B))

def max_pool(x, wsize, stride, name, filters, bs):
  with tf.name_scope(name):
    window_size = [1, 1, wsize, 1]
    strides = [1,1,stride,1]
    x = tf.reshape(x, shape=(bs, 1, -1, filters))
    mp = tf.nn.max_pool(x, ksize=window_size, strides=strides, padding='SAME')
    return tf.reshape(mp, shape=(bs, -1, filters))

def upsample(x, name, out_size, filters, bs):
    with tf.name_scope(name):
        x = tf.reshape(x, shape=(bs, 1, -1, filters))
        size = [1, out_size]
        out = tf.image.resize_bilinear(x, size=size, align_corners=None, name=None)
        return tf.reshape(out, shape=(bs, -1, filters))

# COMMAND ----------


