# Databricks notebook source
# MAGIC %run ../spatiotemporal/func_spatiotemporal

# COMMAND ----------

# MAGIC %run ../temporal/func_AD_evaluation

# COMMAND ----------

from sklearn.decomposition import PCA

# COMMAND ----------

SUBSET = False
fold_ranges = get_fold_ranges(subset=SUBSET)  
df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']

# COMMAND ----------

df_current = load_current_data(subset=SUBSET)
df_seatemp = load_seatemp_data(subset=SUBSET)

df_dts_joined = join_dts_curr_seatemp(df_dts, df_current, df_seatemp)
df_dts.shape, df_dts_joined.shape

# COMMAND ----------

list(df_dts_joined)[-5:]

# COMMAND ----------

dts_seq[6290]

# COMMAND ----------

plt.close()
fig, ax = plt.subplots(1,1, figsize=(15,4.8))
dts_seq = df_dts.iloc[78, :-3]
ax.plot(dts_seq)
ax.set_ylabel('DTS value')
ax.set_xlabel("segment")
ax.set_xlim(right=39516)
ax.set_title('Snapshot of DTS values with segments drawn for the temporal experiments')
ax.annotate('Segment 6284.571', (6290,dts_seq[6290]), (6290-2650,dts_seq[6290]+5), arrowprops={'arrowstyle': '->', 'lw': 4, 'color': 'red'})
ax.annotate('Segment 17696.341', (17712,dts_seq[17712]), (17712-2900,dts_seq[17712]-10), arrowprops={'arrowstyle': '->', 'lw': 4, 'color': 'red'})
ax.annotate('Segment 31999.511', (32028,dts_seq[32028]), (32028-2900,dts_seq[32028]-10), arrowprops={'arrowstyle': '->', 'lw': 4, 'color': 'red'})
display(plt.show())

# COMMAND ----------

df_dts.iloc[78:81, :-3].values.flatten().shape

# COMMAND ----------

list(df_dts)[-5:]

# COMMAND ----------

df_dts_joined['current'].iloc[ts+50:ts+53]

# COMMAND ----------

plt.close()
fig, (ax, ax2, ax3) = plt.subplots(3,1,figsize=(12,4.8), sharex='col')
ts = 9472
ax.plot(df_dts['new_timestamp'].iloc[ts:ts+50], df_dts.iloc[ts:ts+50, 6290], '.')
ax2.plot(df_dts_joined['new_timestamp'].iloc[ts:ts+50], df_dts_joined['current'].iloc[ts:ts+50], '.')
ax3.plot(df_dts_joined['new_timestamp'].iloc[ts:ts+50], df_dts_joined['sea_temp'].iloc[ts:ts+50], '.')
ax.plot(df_dts['new_timestamp'].iloc[ts+51:ts+53], df_dts.iloc[ts+51:ts+53, 6290], '.', color = 'aqua')
ax.plot(df_dts['new_timestamp'].iloc[ts+50:ts+51], df_dts.iloc[ts+50:ts+51, 6290], 'r.')
ax2.plot(df_dts_joined['new_timestamp'].iloc[ts+51:ts+53], df_dts_joined['current'].iloc[ts+51:ts+53], 'g.')
ax2.plot(df_dts_joined['new_timestamp'].iloc[ts+50:ts+51], df_dts_joined['current'].iloc[ts+50:ts+51], '.', color='black')
ax3.plot(df_dts_joined['new_timestamp'].iloc[ts+51:ts+53], df_dts_joined['sea_temp'].iloc[ts+51:ts+53], 'g.')
ax3.plot(df_dts_joined['new_timestamp'].iloc[ts+50:ts+51], df_dts_joined['sea_temp'].iloc[ts+50:ts+51], '.', color='black')
ax.set_ylabel('DTS value')
ax2.set_ylabel('Electric current')
ax3.set_ylabel('Sea temperature')
ax3.set_xlabel("Date")
ax.set_xlim(right=df_dts_joined['new_timestamp'].iloc[ts+54], left = df_dts_joined['new_timestamp'].iloc[ts-1])
ax3.set_ylim(top=5.1, bottom=4.5)
ax.set_title('Input and different target values for prediction models')
ax.axvline(df_dts_joined['new_timestamp'].iloc[ts+49])
ax3.axvline(df_dts_joined['new_timestamp'].iloc[ts+49])
ax2.axvline(df_dts_joined['new_timestamp'].iloc[ts+49])
display(plt.show())

# COMMAND ----------

plt.close()
fig, (ax, ax2, ax3) = plt.subplots(3,1,figsize=(12,4.8), sharex=True)
ax.plot(df_dts['new_timestamp'], df_dts.iloc[:, 6290])
ax2.plot(df_dts_joined['new_timestamp'], df_dts_joined['current'])
ax3.plot(df_dts_joined['new_timestamp'], df_dts_joined['sea_temp'])
ax.set_ylabel('DTS value')
ax2.set_ylabel('Electric current')
ax3.set_ylabel('Sea temperature')
ax3.set_xlabel("Date")
#ax.set_xlim(right=39516)
ax.set_title('Time series modelled by the temporal models')
display(plt.show())

# COMMAND ----------

class PCA_data_by_fold(Data_by_fold):
  def __init__(self, test_fold, val_fold, fold_ranges, steps=3, dts_only=True):
    super().__init__(test_fold, val_fold, fold_ranges, dts=False, curr=False, st=False)
    self.dts_only = dts_only
    if not dts_only:
      df = df_dts_joined
      self.dim = n_segments + 3
    else:
      df = df_dts.drop(columns=['sampling_interval'])
      self.dim = n_segments
    self.mean, self.std = self._get_mean_std([df_by_time_range(df, tr).drop(columns=['timestamp', 'new_timestamp']) for tr in self.train_ranges], axis=0)
    assert self.dim==len(self.mean)==len(self.std) and len(self.mean.shape) == len(self.std.shape) == 1, 'shapes should equal: dim, mean, std {} {} {}'.format(self.dim, self.mean.shape, self.std.shape)
    self.steps = steps
  
  def normalize(self, values):
    return (values - self.mean) / self.std
  
  def get_input_from_df(self, dts_values):
    data = self.normalize(dts_values)
    return self.__multiple_steps_input(data)
  
  def __get_input_data_by_ranges(self, ranges):
    input_data = []
    df = df_dts.drop(columns=['sampling_interval', 'timestamp']) if self.dts_only else df_dts_joined.drop(columns=['timestamp'])
    for tr in ranges:
      range_values = df_by_time_range(df, tr).drop(columns='new_timestamp').values
      assert range_values.shape[1] == self.dim, 'range_values.shape[1]({}) not equal to dim({})'.format(range_values.shape[1], self.dim)
      input_data.append(self.__multiple_steps_input(self.normalize(range_values)))
    return np.concatenate(input_data)
  
  def __multiple_steps_input(self, data):
    size = len(data) - self.steps + 1
    assert data.shape[1] == self.dim
    new_format = np.zeros((size, self.steps*self.dim))
    for i in range(size):
      row = np.concatenate([data[j] for j in range(i, i+self.steps)])
      new_format[i] = row
    return new_format
  
  def get_training_data(self):
    return self.__get_input_data_by_ranges(self.train_ranges)
    
  def get_validation_data(self):
    return self.__get_input_data_by_ranges(self.val_ranges)
  
  
#pca_dbf_with_curr = PCA_data_by_fold(0,1, fold_ranges, 3, dts_only=False)  
#t2 = pca_dbf_with_curr.get_training_data()

# COMMAND ----------

#print(t2.shape, t2.shape[1]/3)

# COMMAND ----------

#pca_dbf = PCA_data_by_fold(0,1, fold_ranges, 3)
#print('pca_dbf generated')
#t1 = pca_dbf.get_training_data()
#print(t1.shape, t1.shape[1]/3)

# COMMAND ----------

#print(t1.shape, t1.shape[1]/3, n_segments)

# COMMAND ----------

#pca = PCA(150, svd_solver='auto')
#pca.fit(t2)

# COMMAND ----------

#if pca:
#  print('pca!')

# COMMAND ----------

class PCA_AnomalyDetector_multidim(AnomalyDetector):
  """
  spatial AD model.
  """
  def __init__(self, pca_components=False, steps=False, quantile=0.999):
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.rec_error_mean = None
    self.rec_error_var = None
    self.anomaly_threshold = None
    self.steps=STEPS if not steps else steps
    
    self.pca = PCA(n_components=PCA_COMPONENTS if not pca_components else pca_components, svd_solver='auto')
    self.quantile=quantile
    self.dts_only=DTS_ONLY
    
    print('steps:{}, dts_only:{}, PCA_components:{}'.format(self.steps, self.dts_only, PCA_COMPONENTS))

  def _is_anomaly_from_reconstruction_error(self, recon_error):
    anomaly_likelihood_score, anomaly_location = self.__rec_error_likelihood(recon_error)
    if anomaly_likelihood_score > self.anomaly_threshold:
      #rec_error_likelihood(recon_error, self.rec_error_mean, self.rec_error_var, True)
      return anomaly_location%self.data_by_fold.dim, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def _get_pca_recs_error(self, data):
    transformed = self.pca.transform(data)
    return data - transformed@self.pca.components_
  
  def __rec_error_likelihood(self, recs, plot=False):
    temp = recs - self.rec_error_mean
    norm_arr = np.square(temp)/self.rec_error_var
    if plot:
        plt.close()
        fig, ax = plt.subplots(2, 1)
        ax[0].plot(recs)
        ax[1].plot(norm_arr)
        assert 5 == 8
    return np.max(norm_arr)/np.mean(norm_arr), np.argmax(norm_arr)
 
  def set_mean_var_anomalyThreshold(self):
    val_data = self.data_by_fold.get_validation_data()
    print('val_data shape', val_data.shape)
    val_rec_errors = self._get_pca_recs_error(val_data)
    self.rec_error_mean = val_rec_errors.mean(axis=0)
    self.rec_error_var = val_rec_errors.var(axis=0)
    scores = np.zeros(len(val_rec_errors))
    for i,val_error in enumerate(val_rec_errors):
      scores[i] = self.__rec_error_likelihood(val_error)[0]
    print('score quantiles 0.95, 0.99, 0.999, 0.9999, 1', np.quantile(scores, [0.95, 0.99, 0.999, 0.9999, 1]))
    #plt.close()
    #fig, ax = plt.subplots(1,2, figsize=(10,5))
    #ax[0].boxplot(scores)
    #ax[1].hist(scores)
    self.anomaly_threshold = np.quantile(scores, self.quantile)
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = PCA_data_by_fold(self.test_fold, self.val_fold, fold_ranges, steps=self.steps, dts_only=self.dts_only)
      print('calculating principal components...')
      self.pca = self.pca.fit(self.data_by_fold.get_training_data())
      print('PCA fit with {} components for {} features! Explained variance ratio: {}'.format(self.pca.n_components_, self.pca.n_features_, sum(self.pca.explained_variance_ratio_)))
      self.set_mean_var_anomalyThreshold()
  

  def get_score(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    if self.dts_only:
      df = df_dts.drop(columns=['sampling_interval'])
    else:
      df = df_dts_joined 
    df_orig = df_by_time_range(df, tr).drop(columns=['new_timestamp']).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    assert df_orig.shape[1] - 1 == self.data_by_fold.dim, '{} not equal {}'.format(df_orig.shape[1], self.data_by_fold.dim)
    rec_errors = self._get_pca_recs_error(self.data_by_fold.get_input_from_df(df_orig.drop(columns='timestamp').values))
    scores = []
    for rec_error in rec_errors:
      scores.append(self.__rec_error_likelihood(rec_error)[0])
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
    if self.dts_only:
      df = df_dts.drop(columns=['sampling_interval'])
    else:
      df = df_dts_joined 
    df_orig = df_by_time_range(df, tr).drop(columns=['new_timestamp']).copy()
    assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
    df_orig[metadata['segments']] = data[metadata['segments']].values
    assert df_orig.shape[1] - 1 == self.data_by_fold.dim, '{} not equal {}'.format(df_orig.shape[1], self.data_by_fold.dim)
    rec_error = self._get_pca_recs_error(self.data_by_fold.get_input_from_df(df_orig.drop(columns='timestamp').values))
    is_anomaly = [(-1, -1) for _ in range(self.steps -1)]
    for rec in rec_error:
      is_anomaly.append(self._is_anomaly_from_reconstruction_error(rec))
    assert len(is_anomaly) == len(data), '{} not equal to {}'.format(len(is_anomaly), len(data))
    return is_anomaly

# COMMAND ----------

PCA_COMPONENTS = 150
STEPS = 3
DTS_ONLY = False
df_segment, segment = None, None
res = get_all_anomaly_scores(PCA_AnomalyDetector_multidim, '', fault_dir, fault_types =['typeI', 'typeII', 'typeIII', 'typeIV'],
                             save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA{}_steps{}_{}.json'.format(PCA_COMPONENTS, STEPS,
                                                                                                                                     'with_control' if not DTS_ONLY else 'dts_only'))

# COMMAND ----------

PCA_COMPONENTS = 150
STEPS = 3
DTS_ONLY = True
df_segment, segment = None, None
res = get_all_anomaly_scores(PCA_AnomalyDetector_multidim, '', fault_dir, fault_types =['typeI', 'typeII', 'typeIII', 'typeIV'],
                             save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/spatial/PCA{}_steps{}_{}.json'.format(PCA_COMPONENTS, STEPS,
                                                                                                                                     'with_control' if not DTS_ONLY else 'dts_only'))

# COMMAND ----------

result = json.loads(res)

tpr = result['true_positive_rates']['typeIII']['random']['2']
fpr = result['false_positive_rates']['typeIII']['random']['2']

#plt.close()
plt.plot(fpr, tpr, color='r')
plt.xlim([-0.05, 1.05])
plt.ylim([-0.05, 1.05])
display(plt.show())

# COMMAND ----------

columns = [6290, 17712, 32028, 'random']
PCA_components = 100
STEPS = 3
quantile = 1
dts_only = False

for column in columns:
  log = Log('/dbfs/mnt/renewablesstudents/user-write/models/p2/evaluation/column_{}'.format(column), '/PCA_{}_steps{}_withoutControl_v2'.format(PCA_components, STEPS))
  log.write('quantile: {}. dts_only: {}'.format(quantile, dts_only))
  if column != 'random':
    df_segment, segment  = load_segment_data(subset=SUBSET, column=column)
  anomaly_detector = PCA_AnomalyDetector_multidim(PCA_components, STEPS, quantile, dts_only=dts_only)
  ad_results = get_AD_results_V2(anomaly_detector, fault_dir, column=column, log=log)

# COMMAND ----------

#typeI: {'FP': 5, 'anomaly_detected': 1.0, 'TN': 8112, 'FN': 0, 'TP': 2917, '1st_found': 1.0, 'healty_detected': 0.038461538461538464} (precision: 0.9982888432580425, recall: 1.0, F: 0.9983057566998853)

# COMMAND ----------

display(plt.show())

# COMMAND ----------

fold = 0
fault_type = 'typeI'
column = 6290

fold_dir = fault_dir + 'test_fold{}/'.format(fold) + 'column_{}/'.format(column)

data_path = glob.glob(fold_dir + fault_type + '/*.csv')[0]
metadata_path = glob.glob(fold_dir + fault_type + '/*.json')[0]
fault_df, metadata = get_fault_data(data_path, metadata_path)
normal_df = get_normal_data(metadata_path, column)

# COMMAND ----------

def get_dts_frame_from(data, metadata):
  tr = {"start_timestamp": pd.to_datetime(metadata['start']), "end_timestamp":pd.to_datetime(metadata['end'])}
  df_orig = df_by_time_range(df_dts, tr).copy()
  assert np.all(data.timestamp.values == df_orig.timestamp.values), "time ranges must match"  
  df_orig[metadata['segments']] = data[metadata['segments']].values
  return dbf.normalize(df_orig.iloc[100,:n_segments].values)
  
f_orig= get_re(fault_df, metadata) 
n_orig= get_re(normal_df, metadata)
f_orig.shape, n_orig.shape

# COMMAND ----------

s_dbf = Spatial_data_by_fold(0, 1, fold_ranges)
pca = PCA(n_components = 100, svd_solver='auto')
dts_norm = s_dbf.normalize(df_dts.iloc[:,:n_segments].values)
