# Databricks notebook source
# MAGIC %run ./func_spatial

# COMMAND ----------

SUBSET = True

fold_ranges = get_fold_ranges(subset=SUBSET)  
df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']

# COMMAND ----------

def convolutionalAutoencoder(lr, log=False):
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()

  x = tf.placeholder("float", [None, n_segments])
  bs = tf.placeholder("int32")

  inp = tf.reshape(x, shape=[-1, n_segments, 1])  
  log.write('learning rate: {}'.format(lr))
  log.write('input {}'.format(inp))
      
  n_filters = 30
  ksize1 = [5, 1, n_filters]
  ksize2 = [5, n_filters, n_filters]
  ksize3 = [5, n_filters, n_filters]
  ksize4 = [5, n_filters, 4]
  p1 = 6
  p2 = 6
  p3 = 6
  p4 = 5
  
  log.write('ksize 1: {}, max pool 1: {}'.format(ksize1, p1))
  log.write('ksize 2: {}, max pool 2: {}'.format(ksize2, p2))
  log.write('ksize 3: {}, max pool 3: {}'.format(ksize3, p3))
  log.write('ksize 4: {}, max pool 4: {}'.format(ksize4, p4))

  W = tf.get_variable('W', shape=(n_filters,1), initializer=tf.random_normal_initializer())
  B = tf.get_variable('B', shape=(1), initializer=tf.zeros_initializer())

  c1, s1 = conv(inp, ksize1, 1, 'conv1')
  log.write('conv1: {}'.format(c1))
  mp1 = max_pool(c1, p1, p1, 'mp1', ksize1[2], bs)
  log.write('mp1: {}'.format(mp1))
  c2, s2 = conv(mp1, ksize2, 1, 'conv2')
  log.write('conv2: {}'.format(c2))
  mp2 = max_pool(c2, p2, p2, 'mp2', ksize2[2], bs)
  log.write('mp2: {}'.format(mp2))
  c3, s3 = conv(mp2, ksize3, 1, 'conv3')
  log.write('c3: {}'.format(c3))
  mp3 = max_pool(c3, p3, p3, 'mp3', ksize3[2], bs)
  log.write('mp3: {}'.format(mp3))
  c4, s4 = conv(mp3, ksize4, 1, 'conv4')
  log.write('c4: {}'.format(c4))
  mp4 = max_pool(c4, p4, p4, 'mp4', ksize4[2], bs)
  log.write('mp4: {}'.format(mp4))
  
  encoding = tf.identity(mp4, name='encoding')
  
  dc4 = deconv(mp4, ksize4, s4, 'deconv4', int(np.ceil(n_segments/(p1*p2*p3*p4))), bs)
  log.write('dc4: {}'.format(dc4))
  up4 = upsample(dc4, 'unpool4', int(np.ceil(n_segments/(p1*p2*p3))), ksize4[1], bs)
  log.write('up4: {}'.format(up4))
  dc3 = deconv(up4, ksize3, s3,'deconv3', int(np.ceil(n_segments/(p1*p2*p3))), bs)
  log.write('dc3: {}'.format(dc3))
  up3 = upsample(dc3, 'unpool0', int(np.ceil(n_segments/(p1*p2))), ksize3[1], bs)
  log.write('up3: {}'.format(up3))
  dc2 = deconv(up3, ksize2, s2,'deconv2', int(np.ceil(n_segments/(p1*p2))), bs)
  log.write('dc2: {}'.format(dc2))
  up2 = upsample(dc2, 'unpool1', n_segments//p1, ksize2[1], bs)
  log.write('up2: {}'.format(up2))
  dc1 = deconv(up2, [ksize1[0],  ksize2[1], n_filters], s1,'deconv1', int(np.ceil(n_segments//p1)), bs)
  log.write('dc1: {}'.format(dc1))
  up1 = upsample(dc1, 'unpool1', n_segments, ksize1[2], bs)
  log.write('up1: {}'.format(up1))

  res = tf.matmul(tf.reshape(up1, shape = [-1, n_filters]), W) + B
  log.write('res: {}'.format(res))

  r = tf.reshape(res, shape=[-1, n_segments])

  loss_op = tf.losses.mean_squared_error(labels=x, predictions=r)
  optimizer = tf.train.AdamOptimizer(learning_rate=lr)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, r, x, bs
convolutionalAutoencoder(lr = 0.001)

# COMMAND ----------

'''
def train_AE_network(data_by_fold, n_epochs=10, batch_size=128, initial_patience=50, file_suffix=False, all_test_folds=True, log=False):
  if not file_suffix:
    print("MODELS ARE NOT SAVED")
  if not log:
    log = Log(False, False)
  model_dir = models_dir + str(file_suffix)
  model_summary(log)

  for test_fold in [2,3,4]:
    val_fold = (test_fold+1) % n_folds
    if val_fold != test_fold:
      log.write('test fold: {}, val fold: {}'.format(test_fold, val_fold))
      patience = initial_patience
      sd = data_by_fold(test_fold, val_fold, fold_ranges)
      train_data = sd.get_training_data()
      val_data = sd.get_validation_data()      
      min_val_loss = 1
      epoch_inputs = train_data
      with tf.Session() as sess:
        if test_fold == 2:
          saver.restore(sess, '/dbfs/mnt/renewablesstudents/user-write/models/p2/spatial_AE/allData/CAE_4650_allDataApril/params_t2v3/model.ckpt')# 
          start_epoch = 208
          min_val_loss = 0.0035830
        else:
          sess.run(tf.global_variables_initializer())
          start_epoch = 0
        log.write('Prior val loss {}'.format(get_ae_loss(sess, val_data)))
        for epoch in range(start_epoch, n_epochs):
          np.random.shuffle(epoch_inputs)
          for batch in range(0, len(epoch_inputs), batch_size):
            batch_inputs = epoch_inputs[batch: batch+batch_size, :]
            sess.run(train_op, feed_dict={x:batch_inputs, bs:len(batch_inputs)})
          epoch_loss = get_ae_loss(sess, train_data)
          epoch_val_loss =  get_ae_loss(sess, val_data)
          log.write('epoch {} finished with train loss {}, val loss {}'.format(epoch, epoch_loss, epoch_val_loss))
          if epoch_val_loss < min_val_loss:
            min_val_loss = epoch_val_loss
            patience = initial_patience
            if file_suffix:
              params_path = save_params(sess, epoch, test_fold, val_fold, log, model_dir)
            else:
              print('calculated val_recs')
              val_recs = get_reconstructions(sess, val_data)
          else:
            patience -= 1
            if patience == 0:
              log.write('Breaking due to no val improvement in {} epochs'.format(initial_patience))
              break
  log.write("Done!")
  if not file_suffix:
    return val_recs
  else:
    return None
'''

# COMMAND ----------

logs_dir, models_dir = set_dirs('test')

file_suffix = False#'CAE_148'
n_epochs = 1000

log = False#Log(logs_dir, file_suffix)
train_op, r, x, bs = convolutionalAutoencoder(lr = 0.001, log=log)
saver = tf.train.Saver()
train_AE_network(data_by_fold=Spatial_data_by_fold, n_epochs=n_epochs, file_suffix=file_suffix, all_test_folds=True, log=log, print_loss_op=False)

# COMMAND ----------


