# Databricks notebook source
# MAGIC %run ./func_temporal

# COMMAND ----------

# MAGIC %run ./func_AD_evaluation

# COMMAND ----------

from sklearn.linear_model import LinearRegression

# COMMAND ----------

SUBSET = False

# COMMAND ----------

fold_ranges = get_fold_ranges(subset=SUBSET)  
df_current = load_current_data(subset=SUBSET)
df_seatemp = load_seatemp_data(subset=SUBSET)

df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']

print(df_current.new_timestamp.iloc[-1], df_seatemp.new_timestamp.iloc[-1])

# COMMAND ----------

df_seg = df_dts.iloc[6290, :]

# COMMAND ----------

index = 5916
length = 10

plt.close()
fig, ax = plt.subplots(1,1, figsize=(6,1.5))
ax.plot(df_seg.values[index:index+length][::-1], '.-', color='green')
ax.set_xlim(-1,length + 1)
ax.set_ylim(23.45, 23.7)

display(plt.show())

# COMMAND ----------

def LR_input(input_, interval):
  return np.concatenate([np.reshape(input_, (input_.shape[0], -1)), interval], 1)
def LR_MSE(reg, input_, target, interval):
  return np.mean(np.square(reg.predict(LR_input(input_, interval))-target))
def get_LR_preds(reg, input_, interval):
  return reg.predict(LR_input(input_, interval))
def get_LR_pred_error_vector(reg, input_, target, interval):
  #pred = get_LR_preds(reg, input_, interval)
  #if global_fault_seq:
  #  print('pred', pred.shape)
  #  global norm_preds
  #  norm_preds = pred
  return get_LR_preds(reg, input_, interval) - target

# COMMAND ----------

# Print LR MSE results per fold.

'''
for test_fold in range(n_folds):
  val_fold = (test_fold+1) % n_folds
  if val_fold != test_fold:
    print('test fold: {}, val fold: {}'.format(test_fold, val_fold))
    td = Temporal_prediction_data_by_fold(test_fold, val_fold, fold_ranges)
    train_inputs, train_targets, train_intervals = td.get_training_data()
    reg = LinearRegression().fit(LR_input(train_inputs, train_intervals), train_targets)
    val_inputs, val_targets, val_intervals = td.get_validation_data()          
    print('train:', LR_MSE(reg, train_inputs, train_targets, train_intervals), 'val:', LR_MSE(reg, val_inputs, val_targets, val_intervals))
'''
pass

# COMMAND ----------

class LR_AnomalyDetector(AnomalyDetector):
  """
  spatial AD model.
  """
  def __init__(self, models_dir=None):
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.preds_error_mean = None
    self.lu_piv = None
    self.anomaly_threshold = None
    
    self.reg = None

  def _is_anomaly_from_prediction_error(self, preds_error):
    anomaly_likelihood_score = self.__pred_error_likelihood(preds_error)
    #print('anomaly likelihood score', anomaly_likelihood_score)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return column, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def __pred_error_likelihood(self, pred_errors):
    tmp = (pred_errors - self.preds_error_mean).T    
    return tmp.T @ lu_solve(self.lu_piv, tmp)
  
  def set_mean_var_anomalyThreshold(self):
    
    train_inputs, train_targets, train_intervals = self.data_by_fold.get_training_data()
    #print('train_inputs, train_targets, train_intervals', train_inputs.shape, train_targets.shape, train_intervals.shape)
    #print('input, target', LR_input(train_inputs, train_intervals).shape, train_targets.reshape((len(train_targets), -1)).shape)
    
    self.reg = LinearRegression().fit(LR_input(train_inputs, train_intervals), train_targets.reshape((len(train_targets), -1)))
    
    val_inputs, val_targets, val_intervals = self.data_by_fold.get_validation_data()
    print('train:', LR_MSE(self.reg, train_inputs, train_targets.reshape((len(train_targets), -1)), train_intervals), 'val:', LR_MSE(self.reg, val_inputs, val_targets.reshape((len(val_targets), -1)), val_intervals))
    
    print('val_inputs, val_targets, val_intervals', val_inputs.shape, val_targets.shape, val_intervals.shape)
    print('LR input', LR_input(val_inputs, val_targets.reshape((len(val_targets),-1))).shape)
    val_pred_errors = get_LR_pred_error_vector(self.reg, val_inputs, val_targets.reshape((len(val_targets), -1)), val_intervals)
    self.preds_error_mean = val_pred_errors.mean(axis=0)
    preds_error_cov = np.atleast_2d(np.cov(val_pred_errors, rowvar=False))
    self.lu_piv = lu_factor(preds_error_cov)
    scores = np.zeros(len(val_pred_errors))
    for i,val_error in enumerate(val_pred_errors):
      scores[i] = self.__pred_error_likelihood(val_error)
    self.anomaly_threshold = np.quantile(scores, 0.999)
    print('anomaly threshold set to:', self.anomaly_threshold)

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = Temporal_prediction_data_by_fold(self.test_fold, self.val_fold, fold_ranges)
      self.set_mean_var_anomalyThreshold()
      #dump_path = '/dbfs/mnt/renewablesstudents/user-write/test/LR_reg_{}.json'.format(INPUT_STEPS)
      #with open(dump_path, 'a') as file:
      #  json.dump(self.reg.coef_.tolist(), file)
      #  print('coef_s dumped to ' + dump_path)
      
  def get_score(self, data, metadata,test_fold):
    self._update_test_fold_if_changed(test_fold)
    input_dataframe = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs, targets, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(input_dataframe)
    pred_errors = get_LR_pred_error_vector(self.reg, inputs, targets.reshape((len(targets),-1)), intervals)
    #pred_errors = pred_errors.reshape((len(pred_errors), -1))
    scores = []
    for pred_error in pred_errors:
      scores.append(self.__pred_error_likelihood(pred_error))
    #global global_fault_seq
    #global global_norm_seq
    #global global_norm_scores
    #global global_fault_scores
    #if not len(global_fault_seq):
    #  global_fault_seq = (inputs, targets, intervals)
    #  global_fault_scores = scores
    #else:
    #  global_norm_seq = (inputs, targets, intervals)
    #  global_norm_scores = scores
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    input_dataframe = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs, targets, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(input_dataframe)
    #print('shape targets', targets.shape)
    pred_errors = get_LR_pred_error_vector(self.reg, inputs, targets.reshape((len(targets),-1)), intervals)
    is_anomaly = [(-1, -1) for _ in range(INPUT_STEPS+OUTPUT_STEPS-1)]
    for pred_error in pred_errors:
      is_anomaly.append(self._is_anomaly_from_prediction_error(pred_error))
    assert len(is_anomaly) == len(data), 'is_anomaly should be same length as input dataframe. Was {} and {}. Check INPUT_STEPS ({}) and OUTPUT_STEPS({})'.format(len(is_anomaly), len(data), INPUT_STEPS, OUTPUT_STEPS)
    return is_anomaly

# COMMAND ----------

#global_norm_seq = None
#global_norm_scores = None
#global_fault_seq = np.asarray([])
#global_fault_scores = None
#norm_preds = []

df_segment, segment  = None, None
INPUT_STEPS = 50
OUTPUT_STEPS = 1
res = get_all_anomaly_scores(LR_AnomalyDetector, '', fault_dir, fault_types=['typeIV'],
#                               save_path=False, columns=[6290], folds=[0])
                               save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/temporal/IV/LR_i{}_o{}'.format(INPUT_STEPS, OUTPUT_STEPS))

# COMMAND ----------

len(result['score_list_normal']['typeIV']['6290']['0'])

# COMMAND ----------

result = json.loads(res)

tpr = result['true_positive_rates']['typeIV']['6290']['0']
fpr = result['false_positive_rates']['typeIV']['6290']['0']

plt.close()
plt.plot(fpr, tpr)
plt.xlim([-0.05, 1.05])
plt.ylim([-0.05, 1.05])
display(plt.show())

# COMMAND ----------

dump_dir = {'norm_preds':norm_preds.squeeze().tolist(),
            'global_norm_scores':global_norm_scores.squeeze().tolist()
           }

#dump_path = '/dbfs/mnt/renewablesstudents/user-write/test/LR_preds_score_28.json'.format(INPUT_STEPS)
      #with open(dump_path, 'a') as file:
      #  json.dump(dump_dir, file)
      #  print('coef_s dumped to ' + dump_path)

# COMMAND ----------

len(global_norm_scores), len(global_norm_seq), len(global_fault_seq), len(global_fault_scores)

f_inputs, f_targets, f_intervals = global_fault_seq
n_inputs, n_targets, n_intervals = global_norm_seq

curr_plot = f_inputs[:,0,1].tolist() + f_inputs[-1,1:,1].tolist()
st_plot = f_inputs[:,0,2].tolist() + f_inputs[-1,1:,2].tolist()

n_inputs.shape, n_targets.shape, n_intervals.shape 

# COMMAND ----------

norm_preds

# COMMAND ----------

global_norm_scores

# COMMAND ----------

plt.close()
fix, ax = plt.subplots(3,1, sharex=True, figsize=(10,4.7))
for a in ax:
  a.axvline(60, color='darkorange', linestyle=':')
ax[0].set_title('LR predictions')
ax[0].plot(np.concatenate([n_inputs[:,0,0], n_targets[-50:,0,0]]), '.', label='observed')
ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds.tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend()
ax[0].set_ylabel('DTS')
ax[1].plot(curr_plot, '.')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
ax[2].plot([i for i in range(50, len(global_norm_scores)+50)], global_norm_scores, 'r.')# color='red')
ax[2].set_ylabel('Anomaly score')
ax[2].set_xlabel('Time step')
display(plt.show())

# COMMAND ----------

plt.close()
fix, ax = plt.subplots(3,1, sharex=True, figsize=(10,4.7))
for a in ax:
  a.axvline(60, color='darkorange', linestyle=':')
ax[0].set_title('LR predictions')
ax[0].plot(np.concatenate([n_inputs[:,0,0], n_targets[-50:,0,0]]), '.', label='observed')
ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds.tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend()
ax[0].set_ylabel('DTS')
ax[1].plot(curr_plot, '.')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
ax[2].plot([i for i in range(50, len(global_norm_scores)+50)], global_norm_scores, color='red')
ax[2].set_ylabel('Anomaly score')
ax[2].set_xlabel('Time step')
display(plt.show())

# COMMAND ----------

plt.close()
fix, ax = plt.subplots(3,1, sharex=True, figsize=(10,4.7))
for a in ax:
  a.axvline(50, color='darkorange', linestyle=':')
ax[0].set_title('LR predictions')
ax[0].plot(np.concatenate([n_inputs[:,0,0], n_targets[-50:,0,0]]), '.', label='observed')
ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds.tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend()
ax[0].set_ylabel('DTS')
ax[1].plot(curr_plot, '.')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
ax[2].plot([i for i in range(50, len(global_norm_scores)+50)], global_norm_scores, 'r.')# color='red')
ax[2].set_ylabel('Anomaly score')
ax[2].set_xlabel('Time step')
display(plt.show())

# COMMAND ----------

with open ('/dbfs/mnt/renewablesstudents/user-write/test/LR_reg_50.json') as file:
  coefs = np.array(json.load(file)).squeeze()

# COMMAND ----------

interval = coefs[-1]
print('interval', interval)
reshaped = coefs[:-1].reshape(5, 4)

# COMMAND ----------

def plot_LR_coefficients(input_steps, display_plot=False):
  with open ('/dbfs/mnt/renewablesstudents/user-write/test/LR_reg_{}.json'.format(input_steps)) as file:
    coefs = np.array(json.load(file)).squeeze()
  if input_steps == 1 and display_plot:
    print(coefs, 'sum:', coefs.sum())
  interval = coefs[-1]
  print('interval for LR', input_steps, interval)
  reshaped = coefs[:-1].reshape(input_steps, 4)

  plt.close()
  plt.figure(figsize=(5,5))
  plt.plot(reshaped[:,0],  label='dts')
  plt.plot(reshaped[:,1], label='current')
  plt.plot(reshaped[:,2], label='seatemp')
  plt.plot(reshaped[:,3], label='int')
  plt.legend(loc=2)
  if display_plot and not input_steps==1:
    display(plt.show())
  print('Sum dts for LR{}: {}, avg {}, std {}'.format(input_steps, reshaped[:,0].sum(),reshaped[:,0].mean(), reshaped[:,0].std()))
  print('Sum current for LR{}: {}, avg {}, std {}'.format(input_steps, reshaped[:,1].sum(),reshaped[:,1].mean(), reshaped[:,1].std()))
  print('Sum seatemp for LR{}: {}, avg {}, std {}'.format(input_steps, reshaped[:,2].sum(),reshaped[:,2].mean(), reshaped[:,2].std()))
  print('Sum interval for LR{}: {}, avg {}, std {}'.format(input_steps, reshaped[:,3].sum(),reshaped[:,3].mean(), reshaped[:,3].std()))
  print('')
  
  
PLOT = False
plot_LR_coefficients(1, PLOT)
if not PLOT:
  plot_LR_coefficients(5, PLOT)
  plot_LR_coefficients(25, PLOT)
  plot_LR_coefficients(50, PLOT)
  plot_LR_coefficients(100, PLOT)

# COMMAND ----------

plot_LR_coefficients(5)

# COMMAND ----------

plot_LR_coefficients(25)

# COMMAND ----------

plot_LR_coefficients(50)

# COMMAND ----------

plot_LR_coefficients(100)

# COMMAND ----------

for column in []: #[6290, 17712, 32028]:
  log = Log('/dbfs/mnt/renewablesstudents/user-write/models/p2/evaluation/column_{}'.format(column), '/LR_outsteps{}_insteps{}_999quantile'.format(OUTPUT_STEPS, INPUT_STEPS))  
  df_segment, segment = load_segment_data(column, subset=SUBSET)
  print(segment)
  df_joined = join_temp_curr_seatemp(df_segment, df_current, df_seatemp)
  ad = LR_AnomalyDetector()
  ad_results = get_AD_results_V2(ad, fault_dir, column=column, log=log)

# COMMAND ----------

#dbutils.fs.rm('dbfs:/mnt/renewablesstudents/user-write/models/p2/evaluation/column_32028/LR_outstpes1_999quantile.txt')

# COMMAND ----------

# Interpret weights. 
