# Databricks notebook source
import tensorflow as tf
import tensorflow.contrib.rnn as rnn_cell

# COMMAND ----------

# MAGIC %run ./func_temporal

# COMMAND ----------

# MAGIC %run ./func_AD_evaluation

# COMMAND ----------

# Prediction models:
#1 step
#  log: /dbfs/mnt/renewablesstudents/user-write/logfiles/p2/temporal_model/LSTM_prediction/column_{}/one_step_predictions.txt
#  params: /dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/one_step_predictions/params_t{}v{}/model.ckpt

#3 steps
#  log: /dbfs/mnt/renewablesstudents/user-write/logfiles/p2/temporal_model/LSTM_prediction/column_{}/three_steps_predictions.txt 
#  params: /dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/three_steps_predictions/params_t{}v{}/model.ckpt

# With control variable:
#1 steps
#  log: /dbfs/mnt/renewablesstudents/user-write/logfiles/p2/temporal_model/LSTM_prediction/column_{}/LSTM_predictions_with_control_os1.txt
#  params: /dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/LSTM_predictions_with_control_os1/params_t{}v{}/model.ckpt

#3 steps
#  log: /dbfs/mnt/renewablesstudents/user-write/logfiles/p2/temporal_model/LSTM_prediction/column_{}/LSTM_predictions_with_control_os3.txt
#  params: /dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/LSTM_predictions_with_control_os3/pa

# COMMAND ----------

PREDICT_CONTROL = False
OUTPUT_STEPS = 1
INPUT_STEPS = 50

train_op, p, x, t, y = LSTM_prediction_graph(n_hidden=64, n_output_steps=OUTPUT_STEPS, predict_control=PREDICT_CONTROL)
saver = tf.train.Saver()

# COMMAND ----------

SUBSET = False

fold_ranges = get_fold_ranges(subset=SUBSET)
df_current  = load_current_data(subset=SUBSET)
df_seatemp  = load_seatemp_data(subset=SUBSET)

df_dts = load_DTS_data(subset=SUBSET)  # First 39516 columns segments, last 3 columns 'timestamp', 'new_timestamp', 'sampling_interval'
assert len(list(df_dts)) == n_segments + 3 and list(df_dts)[-3:] == ['timestamp', 'new_timestamp', 'sampling_interval']

# COMMAND ----------

def get_pred_error_from_saved_params(params_path, inputs, targets, intervals):
   with tf.Session() as sess:
    saver.restore(sess, params_path)
    pred = get_predictions(sess, inputs, intervals)
    #if global_fault_seq:
    #  norm_preds.append(pred)
    return targets - pred

class LSTMPredictionsAnomalyDetector():
  """
  spatial AD model.
  """
  def __init__(self, model_dir, dbf=False, quantile=0.999, th_multiplier=1):
    if not dbf:
      dbf = Temporal_prediction_data_by_fold
    self.model_dir = model_dir
    
    self.data_by_fold = None
    self.test_fold = -1
    self.val_fold = -1
    
    self.preds_error_mean = None
    self.lu_piv = None
    self.anomaly_threshold = None
    
    self.params_path = None
    self.dbf = dbf
    
    self.th_multiplier = th_multiplier
    self.quantile = quantile

  def _is_anomaly_from_prediction_error(self, preds_error):
    anomaly_likelihood_score = self.__pred_error_likelihood(preds_error)
    if anomaly_likelihood_score > self.anomaly_threshold:
      return column, anomaly_likelihood_score
    return -1, anomaly_likelihood_score
  
  def __pred_error_likelihood(self, pred_errors):
    tmp = (pred_errors - self.preds_error_mean).T    
    return tmp.T @ lu_solve(self.lu_piv, tmp)
  
  def set_mean_var_anomalyThreshold(self):
    val_inputs, val_targets, val_intervals = self.data_by_fold.get_validation_data()
    val_pred_errors = get_pred_error_from_saved_params(self.params_path, val_inputs, val_targets, val_intervals)
    val_pred_errors = val_pred_errors.reshape((len(val_pred_errors), -1))
    self.preds_error_mean = val_pred_errors.mean(axis=0)
    preds_error_cov = np.atleast_2d(np.cov(val_pred_errors, rowvar=False))
    self.lu_piv = lu_factor(preds_error_cov)
    scores = np.zeros(len(val_pred_errors))
    for i,val_error in enumerate(val_pred_errors):
      scores[i] = self.__pred_error_likelihood(val_error)
    self.anomaly_threshold = self.th_multiplier * np.quantile(scores, self.quantile)#scores.mean() + 3 * scores.std()
    print('anomaly threshold set to:', self.anomaly_threshold)
    
    
    test_inputs, test_targets, test_intervals = self.data_by_fold.get_test_data()
    test_pred_errors = get_pred_error_from_saved_params(self.params_path, test_inputs, test_targets, test_intervals)
    
    print('MSE val', np.mean(np.square(val_pred_errors)))
    print('MSE test', np.mean(np.square(test_pred_errors)))
    assert 5 == 7
    

  def _update_test_fold_if_changed(self, test_fold):
    if test_fold != self.test_fold:
      val_fold = (test_fold+1) % n_folds
      self.test_fold = test_fold
      self.val_fold = val_fold
      self.data_by_fold = self.dbf(self.test_fold, self.val_fold, fold_ranges)
      training_inputs, training_targets, training_intervals = self.data_by_fold.get_training_data()
      print('training inputs. mean: {}, max: {}'.format(training_inputs.mean(axis=(0,1)), training_inputs.max(axis=(0,1))))
      print('training targets mean:{}, max:{}'.format(training_targets.mean(), training_targets.max()))
      self.params_path = self.model_dir.format(test_fold, val_fold)
      self.set_mean_var_anomalyThreshold()
  
  def get_score(self,data,metadata,test_fold):
    self._update_test_fold_if_changed(test_fold)
    print('data', data.shape)
    input_dataframe = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs, targets, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(input_dataframe)
    pred_errors = get_pred_error_from_saved_params(self.params_path, inputs, targets, intervals)
    pred_errors = pred_errors.reshape((len(pred_errors), -1))
    scores = []
    for pred_error in pred_errors:
      scores.append(self.__pred_error_likelihood(pred_error))
    global global_fault_seq
    global global_norm_seq
    global global_norm_scores
    global global_fault_scores
    if not len(global_fault_seq):
      global_fault_seq = (inputs, targets, intervals)
      global_fault_scores = scores
    else:
      global_norm_seq = (inputs, targets, intervals)
      global_norm_scores = scores
    
    return np.max(scores)
  
  def is_anomalous(self, data, metadata, test_fold):
    self._update_test_fold_if_changed(test_fold)
    input_dataframe = join_temp_curr_seatemp(data, df_current, df_seatemp)
    inputs, targets, intervals = self.data_by_fold.get_inputs_targets_intervals_from_df(input_dataframe)
    pred_errors = get_pred_error_from_saved_params(self.params_path, inputs, targets, intervals)
    pred_errors = pred_errors.reshape((len(pred_errors), -1))
    is_anomaly = [(-1, -1) for _ in range(INPUT_STEPS+OUTPUT_STEPS-1)]
    for pred_error in pred_errors:
      is_anomaly.append(self._is_anomaly_from_prediction_error(pred_error))
    assert len(is_anomaly) == len(input_dataframe), 'is_anomaly should be same length as input dataframe. Check INPUT_STEPS ({}) and OUTPUT_STEPS({})'.format(INPUT_STEPS, OUTPUT_STEPS)
    return is_anomaly

# COMMAND ----------

models_dir ='/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/one_step_predictions/params_t{}v{}/model.ckpt'
# '/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/three_steps_predictions/params_t{}v{}/model.ckpt'
         #'/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/LSTM_predictions_with_control_os1/params_t{}v{}/model.ckpt'

PREDICT_CONTROL = False  
OUTPUT_STEPS = 1

#global_norm_seq = None
#global_norm_scores = None
#global_fault_seq = np.asarray([])
#global_fault_scores = None
#norm_preds = []
  
train_op, p, x, t, y = LSTM_prediction_graph(n_hidden=128 if PREDICT_CONTROL else 64, n_output_steps=OUTPUT_STEPS, predict_control=PREDICT_CONTROL)
saver = tf.train.Saver()
df_segment, segment, df_joined  = None, None, None
res = get_all_anomaly_scores(LSTMPredictionsAnomalyDetector, models_dir, fault_dir, fault_types =['typeIV'],
                             columns=[6290], folds=[1], save_path=False) 
                             #save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/temporal/IV/LSTM_predictions_i{}_o{}{}.json'.format(INPUT_STEPS, OUTPUT_STEPS, '_c' if PREDICT_CONTROL else ''))
norm_preds = np.asarray(norm_preds).squeeze()

# COMMAND ----------

 #'/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/one_step_predictions/params_t{}v{}/model.ckpt'
          # '/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/three_steps_predictions/params_t{}v{}/model.ckpt'
models_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/LSTM_predictions_with_control_os3/params_t{}v{}/model.ckpt'

PREDICT_CONTROL = True
OUTPUT_STEPS = 3
  
train_op, p, x, t, y = LSTM_prediction_graph(n_hidden=128 if PREDICT_CONTROL else 64, n_output_steps=OUTPUT_STEPS, predict_control=PREDICT_CONTROL)
saver = tf.train.Saver()
df_segment, segment, df_joined  = None, None, None
res = get_all_anomaly_scores(LSTMPredictionsAnomalyDetector, models_dir, fault_dir, fault_types =['typeIV'],
                             #columns=[6290], folds=[1], save_path=False) 
                             save_path='/dbfs/mnt/renewablesstudents/user-write/anomaly_scores/temporal/IV/LSTM_predictions_i{}_o{}{}.json'.format(INPUT_STEPS, OUTPUT_STEPS, '_c' if PREDICT_CONTROL else ''))
norm_preds = np.asarray(norm_preds).squeeze()

# COMMAND ----------

len(global_norm_scores), len(global_norm_seq), len(global_fault_seq), len(global_fault_scores)

f_inputs, f_targets, f_intervals = global_fault_seq
n_inputs, n_targets, n_intervals = global_norm_seq

curr_plot = f_inputs[:,0,1].tolist() + f_inputs[-1,1:,1].tolist()
st_plot = f_inputs[:,0,2].tolist() + f_inputs[-1,1:,2].tolist()

n_inputs.shape, n_targets.shape, n_intervals.shape 

# COMMAND ----------

np.argmax(global_norm_scores)

# COMMAND ----------

norm_preds.shape

# COMMAND ----------

plt.close()
fix, ax = plt.subplots(3,1, sharex=True, figsize=(10,4.7))
for a in ax[:-1]:
  a.axvline(60, color='darkorange', linestyle=':')
ax[0].set_title('LSTM predictions')
ax[0].plot(np.concatenate([n_inputs[:,0,0], n_targets[-50:,0,0]]), '.', label='observed')
ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds[:,0].tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend()
ax[0].set_ylabel('DTS')
ax[1].plot(curr_plot, '.')
ax[1].plot([i for i in range(50, len(norm_preds)+50)], norm_preds[:,1].tolist(),'r.', alpha=0.9, label='predicted')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
ax[2].plot([i for i in range(50, len(global_norm_scores)+50)], global_norm_scores, color='red')
ax[2].set_ylabel('Anomaly score')
ax[2].set_xlabel('Time step')
display(plt.show())

# COMMAND ----------

plt.close()
fix, ax = plt.subplots(3,1, sharex=True, figsize=(10,4.7))
for a in ax:
  a.axvline(60, color='darkorange', linestyle=':')
ax[0].set_title('LSTM predictions')
ax[0].plot(np.concatenate([n_inputs[:,0,0], n_targets[-50:,0,0]]), '.', label='observed')
ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds.tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend()
ax[0].set_ylabel('DTS')
ax[1].plot(curr_plot, '.')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
ax[2].plot([i for i in range(50, len(global_norm_scores)+50)], global_norm_scores, 'r.')# color='red')
ax[2].set_ylabel('Anomaly score')
ax[2].set_xlabel('Time step')
display(plt.show())

# COMMAND ----------

plt.close()
fix, ax = plt.subplots(3,1, sharex=True, figsize=(10,4.7))
for a in ax:
  a.axvline(50, color='darkorange', linestyle=':')
ax[0].set_title('LSTM predictions')
ax[0].plot(np.concatenate([n_inputs[:,0,0], n_targets[-50:,0,0]]), '.', label='observed')
ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds.tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend()
ax[0].set_ylabel('DTS')
ax[1].plot(curr_plot, '.')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
ax[2].plot([i for i in range(50, len(global_norm_scores)+50)], global_norm_scores, 'r.')# color='red')
ax[2].set_ylabel('Anomaly score')
ax[2].set_xlabel('Time step')
display(plt.show())

# COMMAND ----------

np.sort(global_norm_scores), np.argsort(global_norm_scores)

# COMMAND ----------

plt.close()
fig, axis = plt.subplots(2,2)
axis[0][0].plot(global_norm_seq[segment].values)
axis[1][0].plot([0 for _ in range(50)] + global_norm_scores)
axis[0][1].plot(global_fault_seq[segment].values)
axis[1][1].plot([0 for _ in range(50)] + global_fault_scores)
display(plt.show())

# COMMAND ----------

for column in []: #[6290, 17712, 32028]:
  log =  Log('/dbfs/mnt/renewablesstudents/user-write/models/p2/evaluation/column_{}'.format(column), '/LSTM_predictions_with_control_os{}'.format(OUTPUT_STEPS))
  models_dir = '/dbfs/mnt/renewablesstudents/user-write/models/p2/temporal_model/LSTM_prediction/column_{}/'.format(column) + 'LSTM_predictions_with_control_os1/params_t{}v{}/model.ckpt'
  log.write('models_dir: ' + models_dir)
  
  df_segment, segment  = load_segment_data(subset=SUBSET, column=column)
  df_joined = join_temp_curr_seatemp(df_segment, df_current, df_seatemp)
  ad = LSTMPredictionsAnomalyDetector(models_dir, dbf=Temporal_prediction_data_by_fold_with_control if PREDICT_CONTROL else Temporal_prediction_data_by_fold)
  _ = get_AD_results_V2(ad, fault_dir, column=column, log=log)

# COMMAND ----------

with open('/dbfs/mnt/renewablesstudents/user-write/logfiles/p2/temporal_model/LSTM_prediction/column_6290/one_step_predictions.txt') as file:
  fold_1 = False
  val_losses = []
  train_losses = []
  for line in file:
    if 'test fold: 1' in line:
      fold_1 = True
    if 'test fold: 2' in line:
      fold_1 = False
    if fold_1 and 'epoch ' in line:
      val_losses.append(float(line.split()[-1]))
      train_losses.append(float(line.split()[-4][:-1]))

plt.close()
plt.figure(figsize=(8,6))
plt.axvline(len(val_losses)-51, color='red')
plt.plot(train_losses, label='train loss')
plt.plot(val_losses, label='val loss')
plt.xlabel('epoch')
plt.ylabel('MSE')
plt.xlim((0,len(val_losses)))
plt.legend()
display(plt.show())
      

# COMMAND ----------

import matplotlib.pyplot as plt
import numpy as np

# COMMAND ----------

val_losses[-51], train_losses[-51]

# COMMAND ----------

np.min(val_losses), np.argmin(val_losses)

# COMMAND ----------

np.min(train_losses), np.argmin(train_losses)

# COMMAND ----------

df_segment.shape

# COMMAND ----------

t0 = df_segment.new_timestamp.iloc[0]
t1 = df_segment.new_timestamp.iloc[50]
t2 = df_segment.new_timestamp.iloc[51]

# COMMAND ----------



# COMMAND ----------



plt.close()
fix, ax = plt.subplots(2,1, sharex=True, figsize=(10,4.7))
ax[0].set_title('Proposed additional electric current input')
ax[0].plot(df_segment['new_timestamp'].iloc[1:51], df_segment[segment].iloc[1:51], '.', label='input')
ax[0].set_xlim(df_segment['new_timestamp'].iloc[0], df_segment['new_timestamp'].iloc[53])
ax[0].set_ylim(top=41.3)

ax[0].plot(df_segment['new_timestamp'].iloc[51:52], df_segment[segment].iloc[51:52], '.', color='red', label='prediction')
cuurr = df_current.loc[(df_current.new_timestamp <= t1) & (df_current.new_timestamp >= t0)]
ax[1].plot(cuurr.new_timestamp, cuurr.v, '.')
curr2 = df_current.loc[(df_current.new_timestamp <= t2) & (df_current.new_timestamp >= t1)]
ax[1].plot(curr2.new_timestamp, curr2.v, '.', color='b')


ax[0].axvline(df_segment['new_timestamp'].iloc[50])
ax[1].axvline(df_segment['new_timestamp'].iloc[50])
ax[0].axvline(df_segment['new_timestamp'].iloc[51], color='red')
ax[1].axvline(df_segment['new_timestamp'].iloc[51], color='red')

#ax[0].plot([i for i in range(50, len(norm_preds)+50)], norm_preds.tolist(),'r.', alpha=0.9, label='predicted')
ax[0].legend(loc=3)
ax[0].set_ylabel('DTS')
#ax[1].plot(curr_plot, '.')
ax[1].set_ylabel('Current')
#ax[2].plot([i for i in range(50, len(norm_preds)+50)], norm_preds - n_targets.squeeze())
#ax[2].set_ylim((-0.1, 0.2))
#ax[2].axhline(0, color='orange')
display(plt.show())

# COMMAND ----------


