# Databricks notebook source
# MAGIC %run ../func_p2

# COMMAND ----------

from scipy.linalg import lu_factor, lu_solve

# COMMAND ----------

class Temporal_prediction_data_by_fold_base(Data_by_fold):
  def __init__(self, test_fold, val_fold, fold_ranges, predict_control):
    super().__init__(test_fold, val_fold, fold_ranges, curr=True, st=True)
    self.segment_mean, self.segment_std = self._get_mean_std([df_by_time_range(df_segment, tr)[segment] for tr in self.train_ranges])
    self.interval_mean, self.interval_std = self._get_mean_std([df_by_time_range(df_joined, tr)['sampling_int'] for tr in self.train_ranges])
    self.predict_control = predict_control
  
  def __normalize(self, data, mean, std):
    return (data - mean) / std
  
  def __get_inputs_targets(self, data, interval_list):
    out_dim = 3 if self.predict_control else 1
    size = len(data) - INPUT_STEPS - OUTPUT_STEPS + 1
    dim = data.shape[1]
    inputs = np.zeros((size, INPUT_STEPS, dim))
    targets = np.zeros((size, OUTPUT_STEPS, out_dim))
    intervals = np.zeros((size, OUTPUT_STEPS))

    for i in range(size):
      inputs[i] = data[i:i+INPUT_STEPS,:]
      targets[i] = data[i+INPUT_STEPS:i+INPUT_STEPS+OUTPUT_STEPS, :out_dim]
      intervals[i] = interval_list[i+INPUT_STEPS:i+INPUT_STEPS+OUTPUT_STEPS]
    return inputs, targets, intervals
  
  def get_inputs_targets_intervals_from_df(self, df):
    df.loc[:,segment] = self.__normalize(df.loc[:,segment].values, self.segment_mean, self.segment_std)
    df.loc[:,'current'] = self.__normalize(df.loc[:,'current'].values, self.curr_mean, self.curr_std)
    df.loc[:,'sea_temp'] = self.__normalize(df.loc[:,'sea_temp'].values, self.st_mean, self.st_std)
    df.loc[:,'sampling_int'] = self.__normalize(df.loc[:,'sampling_int'].values, self.interval_mean, self.interval_std)
    return self.__get_inputs_targets(df.loc[:,[segment, 'current', 'sea_temp', 'sampling_int']].values, df['sampling_int'].values)
  
  def __get_input_data_by_ranges(self, ranges):
    inputs = []
    targets = []
    intervals = []
    for tr in ranges:
      range_df = df_by_time_range(df_joined, tr).loc[:,[segment, 'current', 'sea_temp', 'sampling_int']]
      #normalize each feature
      range_df.loc[:,segment] = self.__normalize(range_df[segment].values, self.segment_mean, self.segment_std)
      range_df.loc[:,'current'] = self.__normalize(range_df['current'].values, self.curr_mean, self.curr_std)
      range_df.loc[:,'sea_temp'] = self.__normalize(range_df['sea_temp'].values, self.st_mean, self.st_std)
      range_df.loc[:,'sampling_int'] = self.__normalize(range_df['sampling_int'].values, self.interval_mean, self.interval_std)
      range_inputs, range_targets, range_intervals = self.__get_inputs_targets(range_df.values, range_df['sampling_int'].values)

      inputs.append(range_inputs)
      targets.append(range_targets)
      intervals.append(range_intervals)
    return np.concatenate(inputs), np.concatenate(targets), np.concatenate(intervals)
  
  def get_training_data(self):
    return self.__get_input_data_by_ranges(self.train_ranges)
    
  def get_validation_data(self):
    return self.__get_input_data_by_ranges(self.val_ranges)
  
  def get_test_data(self):
    return self.__get_input_data_by_ranges(self.test_ranges)
  
class Temporal_prediction_data_by_fold(Temporal_prediction_data_by_fold_base):
  def __init__(self, test_fold, val_fold, fold_ranges):
    super().__init__(test_fold, val_fold, fold_ranges, False)

class Temporal_prediction_data_by_fold_with_control(Temporal_prediction_data_by_fold_base):
  def __init__(self, test_fold, val_fold, fold_ranges):
    super().__init__(test_fold, val_fold, fold_ranges, True)
  

# COMMAND ----------

def get_predictions(sess, inputs, intervals, BS=False, batch_size=batch_size):
  predictions = []
  for batch in range(0, len(inputs), batch_size):
    if BS:
      predictions.append(sess.run(p, feed_dict={x:inputs[batch:batch+batch_size], t:intervals[batch:batch+batch_size], bs:len(inputs[batch:batch+batch_size])}))
    else:
      predictions.append(sess.run(p, feed_dict={x:inputs[batch:batch+batch_size], t:intervals[batch:batch+batch_size]}))
  return np.concatenate(predictions)

def get_prediction_loss(sess, inputs, targets, intervals, BS, batch_size):
  predictions = get_predictions(sess, inputs, intervals, BS, batch_size)
  losses = np.square(targets - predictions)
  return np.mean(losses)

def trainPredictionNetwork(n_epochs=10, batch_size=128, initial_patience=50, file_suffix=False, all_test_folds=True, log=False, dbf=Temporal_prediction_data_by_fold, BS=False, predict_control=False, save_model=True, best_val=True):
  if not log:
    log = Log(False, False)
  model_dir = models_dir + str(file_suffix)
  model_summary(log)
  
  params = {'save_model':save_model, 'n_epochs':n_epochs, 'batch_size':batch_size, 'initial_patience':initial_patience, 'file_suffix':file_suffix, 'all_test_folds':all_test_folds,  'BS':BS, 'predict_control':predict_control, 'best_val':best_val}
  log.write(str(params))
  
  for test_fold in range(n_folds if all_test_folds else 1):
    val_fold = (test_fold+1) % n_folds
    if val_fold != test_fold:
      log.write('test fold: {}, val fold: {}'.format(test_fold, val_fold))
      val_predictions = False
      patience = initial_patience
      td = dbf(test_fold, val_fold, fold_ranges)
      train_inputs, train_targets, train_intervals = td.get_training_data()
      val_inputs, val_targets, val_intervals = td.get_validation_data() 
      if test_fold == 0: print("val inputs, targets, intervals:", val_inputs.shape, val_targets.shape, val_intervals.shape)
      min_val_loss = 1
      epoch_indicies = [i for i in range(len(train_inputs))]
      with tf.Sesget_prediction_losssion() as sess:
        sess.run(tf.global_variables_initializer())
        log.write('Prior val loss {}'.format(get_prediction_loss(sess, val_inputs, val_targets, val_intervals, BS, batch_size)))
        for epoch in range(n_epochs):
          np.random.shuffle(epoch_indicies)
          epoch_inputs = train_inputs[epoch_indicies]
          epoch_targets = train_targets[epoch_indicies]
          epoch_intervals = train_intervals[epoch_indicies]
          for batch in range(0, len(epoch_indicies), batch_size):
            batch_inputs = epoch_inputs[batch: batch+batch_size, :]
            batch_targets = epoch_targets[batch: batch+batch_size, :]
            batch_intervals = epoch_intervals[batch: batch+batch_size, :]
            if BS:
              sess.run(train_op, feed_dict={x:batch_inputs, y:batch_targets, t:batch_intervals, bs:len(batch_inputs)})
            else:
              sess.run(train_op, feed_dict={x:batch_inputs, y:batch_targets, t:batch_intervals})
          epoch_train_loss = get_prediction_loss(sess, train_inputs, train_targets, train_intervals, BS, batch_size)
          epoch_val_loss =  get_prediction_loss(sess, val_inputs, val_targets, val_intervals, BS, batch_size)
          log.write('epoch {} finished with train loss {}, val loss {}'.format(epoch, epoch_train_loss, epoch_val_loss))
          if epoch_val_loss < min_val_loss or not best_val:
            min_val_loss = epoch_val_loss
            patience = initial_patience
            if file_suffix and save_model:
              params_path = save_params(sess, epoch, test_fold, val_fold, log, model_dir)
            else:
              #pass
              val_predictions = get_predictions(sess, val_inputs, val_intervals, BS, batch_size)
              train_predictions = get_predictions(sess, train_inputs, train_intervals, BS, batch_size)
              print('preds calculated')
          else:
            patience -= 1
            if patience == 0:
              log.write('Breaking due to no val improvement in {} epochs'.format(initial_patience))
              print('val_predictions', val_predictions)
              if not val_predictions:
                val_predictions = get_predictions(sess, val_inputs, val_intervals, BS, batch_size)
                train_predictions = get_predictions(sess, train_inputs, train_intervals, BS, batch_size)
                print('preds calculated')
              break
  log.write("Done!")
  if not (file_suffix and save_model):

    return val_predictions, train_predictions
  else:
    return None, None

# COMMAND ----------

class Temporal_Autoencoder_data_by_fold(Data_by_fold):
  def __init__(self, test_fold, val_fold, fold_ranges):
    super().__init__(test_fold, val_fold, fold_ranges, curr=not ONLY_SEGMENT, st=not ONLY_SEGMENT)
    self.segment_mean, self.segment_std = self._get_mean_std([df_by_time_range(df_segment, tr)[segment] for tr in self.train_ranges])
    #if not ONLY_SEGMENT:
    #  self.interval_mean, self.interval_std = self._get_mean_std([df_by_time_range(df_joined, tr)['sampling_int'] for tr in self.train_ranges])
  
  def __normalize(self, data, mean, std):
    return (data - mean) / std
  
  def __get_inputs(self, data, interval_list=False, window_size=50):
    size = len(data) - window_size +1
    dim = data.shape[1]
    inputs = np.zeros((size, window_size, dim))
    for i in range(size):
      inputs[i] = data[i:i+window_size,:]
    return inputs
  
  def get_inputs(self, df):
    columns = [segment]
    if not ONLY_SEGMENT:
      df.loc[:,'current'] = self.__normalize(df.loc[:,'current'].values, self.curr_mean, self.curr_std)
      df.loc[:,'sea_temp'] = self.__normalize(df.loc[:,'sea_temp'].values, self.st_mean, self.st_std)
      #df.loc[:,'sampling_int'] = self.__normalize(df.loc[:,'sampling_int'].values, self.interval_mean, self.interval_std)
      columns.extend(['current', 'sea_temp']) #, 'sampling_int'])
    df.loc[:,segment] = self.__normalize(df.loc[:, segment].values, self.segment_mean, self.segment_std)
    return self.__get_inputs(df.loc[:, columns].values)
  
  def __get_input_data_by_ranges(self, ranges):
    inputs = []
    for tr in ranges:
      if ONLY_SEGMENT:
        range_df = df_by_time_range(df_segment, tr).loc[:,[segment]]
      else:
        range_df = df_by_time_range(df_joined, tr).loc[:,[segment, 'current', 'sea_temp', 'sampling_int']]        
      inputs.append(self.get_inputs(range_df))
    return np.concatenate(inputs)
  
  def get_training_data(self):
    return self.__get_input_data_by_ranges(self.train_ranges)
    
  def get_validation_data(self):
    return self.__get_input_data_by_ranges(self.val_ranges)

# COMMAND ----------

def LSTM_prediction_graph(predict_control, log=None, learning_rate=0.01, n_hidden=64, dim_input=4, n_input_steps=50, n_output_steps=3):
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()
  
  dim_output = 3 if predict_control else 1
  
  params = {'learning rate': learning_rate, 'LSTM hidden units':n_hidden, 'dim input': dim_input, 'input steps': n_input_steps, 'output steps': n_output_steps, 'predict_control':predict_control}
  log.write(str(params))
  
  # Graph input
  x = tf.placeholder("float", [None, n_input_steps, dim_input], name='x')  #input
  y = tf.placeholder("float", [None, n_output_steps, dim_output], name='y')  #target
  t = tf.placeholder("float", [None, n_output_steps], name='t')  #time


  # Define weights
  weights = {'out': tf.Variable(tf.random_normal([n_hidden+n_output_steps, n_output_steps*dim_output]))}   # +n_output_steps because of concatenated time until prediction input
  biases = {'out': tf.Variable(tf.zeros([n_output_steps*dim_output]))}

  cell = rnn_cell.LSTMCell(n_hidden, name='LSTM_cell')
  outputs, _ = tf.nn.dynamic_rnn(cell, x, dtype=tf.float32)  
  p = tf.matmul(tf.concat([outputs[:, -1, :], t], axis=1), weights['out']) + biases['out']
  p = tf.identity(p, name='p')
  
  p = tf.reshape(p, shape=[-1, n_output_steps, dim_output])

  # Define loss and optimizer
  loss_op = tf.losses.mean_squared_error(labels=y, predictions=p)
  optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, p, x, t, y

# COMMAND ----------

def LSTM_autoencoder_graph(lr, log=None,  window_size=50):
  dim_input=1 if ONLY_SEGMENT else 3
  n_hidden=64 if ONLY_SEGMENT else 128
  if not log:
    log = Log(False, False)
  tf.reset_default_graph()
  
  params = {'learning rate': lr, 'LSTM hidden units':n_hidden, 'dim input': dim_input, 'window size': window_size}
  log.write('--------')
  log.write(str(params))
  
  # Graph input
  x = tf.placeholder(tf.float32, [None, window_size, dim_input], name='x')
  # Batch size
  bs = tf.placeholder(tf.int32, [])

  # Define weights
  out_weights =  tf.Variable(tf.random_normal([n_hidden, dim_input]))
  out_biases = tf.Variable(tf.zeros([dim_input]))

  encoder_cell = rnn_cell.LSTMCell(n_hidden, name='encoder_LSTM_cell')
  _, encoder_state = tf.nn.dynamic_rnn(encoder_cell, tf.reverse(x, axis=[1]), dtype=tf.float32)
  
  # TODO experiment with what state to keep
  decoder_cell = rnn_cell.LSTMCell(n_hidden, name='decoder_LSTM_cell')
  encoding = rnn_cell.LSTMStateTuple(decoder_cell.zero_state(bs, tf.float32)[0], encoder_state[1])
  log.write('encoding: {}'.format(encoding))  
  
  initial_input = tf.fill((bs, dim_input), 0.0)
  print('initial_input', initial_input)
  
  #Training: Use true inputs.
  state = encoding
  training_outputs = []
  training_inputs = [initial_input] + [x[:,i,:] for i in range(window_size-1)]
  print('len training inputs', len(training_inputs))
  for inp in training_inputs:
    output_, state = decoder_cell(inp, state)
    output = tf.matmul(output_, out_weights) + out_biases
    training_outputs.append(output)
  
  # Inference: Feed back predicted value.
  state = encoding
  inference_outputs = []
  input_ = initial_input
  for i in range(window_size):
    output, state = decoder_cell(input_, state)
    input_ = tf.matmul(output, out_weights) + out_biases
    inference_outputs.append(input_)

  print('len inference_outputs, training_outputs', len(inference_outputs), len(training_outputs))  
  
  r = inference_outputs

  labels = tf.transpose(x[:,:,:], [1,0,2])
  train_r = training_outputs
  print('labels', labels)
  
  # Define loss and optimizer
  loss_op = tf.losses.mean_squared_error(labels=labels, predictions=train_r)
  optimizer = tf.train.AdamOptimizer(learning_rate=lr)
  train_op = optimizer.minimize(loss_op)
  
  return train_op, r, x, bs, loss_op

#train_op, r, x, bs, loss_op = LSTM_autoencoder_graph(0.001)
